
/*****************************************************************************
 *  California Instruments Cs Series AC Current Source Instrument Driver                               
 *  LabWindows/CVI 2009 Instrument Driver                                     
 *  Original Release: Monday, February 28, 2011
 *  By: Kevin Lam                              
 *  PH. (858)450-0085   Fax (858)458-0257                              
 *                                                                           
 *****************************************************************************/
 
#include <string.h>
#include <stdio.h>  
#include <ctype.h>
#include <stdlib.h> 
#include <limits.h> 
#include <math.h>
#include <formatio.h> 
#include "ciCs.h"

#if defined (_CVI_)
    #include <utility.h>
#elif defined (_MSC_VER)
    #include <windows.h>

#define Delay(timeInSeconds) Sleep((DWORD)((timeInSeconds)*1000));
#else
#error Delay is not defined for selected compiler
#endif

#define MULTI_THREAD_USED
#define PI                          3.14159265

/*****************************************************************************
 *--------------------- Hidden Attribute Declarations -----------------------*
 *****************************************************************************/

#define CICS_ATTR_OPC_TIMEOUT               				(IVI_SPECIFIC_PRIVATE_ATTR_BASE + 1L)   /* ViInt32 */
#define CICS_ATTR_INSTR_MODEL_TYPE          				(IVI_SPECIFIC_PRIVATE_ATTR_BASE + 2L)   /* ViInt32 */
#define CICS_ATTR_INSTR_ACTIVE_PHASE        				(IVI_SPECIFIC_PRIVATE_ATTR_BASE + 3L)   /* ViInt32 */
#define CICS_ATTR_INSTR_PHASE_COUPLING      				(IVI_SPECIFIC_PRIVATE_ATTR_BASE + 4L)   /* ViInt32 */
#define CICS_ATTR_INSTR_VOLTAGE_RANGE        				(IVI_SPECIFIC_PRIVATE_ATTR_BASE + 5L)   /* ViReal64, Phase-Based */
#define CICS_ATTR_INSTR_CURRENT_RANGE        				(IVI_SPECIFIC_PRIVATE_ATTR_BASE + 6L)   /* ViReal64, Phase-Based */
#define CICS_ATTR_INSTR_VOLTAGE_LIMIT_BEHAVIOR 				(IVI_SPECIFIC_PRIVATE_ATTR_BASE + 7L)  	/* ViInt32 */
#define CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT  				(IVI_SPECIFIC_PRIVATE_ATTR_BASE + 8L)  /* ViInt32 */
#define CICS_ATTR_INSTR_OVP_ENABLED         				(IVI_SPECIFIC_PRIVATE_ATTR_BASE + 9L)  /* ViBoolean */
#define CICS_ATTR_INSTR_OUTPUT_PHASE_C  					(IVI_SPECIFIC_PRIVATE_ATTR_BASE + 10L)  	/* ViInt32 */

ViInt32 	busInterface;
ViReal64	serialDelay=0.05;

/**
 * Define supported instrument models.
 */
enum
{
    NOT_SUPPORTED = -1,
    CI3000Cs = 3000,
    CI4500Cs = 4500,
    CI6000Cs = 6000,
    CI9000Cs = 9000,
    CI12000Cs = 12000,
    CI13500Cs = 13500,
    CI18000Cs = 18000,
    
} ESupportedInstruments;
    
/**
 * Define table of the supported instrument models
 * and their corresponding strings in IDN?
 */
static IviStringValueTable gs_ciCsModelTable =
{
        {CI3000Cs,  	"3000CS"},
        {CI4500Cs,  	"4500CS"},
        {CI6000Cs,  	"6000CS"},
        {CI9000Cs,  	"9000CS"},
        {CI12000Cs,  	"12000CS"},
        {CI13500Cs,  	"13500CS"},
        {CI18000Cs,  	"18000CS"},
        {VI_NULL,   	VI_NULL }
};

#define IS_3000Cs 	(model_type == CI3000Cs)
#define IS_4500Cs 	(model_type == CI4500Cs)
#define IS_6000Cs 	(model_type == CI6000Cs)
#define IS_9000Cs 	(model_type == CI9000Cs)
#define IS_12000Cs 	(model_type == CI12000Cs)
#define IS_13500Cs 	(model_type == CI13500Cs)
#define IS_18000Cs 	(model_type == CI18000Cs)

#define SETUP_MODEL_TAG             "Model"

    /*- I/O buffer size -----------------------------------------------------*/
#define BUFFER_SIZE                             512L        
#define CICS_IO_SESSION_TYPE   IVI_VAL_VISA_SESSION_TYPE

#define DOUBLE_FORMAT_LENGTH                15          /* Length of string of formated double value (sprintf %Le))
                                                         * This value is 12, but for sure we are using 15 */

#define MAX_USER_DEFINED_WAVEFORMS          12
#define SWITCH_PHASE_DELAY_MILISECONDS      15000
    /*- 488.2 Event Status Register (ESR) Bits ------------------------------*/
#define IEEE_488_2_ERROR_BITS                   0x3C

    /*- Defined macros for enum attributes ----------------------------------*/
#define ciCs_GetCmdFromIntValue(value, table, cmd) \
        Ivi_GetViInt32EntryFromValue (value, table, VI_NULL, VI_NULL, VI_NULL, VI_NULL, cmd, VI_NULL)

#define ciCs_GetValueFromCmd(buffer, table, value) \
        Ivi_GetViInt32EntryFromString (buffer, table, value, VI_NULL, VI_NULL, VI_NULL, VI_NULL)

    /*- List of channels passed to the Ivi_BuildChannelTable function -------*/ 
#define PHASE_REP_CAP_NAME                  "Phase"
#define PHASE_LIST_3_PHASES                 "ALLPHASES,PHASE1,PHASE2,PHASE3"
#define PHASE_LIST_1_PHASE                  "PHASE1"
#define CHANNEL_LIST                        "1"

/*- Table of commands for enum attributes (ViInt32) -------------------------*/
static IviStringValueTable gs_EnumCommands =
{
    {CICS_ATTR_INSTR_PHASE_COUPLING,            			":INST:COUP"        },
    {CICS_ATTR_INSTR_VOLTAGE_LIMIT_BEHAVIOR,   				":VOLT:PROT:STAT"   },
    {CICS_ATTR_TRIGGER_SOURCE,                  			":TRIG:SOUR"        },
    {CICS_ATTR_TRIGGERED_CURRENT_MODE,          			":CURR:MODE"        },
    {CICS_ATTR_TRIGGERED_FREQUENCY_MODE,        			":FREQ:MODE"        },
    {CICS_ATTR_TRIGGERED_FUNCTION_MODE,         			":FUNC:MODE"        },
    {CICS_ATTR_TRIGGERED_PHASE_ANGLE_MODE,      			":PHAS:MODE"        },
    {CICS_ATTR_TRIGGERED_SLEW_CURRENT_RATE_MODE,			":CURR:SLEW:MODE"   },
    {CICS_ATTR_TRIGGERED_SLEW_FREQUENCY_RATE_MODE,			":FREQ:SLEW:MODE"   },
    {CICS_ATTR_TRIGGER_LIST_MODE,               			":LIST:STEP"        },
    {CICS_ATTR_ACQUISITION_TRIGGER_SOURCE,      			":TRIG:SEQ3:SOUR"   },
    {CICS_ATTR_WINDOW_TYPE,                     			":SENS:WIND"        },
    {CICS_ATTR_OUTPUT_TRIGGER_SOURCE,           			":OUTP:TTLT:SOUR"   },
    {CICS_ATTR_TRIGGER_SYNCHRONIZATION_SOURCE,  			":TRIG:SEQ2:SOUR"   },
    {CICS_ATTR_OUTPUT_PHASE_MODE,        		 			":SYST:CONF:NOUT"   },
    {CICS_ATTR_OUTPUT_ALC_STATE,          		 			":CURR:ALC"         },
    {VI_NULL,                                    						 VI_NULL}
};

/*- Table of commands for ViInt32 attributes --------------------------------*/
static IviStringValueTable gs_ViInt32Commands =
{
    {CICS_ATTR_INSTR_ACTIVE_PHASE,              ":INST:NSEL"        },
	{CICS_ATTR_INSTR_OUTPUT_PHASE_C,     	   	":LIM:PHASE"     	},
    {CICS_ATTR_TRIGGER_PULSE_COUNT,             ":PULS:COUN"        },
    {CICS_ATTR_TRIGGER_LIST_COUNT,              ":LIST:COUN"        },
    {VI_NULL,                                    VI_NULL  			}
};

/*- Table of commands for ViBoolean attributes ------------------------------*/
static IviStringValueTable gs_ViBooleanCommands =
{
    {CICS_ATTR_OUTPUT_ENABLED,                  ":OUTP"             },
    {CICS_ATTR_OUTPUT_TRIGGER_ENABLED,          ":OUTP:TTLT"        },

#if defined(OVERVOLTAGE_PROTECTION_STATE_SUPPORTED) 
    {CICS_ATTR_INSTR_OVP_ENABLED,               ":VOLT:PROT:STAT"   },
#endif
    {VI_NULL,                                    VI_NULL            }
};

/*- Table of commands for ViString attributes -------------------------------*/
static IviStringValueTable gs_ViStringCommands =
{
    {CICS_ATTR_FUNCTION,                        ":FUNC"     		},
    {CICS_ATTR_TRIGGERED_FUNCTION,              ":FUNC:TRIG"		},
    {CICS_ATTR_INSTR_USER_TABLE_LIST,           ":TRAC:CAT" 		},
	{CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT,        ":SYST:CONF:NOUT"	},
    {VI_NULL,                                    VI_NULL     		}
};

/*- Table of commands for ViReal64 attributes -------------------------------*/
static IviStringValueTable gs_ViReal64Commands =
{
	{CICS_ATTR_CURRENT_LEVEL,                   ":CURR"         	},
    {CICS_ATTR_VOLTAGE_LIMIT,		            ":VOLT"         	},
	{CICS_ATTR_PHASE_ANGLE,                     ":PHAS"         	},
    {CICS_ATTR_FREQUENCY,                       ":FREQ"         	},
    {CICS_ATTR_OVP_LIMIT,                       ":VOLT:PROT"    	},
    {CICS_ATTR_SLEW_CURRENT_RATE,               ":CURR:SLEW"    	},
    {CICS_ATTR_SLEW_FREQUENCY_RATE,             ":FREQ:SLEW"    	},
    {CICS_ATTR_CLIPPING_LEVEL,                  ":FUNC:CSIN THD"	},
    {CICS_ATTR_TRIGGER_DELAY,                   ":TRIG:DEL"     	},
    {CICS_ATTR_TRIGGERED_CURRENT_LEVEL,         ":CURR:TRIG"    	},
    {CICS_ATTR_TRIGGERED_FREQUENCY,             ":FREQ:TRIG"    	},
    {CICS_ATTR_TRIGGERED_PHASE_ANGLE,           ":PHAS:TRIG"    	},
    {CICS_ATTR_TRIGGERED_SLEW_CURRENT_RATE,     ":CURR:SLEW:TRIG"	},
    {CICS_ATTR_TRIGGERED_SLEW_FREQUENCY_RATE,   ":FREQ:SLEW:TRIG"	},
    {CICS_ATTR_TRIGGER_PULSE_WIDTH,             ":PULS:WIDT"    	},
    {CICS_ATTR_TRIGGER_PULSE_PERIOD,            ":PULS:PER"     	},
    {CICS_ATTR_ACQUISITION_TIME_INTERVAL,       ":SENS:SWE:TINT"	},
    {CICS_ATTR_INSTR_VOLTAGE_RANGE,             ":VOLT:RANG"    	},
    {CICS_ATTR_TRIGGER_SYNCHRONIZATION_PHASE,   ":TRIG:SEQ2:PHAS"	},
    {CICS_ATTR_OUTPUT_PROTECTION_DELAY,         ":OUTP:PROT:DEL"	},
    {CICS_ATTR_VOLTAGE_RANGE_MINIMUM,			":VOLT:RANG? MIN"	},
    {CICS_ATTR_VOLTAGE_RANGE_MAXIMUM,			":VOLT:RANG? MAX"	},
    {CICS_ATTR_VOLTAGE_MAXIMUM,				 	":VOLT? MAX"	 	},
    {CICS_ATTR_CURRENT_MAXIMUM,				 	":CURR? MAX"	 	},
    {CICS_ATTR_FREQUENCY_MINIMUM,				":FREQ? MIN"		},
    {CICS_ATTR_FREQUENCY_MAXIMUM,				":FREQ? MAX"		},
    
    {VI_NULL,                                    VI_NULL}
};

/*****************************************************************************
 *------ Typedefs for string/string tables used in function callbacks -------*
 *****************************************************************************/
typedef struct
        {
        ViString col1;
        ViString col2;
        } ciCsStringTableEntries;

typedef ciCsStringTableEntries* ciCsStringTable;

/*- List of predefined waverorms --------------------------------------------*/
static ciCsStringTableEntries functionStringTableEntries[] =
    {
        { CICS_VAL_SINE        ,   "SIN"    }, 
        { CICS_VAL_SQUARE      ,   "SQU"    }, 
        { CICS_VAL_CLIPPED_SINE,   "CSIN"   }, 
        { VI_NULL              ,    VI_NULL }
    };

static ciCsStringTable gs_FunctionStringTable = functionStringTableEntries;

/*****************************************************************************
 *-------------- Utility Function Declarations (Non-Exported) ---------------*
 *****************************************************************************/
static void RemoveSurroundingWhiteSpace(ViChar string[]);
static void RemoveQuotationMarks(ViChar string[]);

static ViStatus ciCs_GetDriverSetupOption(
    ViSession   vi,
    ViInt32*    pModelType
);

static ViStatus ciCs_GetIDNInformation(
    ViSession       vi,
    ViAttr          attributeId,
    ViConstString   formatString
);

static ViStatus ciCs_QueryID(
    ViSession   vi,
    ViSession   io,
    ViInt32     bufferSize,
    ViChar      buffer[]
);

static ViStatus ciCs_GetModelFromBuffer(
    ViSession       vi,
    ViConstString   buffer,
    ViInt32*        pModelType
);

static ViStatus ciCs_GetPhaseIndex(
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32*        pPhaseIndex,
    ViBoolean*      pIsSinglePhase
);
static ViStatus ciCs_SetActivePhase(
    ViSession       vi,
    ViConstString   phaseName
);

static ViStatus ciCs_CheckWaveformName(
    ViSession       vi,
    ViConstString   waveformName
);
static ViStatus ciCs_IsUserWaveformDefined(
    ViSession       vi,
    ViConstString   waveformName,
    ViBoolean*      isDefined
);

static ViStatus ciCs_GetNumberUserDefinedWaveforms(
    ViSession       vi,
    ViConstString   waveformName,
    ViInt32*        numberOfWaves
);

static ViStatus ciCs_GetStr2FromStr1 (
    ciCsStringTable strTable, 
    ViConstString   str1,
    ViConstString*  str2
);

static ViStatus ciCs_GetStr1FromStr2 (
    ciCsStringTable strTable, 
    ViConstString*  str1,
    ViConstString   str2
);

static ViStatus ciCs_GetMaximumRange (
    ViSession       vi,
    ViInt32         rangeType,
    ViReal64*       maximumValue
);

static ViStatus ciCs_GetMeasurementSimulation(
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         measurementType,
    ViReal64*       pMeasurement
);

static ViStatus ciCs_GetHarmonicMeasurementSimulation(
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         measurementType,
    ViInt32         harmonic,
    ViReal64*       measurement
);

static ViStatus ciCs_GetArrayMeasurementSimulation(
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         measurementType,
    ViInt32         arraySize,
    ViReal64        measurement[],
    ViInt32*        number_of_values
);

static ViStatus ciCs_ReadDataUnknownLength (
    ViSession       io,
    ViChar**        pdata,
    ViInt32*        readed_count
);

static ViStatus ciCsDisableOutputEnabledCache (
    ViSession       vi
);

static ViStatus ciCs_DefaultInstrSetup (ViSession openInstrSession);
static ViStatus ciCs_CheckStatus (ViSession vi);
static ViStatus ciCs_WaitForOPC (ViSession vi, ViInt32 maxTime);

static ViStatus ciCs_InitAttributes (
    ViSession   vi,
    ViInt32     modelType
);
static ViStatus ciCs_SetModelSpecificCallbacks(
    ViSession   vi,
    ViInt32     modelType
 );
 

/*****************************************************************************
 *----------------- Callback Declarations (Non-Exported) --------------------*
 *****************************************************************************/

    /*- Global Session Callbacks --------------------------------------------*/
    
static ViStatus _VI_FUNC ciCs_CheckStatusCallback (ViSession vi, ViSession io);
static ViStatus _VI_FUNC ciCs_WaitForOPCCallback (ViSession vi, ViSession io);

    /*- Shared callbacks ----------------------------------------------------*/
static ViStatus _VI_FUNC ciCsViInt32_ReadCallback (ViSession vi, ViSession io,
                                                    ViConstString channelName,
                                                    ViAttr attributeId,
                                                    ViInt32 *value);
static ViStatus _VI_FUNC ciCsViInt32_WriteCallback (ViSession vi, ViSession io,
                                                     ViConstString channelName,
                                                     ViAttr attributeId,
                                                     ViInt32 value);

static ViStatus _VI_FUNC ciCsViReal64_ReadCallback (ViSession vi, ViSession io,
                                                     ViConstString channelName,
                                                     ViAttr attributeId,
                                                     ViReal64 *value);
static ViStatus _VI_FUNC ciCsViReal64_WriteCallback (ViSession vi, ViSession io,
                                                      ViConstString channelName,
                                                      ViAttr attributeId,
                                                      ViReal64 value);

static ViStatus _VI_FUNC ciCsEnum_WriteCallback (ViSession vi, ViSession io,
                                                  ViConstString channelName,
                                                  ViAttr attributeId,
                                                  ViInt32 value);
static ViStatus _VI_FUNC ciCsEnum_ReadCallback (ViSession vi, ViSession io,
                                                 ViConstString channelName,
                                                 ViAttr attributeId,
                                                 ViInt32 *value);

static ViStatus _VI_FUNC ciCsViBoolean_WriteCallback (ViSession vi,
                                                       ViSession io,
                                                       ViConstString channelName,
                                                       ViAttr attributeId,
                                                       ViBoolean value);
static ViStatus _VI_FUNC ciCsViBoolean_ReadCallback (ViSession vi, ViSession io,
                                                      ViConstString channelName,
                                                      ViAttr attributeId,
                                                      ViBoolean *value);

static ViStatus _VI_FUNC ciCsViString_WriteCallback (ViSession vi, ViSession io,
                                                      ViConstString channelName,
                                                      ViAttr attributeId,
                                                      ViConstString value);
static ViStatus _VI_FUNC ciCsViString_ReadCallback (ViSession vi, ViSession io,
                                                     ViConstString channelName,
                                                     ViAttr attributeId,
                                                     const ViConstString cacheValue);

    /*- Attribute callbacks -------------------------------------------------*/
static ViStatus _VI_FUNC ciCsAttrDriverRevision_ReadCallback (
    ViSession           vi,
    ViSession           io, 
    ViConstString       phaseName,
    ViAttr              attributeId, 
    const ViConstString cacheValue
);

static ViStatus _VI_FUNC ciCsAttrInstrumentManufacturer_ReadCallback (
    ViSession           vi,
    ViSession           io, 
    ViConstString       phaseName,
    ViAttr              attributeId, 
    const ViConstString cacheValue
);
static ViStatus _VI_FUNC ciCsAttrInstrumentModel_ReadCallback (
    ViSession           vi,
    ViSession           io, 
    ViConstString       phaseName,
    ViAttr              attributeId, 
    const ViConstString cacheValue
);
static ViStatus _VI_FUNC ciCsAttrFirmwareRevision_ReadCallback (
    ViSession           vi, 
    ViSession           io,
    ViConstString       phaseName, 
    ViAttr              attributeId,
    const ViConstString cacheValue
);
static ViStatus _VI_FUNC ciCsAttrIdQueryResponse_ReadCallback (ViSession vi,
                                                                ViSession io,
                                                                ViConstString channelName,
                                                                ViAttr attributeId,
                                                                const ViConstString cacheValue);

static ViStatus _VI_FUNC ciCsAttrOutputPhaseCount_WriteCallback (ViSession vi,
                                                                  ViSession io,
                                                                  ViConstString channelName,
                                                                  ViAttr attributeId,
                                                                  ViInt32 value);

static ViStatus _VI_FUNC ciCsAttrOvpEnabled_WriteCallback (
    ViSession       vi, 
    ViSession       io,
    ViConstString   phaseName, 
    ViAttr          attributeId,
    ViBoolean       value
);
static ViStatus _VI_FUNC ciCsAttrOvpEnabled_ReadCallback (
    ViSession       vi, 
    ViSession       io,
    ViConstString   phaseName, 
    ViAttr          attributeId,
    ViBoolean*      value
);

static ViStatus _VI_FUNC ciCsAttrOvpEnabled_CheckCallback(
    ViSession       vi,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViBoolean       value
);

static ViStatus _VI_FUNC ciCsPhaseNameViReal64_CheckCallback(
    ViSession       vi,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViReal64        value
);
static ViStatus _VI_FUNC ciCsAttrOVPLimit_CoerceCallback(
    ViSession       vi,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViReal64        value,
    ViReal64*       coercedValue
);

static ViStatus _VI_FUNC ciCsAttrVoltageLimitBehavior_ReadCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViInt32*        value
);
static ViStatus _VI_FUNC ciCsAttrVoltageLimitBehavior_WriteCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViInt32         value
);

static ViStatus _VI_FUNC ciCsAttrVoltageLimit_CheckCallback(
    ViSession       vi,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViReal64        value
);

static ViStatus _VI_FUNC ciCsAttrCurrentLevel_CheckCallback(
    ViSession       vi,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViReal64        value
);

static ViStatus _VI_FUNC ciCsAttrFrequencyValue_CheckCallback (ViSession vi,
                                                                ViConstString channelName,
                                                                ViAttr attributeId,
                                                                ViReal64 value);

static ViStatus _VI_FUNC ciCsAttrSlewCurrentRate_CheckCallback(
    ViSession       vi,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViReal64        value
);

static ViStatus _VI_FUNC ciCsAttrSlewFrequencyRate_CheckCallback (ViSession vi,
                                                                   ViConstString channelName,
                                                                   ViAttr attributeId,
                                                                   ViReal64 value);

static ViStatus _VI_FUNC ciCsAttrSlewRate_ReadCallback (ViSession vi,
                                                         ViSession io,
                                                         ViConstString channelName,
                                                         ViAttr attributeId,
                                                         ViReal64 *value);

static ViStatus _VI_FUNC ciCsAttrSlewRate_WriteCallback (ViSession vi,
                                                          ViSession io,
                                                          ViConstString channelName,
                                                          ViAttr attributeId,
                                                          ViReal64 value);

static ViStatus _VI_FUNC  ciCsAttrFunction_CheckCallback(
    ViSession       vi,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViConstString   value
);

static ViStatus _VI_FUNC ciCsAttrFunction_ReadCallback (
    ViSession           vi,
    ViSession           io, 
    ViConstString       channelName,
    ViAttr              attributeId, 
    const ViConstString cacheValue
);

static ViStatus _VI_FUNC  ciCsAttrTriggerDelay_CheckCallback(
    ViSession       vi,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViReal64        value
);

static ViStatus _VI_FUNC  ciCsAttrAcquisitionStartTime_CheckCallback(
    ViSession       vi,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViReal64        value
);
static ViStatus _VI_FUNC ciCsAttrAcquisitionStartTime_CoerceCallback (ViSession vi,
                                                                       ViConstString channelName,
                                                                       ViAttr attributeId,
                                                                       ViReal64 value,
                                                                       ViReal64 *coercedValue);

static ViStatus _VI_FUNC  ciCsAttrAcquisitionStartTime_ReadCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViReal64*       value
);

static ViStatus _VI_FUNC  ciCsAttrAcquisitionStartTime_WriteCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViReal64        value
);

static ViStatus _VI_FUNC  ciCsAttrInstrCurrentRange_ReadCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   pahseName,
    ViAttr          attributeId,
    ViReal64*       value
);

static ViStatus _VI_FUNC  ciCsAttrInstrCurrentRange_WriteCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   pahseName,
    ViAttr          attributeId,
    ViReal64        value
);

static ViStatus _VI_FUNC ciCsAttrInstrOutputPhaseCount_WriteCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViInt32         value
);

static ViStatus _VI_FUNC ciCsAttrOutputPhaseCount_ReadCallback (ViSession vi,
                                                                 ViSession io,
                                                                 ViConstString channelName,
                                                                 ViAttr attributeId,
                                                                 ViInt32 *value);

/*****************************************************************************
 *------------ User-Callable Functions (Exportable Functions) ---------------*
 *****************************************************************************/

/*****************************************************************************
 * Function: ciCs_init   
 * Purpose:  VXIplug&play required function.  Calls the   
 *           ciCs_InitWithOptions function.   
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_init (
    ViRsrc      resourceName,
    ViBoolean   IDQuery,
    ViBoolean   resetDevice,
    ViSession*  newVi,
    ViUInt32 	baudRate
)
{   
    ViStatus    error = VI_SUCCESS;

    if (newVi == VI_NULL)
        {
        Ivi_SetErrorInfo (VI_NULL, VI_FALSE, IVI_ERROR_INVALID_PARAMETER, 
                          VI_ERROR_PARAMETER4, "Null address for Instrument Handle");
        checkErr( IVI_ERROR_INVALID_PARAMETER);
        }

    checkErr( ciCs_InitWithOptions (resourceName, IDQuery, resetDevice, 
                                        "", newVi, baudRate));

Error:  
    return error;
}

/*****************************************************************************
 * Function: ciCs_InitWithOptions                                       
 * Purpose:  This function creates a new IVI session and calls the 
 *           IviInit function.                                     
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_InitWithOptions (
    ViRsrc          resourceName,
    ViBoolean       IDQuery,
    ViBoolean       resetDevice,
    ViConstString   optionString,
    ViSession*      newVi,
    ViUInt32 		baudRate
)
{   
    ViStatus    error = VI_SUCCESS;
    ViSession   vi = VI_NULL;
    ViChar      newResourceName[IVI_MAX_MESSAGE_BUF_SIZE];
    ViChar      newOptionString[IVI_MAX_MESSAGE_BUF_SIZE];
    ViBoolean   isLogicalName;
    
    if (newVi == VI_NULL)
        {
        Ivi_SetErrorInfo (VI_NULL, VI_FALSE, IVI_ERROR_INVALID_PARAMETER, 
                          VI_ERROR_PARAMETER5, "Null address for Instrument Handle");
        checkErr( IVI_ERROR_INVALID_PARAMETER);
        }

    *newVi = VI_NULL;
    
    checkErr( Ivi_GetInfoFromResourceName (resourceName, (ViString)optionString, newResourceName,
                                           newOptionString, &isLogicalName));
    
        /* create a new interchangeable driver */
    checkErr( Ivi_SpecificDriverNew ("ciCs", newOptionString, &vi));  
    if (!isLogicalName)
    {
        ViInt32 oldFlag = 0;
        
        checkErr (Ivi_GetAttributeFlags (vi, IVI_ATTR_IO_RESOURCE_DESCRIPTOR, &oldFlag));
        checkErr (Ivi_SetAttributeFlags (vi, IVI_ATTR_IO_RESOURCE_DESCRIPTOR, oldFlag & 0xfffb | 0x0010));
        checkErr (Ivi_SetAttributeViString (vi, "", IVI_ATTR_IO_RESOURCE_DESCRIPTOR, 0, newResourceName));
    }
        /* init the driver */
    checkErr( ciCs_IviInit (newResourceName, IDQuery, resetDevice, vi, baudRate)); 
    if (isLogicalName)
        checkErr( Ivi_ApplyDefaultSetup (vi));
    *newVi = vi;
    
Error:
    if (error < VI_SUCCESS) 
        Ivi_Dispose (vi);
        
    return error;
}

/*****************************************************************************
 * Function: ciCs_IviInit                                                       
 * Purpose:  This function is called by ciCs_InitWithOptions  
 *           or by an IVI class driver.  This function initializes the I/O 
 *           interface, optionally resets the device, optionally performs an
 *           ID query, and sends a default setup to the instrument.                
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_IviInit (
    ViRsrc      resourceName,
    ViBoolean   IDQuery,
    ViBoolean   reset,
    ViSession   vi,
    ViUInt32 	baudRate
)
{
    ViStatus    error = VI_SUCCESS;
    ViSession   io = VI_NULL;
    ViInt32     model_type;
    ViChar      id_query_buffer[BUFFER_SIZE] = "";
    
    checkErr( ciCs_GetDriverSetupOption (vi, &model_type));
	 
    if (!Ivi_Simulating(vi))
    {
        ViSession   rmSession = VI_NULL;
        ViUInt16    interface_type;// = VI_INTF_GPIB;

        /* Open instrument session */
        checkErr( Ivi_GetAttributeViSession (vi, VI_NULL, IVI_ATTR_VISA_RM_SESSION, 0,
                                             &rmSession)); 
        viCheckErr( viOpen (rmSession, resourceName, VI_NULL, VI_NULL, &io));
        
        if (IDQuery){
        	ViInt32 actual_model = model_type;

        	if (model_type != actual_model)
            	viCheckErr( VI_ERROR_FAIL_ID_QUERY);
    	}
    
        /* io session owned by driver now */
        viCheckErr( viGetAttribute (io, VI_ATTR_INTF_TYPE, &interface_type));
        if (interface_type == VI_INTF_ASRL)  // Serial Port /
        {
            busInterface = VI_INTF_ASRL;
            viCheckErr( viSetAttribute (io, VI_ATTR_TERMCHAR, '\n'));
            viCheckErr( viSetAttribute (io, VI_ATTR_TERMCHAR_EN, VI_TRUE));
            viCheckErr( viSetAttribute (io, VI_ATTR_ASRL_END_IN, VI_ASRL_END_TERMCHAR));
            viCheckErr( viSetAttribute (io, VI_ATTR_ASRL_END_OUT, VI_ASRL_END_TERMCHAR));
            viCheckErr( viSetAttribute (io, VI_ATTR_TMO_VALUE, 2000));//v1009
        
			viCheckErr( viSetBuf (io, VI_ASRL_OUT_BUF, 60000) );
			
        	// Add for RS-232 
			#ifdef VI_ATTR_ASRL_DATA_BITS
	    	// Configure serial COM, baud rate, data bit 8, parity 0, stop 1	
		    
			viCheckErr( viSetAttribute (io, VI_ATTR_ASRL_BAUD, baudRate )); 
	       	viCheckErr( viSetAttribute (io, VI_ATTR_ASRL_DATA_BITS, 8 )); 
	       	viCheckErr( viSetAttribute (io, VI_ATTR_ASRL_PARITY, VI_ASRL_PAR_NONE )); 
	       	viCheckErr( viSetAttribute (io, VI_ATTR_ASRL_STOP_BITS, VI_ASRL_STOP_ONE )); 
	       	viCheckErr( viSetAttribute (io, VI_ATTR_ASRL_FLOW_CNTRL, VI_ASRL_FLOW_RTS_CTS ));
		   																	
	        #endif
        }
		else if (interface_type == VI_INTF_TCPIP)  // LAN /
        {
            busInterface = VI_INTF_TCPIP;
            
			viCheckErr( viSetAttribute (io, VI_ATTR_TERMCHAR, '\n'));
            viCheckErr( viSetAttribute (io, VI_ATTR_TERMCHAR_EN, VI_TRUE));
			
			viCheckErr( viSetAttribute (io, VI_ATTR_TCPIP_KEEPALIVE, VI_TRUE ));
			
            viCheckErr( viSetAttribute (io, VI_ATTR_TMO_VALUE, 2000));//v1009
        }	  
        else
        {
		   	busInterface = VI_INTF_GPIB;
		   	viCheckErr( viSetAttribute (io, VI_ATTR_TMO_VALUE, 2000 ));//v1009
		}
            
        // io session owned by driver now 
        checkErr( Ivi_SetAttributeViSession (vi, VI_NULL, IVI_ATTR_IO_SESSION, 0, io));  

        // Configure VISA Formatted I/O 
        viCheckErr( viSetBuf (io, VI_READ_BUF | VI_WRITE_BUF, 4000));
        viCheckErr( viSetAttribute (io, VI_ATTR_WR_BUF_OPER_MODE, VI_FLUSH_ON_ACCESS));
        viCheckErr( viSetAttribute (io, VI_ATTR_RD_BUF_OPER_MODE, VI_FLUSH_ON_ACCESS));

        if (IDQuery && model_type == NOT_SUPPORTED)
        {
            checkErr( ciCs_QueryID (vi, io, BUFFER_SIZE, id_query_buffer));
            checkErr( ciCs_GetModelFromBuffer (vi,
                                               id_query_buffer,
                                               &model_type));
        }
    }

    if (model_type == NOT_SUPPORTED)
        model_type = CI4500Cs;   

	checkErr( Ivi_BuildRepCapTable( vi,
                                    PHASE_REP_CAP_NAME,
                                        PHASE_LIST_3_PHASES  //always for three phases
                                  )); 
	
    checkErr( Ivi_BuildChannelTable (vi, CHANNEL_LIST, VI_FALSE, VI_NULL));    
     
    // Add attributes 
    checkErr( ciCs_InitAttributes (vi, model_type));
        
    // Reset instrument ----------------------------------------------------
    if (reset) 
        error = ciCs_reset (vi);
    else  // Send Default Instrument Setup ---------------------------------
        error = ciCs_DefaultInstrSetup (vi);
    checkErr( error );
    
    // Identification Query ------------------------------------------------
    if (IDQuery)                               
    {
        ViInt32 actual_model = model_type;

        if (!id_query_buffer[0])
        {
            checkErr( Ivi_GetAttributeViString (vi,
                                                VI_NULL,
                                                CICS_ATTR_ID_QUERY_RESPONSE,
                                                0, BUFFER_SIZE,
                                                id_query_buffer));
            
            checkErr( ciCs_GetModelFromBuffer (vi,
                                               id_query_buffer,
                                               &actual_model));
        }
        else
        {
            checkErr( Ivi_SetAttributeViString (vi,
                                                VI_NULL,
                                                CICS_ATTR_ID_QUERY_RESPONSE,
                                                IVI_VAL_DONT_MARK_AS_SET_BY_USER,
                                                id_query_buffer));
        }

        if (model_type != actual_model)
            viCheckErr( VI_ERROR_FAIL_ID_QUERY);
    }  

    checkErr( ciCs_CheckStatus (vi));

Error:
    if (error < VI_SUCCESS)
        {
        if (!Ivi_Simulating (vi) && io)
            viClose (io);
        }
    return error;
}

/*****************************************************************************
 * Function: ciCs_close                                                           
 * Purpose:  This function closes the instrument.                            
 *
 *           Note:  This function must unlock the session before calling
 *           Ivi_Dispose.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_close (ViSession vi)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Ivi_LockSession (vi, VI_NULL));
    
    checkErr( ciCs_IviClose (vi));

Error:    
    Ivi_UnlockSession (vi, VI_NULL);
    Ivi_Dispose (vi);  

    return error;
}

/*****************************************************************************
 * Function: ciCs_IviClose                                                        
 * Purpose:  This function performs all of the drivers clean-up operations   
 *           except for closing the IVI session.  This function is called by 
 *           ciCs_close or by an IVI class driver. 
 *
 *           Note:  This function must close the I/O session and set 
 *           IVI_ATTR_IO_SESSION to 0.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_IviClose (ViSession vi)
{
    ViStatus error = VI_SUCCESS;
    ViSession io = VI_NULL;

        /* Do not lock here.  The caller manages the lock. */
    checkErr( Ivi_GetAttributeViSession (vi, VI_NULL, IVI_ATTR_IO_SESSION, 0, &io));

Error:
    Ivi_SetAttributeViSession (vi, VI_NULL, IVI_ATTR_IO_SESSION, 0, VI_NULL);
    if(io)                                                      
        {
        viClose (io);
        }
    return error;   
}

/*****************************************************************************
 * Function: ciCs_reset                                                         
 * Purpose:  This function resets the instrument.                          
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_reset (ViSession vi)
{
    ViStatus    error = VI_SUCCESS;

    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (!Ivi_Simulating(vi))                /* call only when locked */
    {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */
        
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));
        if( busInterface == VI_INTF_TCPIP )
			viCheckErr( viPrintf (io, "*RST\n"));
		else if (busInterface == VI_INTF_ASRL)
			viPrintf (io, "*RST");
		else
        	viCheckErr( viPrintf (io, "*RST"));
        
        Delay( 1 );
        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
    }
    
    checkErr( ciCs_DefaultInstrSetup (vi));                                
    
    checkErr( ciCs_CheckStatus (vi));                                      

Error:
    Ivi_UnlockSession (vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ResetWithDefaults
 * Purpose:  This function resets the instrument and applies default settings
 *           from the IVI Configuration Store based on the logical name
 *           from which the session was created.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ResetWithDefaults (ViSession vi)
{
    ViStatus error = VI_SUCCESS;
    
    checkErr( Ivi_LockSession (vi, VI_NULL));
    
    checkErr( ciCs_reset(vi));
    checkErr( Ivi_ApplyDefaultSetup(vi));
    
Error:
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/**************************************************************************** 
 *  Function: ciCs_Disable
 *  Purpose:  This function places the instrument in a quiescent state as 
 *            quickly as possible.
 ****************************************************************************/
ViStatus _VI_FUNC ciCs_Disable (ViSession vi)
{
    return VI_SUCCESS;
}

/*****************************************************************************
 * Function: ciCs_self_test                                                       
 * Purpose:  This function executes the instrument self-test and returns the 
 *           result.                                                         
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_self_test (
    ViSession   vi,
    ViInt16*    testResult, 
    ViChar      testMessage[]
)
{
    ViStatus    error = VI_SUCCESS;
	ViChar	    tempCmd[BUFFER_SIZE];  
	ViChar      buffer[BUFFER_SIZE];
	ViUInt32  	read_count;
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (testResult == VI_NULL)
        viCheckParm( IVI_ERROR_INVALID_PARAMETER, 2, "Null address for Test Result");

    if (!Ivi_Simulating(vi))                /* call only when locked */
    {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */
    
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));
        
        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    }
        
        if( busInterface == VI_INTF_TCPIP )
        	viCheckErr( viPrintf (io, "*TST?\n"));
		else if(busInterface == VI_INTF_ASRL){		
			strcpy(tempCmd, "*TST?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		}
		else
			viCheckErr( viPrintf (io, "*TST?"));
        
        Delay(10);
        
		if(busInterface == VI_INTF_ASRL){
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
			buffer[read_count] = '\0';
		
			*testResult = atoi(buffer);
		}
		else
        	viCheckErr( viScanf (io, "%hd", testResult));
        
        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    }
        
        if( busInterface == VI_INTF_TCPIP )
			viCheckErr( viQueryf(io, ":SYST:ERR?\n", "%ld,\"%256[^\"]", testResult, testMessage));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, ":SYST:ERR?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, "");
			
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
			buffer[read_count] = '\0';
		
			strcpy(testMessage, buffer);
		}
		else
			viCheckErr( viQueryf(io, ":SYST:ERR?", "%ld,\"%256[^\"]", testResult, testMessage));
    }
    else
    {   /* Simulate Self Test */
        *testResult = 0;
        strcpy(testMessage, "No error");
    }
    
    checkErr( ciCs_CheckStatus (vi));

Error:
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_error_query                                                     
 * Purpose:  This function queries the instrument error queue and returns   
 *           the result.                                                     
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_error_query (
    ViSession       vi,
    ViInt32*        errCode,
    ViChar          errMessage[]
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar	    tempCmd[BUFFER_SIZE];  
	ViChar      buffer[BUFFER_SIZE];
	ViUInt32  	read_count;
	
    checkErr( Ivi_LockSession (vi, VI_NULL));
    
    if (errCode == VI_NULL)
        viCheckParm( IVI_ERROR_INVALID_PARAMETER, 2, "Null address for Error Code");
    if (errMessage == VI_NULL)
        viCheckParm( IVI_ERROR_INVALID_PARAMETER, 3, "Null address for Error Message");

    if (!Ivi_Simulating(vi))                /* call only when locked */
        {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */
    
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));
        
        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    }  
        
        if( busInterface == VI_INTF_TCPIP )
        	viCheckErr( viQueryf(io, ":SYST:ERR?\n", "%ld,\"%256[^\"]", errCode, errMessage));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, ":SYST:ERR?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, "");
			
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
			buffer[read_count] = '\0';
		
			strcpy(errMessage, buffer);
		}
		else
			viCheckErr( viQueryf(io, ":SYST:ERR?", "%ld,\"%256[^\"]", errCode, errMessage));
        }
    else
        {
            /* Simulate Error Query */
        *errCode = 0;
        strcpy (errMessage, "No error");
        }

Error:
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_error_message                                                  
 * Purpose:  This function translates the error codes returned by this       
 *           instrument driver into user-readable strings.  
 *
 *           Note:  The caller can pass VI_NULL for the vi parameter.  This 
 *           is useful if one of the init functions fail.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_error_message (
    ViSession   vi,
    ViStatus    errorCode,
    ViChar      errorMessage[256]
)
{
    ViStatus    error = VI_SUCCESS;
    
    static      IviStringValueTable errorTable = 
        {
            //{CICS_ERROR_CALIBRATION_NOT_ENABLED,        "Calibration is not enabled"},
            {VI_NULL,                               VI_NULL}
        };  
        
    if (vi)
        Ivi_LockSession(vi, VI_NULL);

        /* all VISA and IVI error codes are handled as well as codes in the table */
    if (errorMessage == VI_NULL)
        viCheckParm( IVI_ERROR_INVALID_PARAMETER, 3, "Null address for Error Message");

    checkErr( Ivi_GetSpecificDriverStatusDesc(vi, errorCode, errorMessage, errorTable));

Error:
    if (vi)
        Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_InvalidateAllAttributes
 * Purpose:  This function invalidates the cached value of all attributes.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_InvalidateAllAttributes (ViSession vi)
{
    return Ivi_InvalidateAllAttributes(vi);
}

/*****************************************************************************
 * Function: ciCs_revision_query                                                  
 * Purpose:  This function returns the driver and instrument revisions.      
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_revision_query (
    ViSession   vi,
    ViChar      driverRev[],
    ViChar      instrRev[]
)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (driverRev == VI_NULL)
        viCheckParm( IVI_ERROR_INVALID_PARAMETER, 2, "Null address for Driver Revision");
    if (instrRev == VI_NULL)
        viCheckParm( IVI_ERROR_INVALID_PARAMETER, 3, "Null address for Instrument Revision");

    checkErr( Ivi_GetAttributeViString (vi, VI_NULL, CICS_ATTR_SPECIFIC_DRIVER_REVISION, 
                                        0, 256, driverRev));
    checkErr( Ivi_GetAttributeViString (vi, "", CICS_ATTR_INSTRUMENT_FIRMWARE_REVISION, 
                                        0, 256, instrRev));
    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureOutput
 * Purpose:  This functions configures the number of outputs phases
 *           for the AC source.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureOutput (
    ViSession   vi,
    ViInt32     numberOfPhases
)
{
    return Ivi_SetAttributeViInt32( vi,
                                    VI_NULL,
                                    CICS_ATTR_OUTPUT_PHASE_COUNT,
                                    IVI_VAL_DIRECT_USER_CALL,
                                    numberOfPhases);
}

/*****************************************************************************
 * Function: ciCs_ConfigureOutputEnabled
 * Purpose:  This function enables or disables the AC source output.
  *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureOutputEnabled (
    ViSession   vi,
    ViBoolean   enabled
)
{
    return Ivi_SetAttributeViBoolean( vi,
                                      VI_NULL,
                                      CICS_ATTR_OUTPUT_ENABLED,
                                      IVI_VAL_DIRECT_USER_CALL,
                                      enabled);
}

/*****************************************************************************
 * Function: ciCs_ConfigureOutputALCState
 * Purpose:  This function enables or disables the ALC state of the AC source.
  *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureOutputALCState (
    ViSession   vi,
    ViInt32   	enabled
)
{
    return Ivi_SetAttributeViInt32( vi,
                                    VI_NULL,
                                    CICS_ATTR_OUTPUT_ALC_STATE,
                                    IVI_VAL_DIRECT_USER_CALL,
                                    enabled);
}

/*****************************************************************************
 * Function: ciCs_ConfigureOutputPhaseMode
 * Purpose:  This functions configures the AC source to be single or three
 *           phase mode.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureOutputPhaseMode (
    ViSession   vi,
    ViInt32     phaseMode
)
{
    return Ivi_SetAttributeViInt32( vi,
                                    VI_NULL,
                                    CICS_ATTR_OUTPUT_PHASE_MODE,
                                    IVI_VAL_DIRECT_USER_CALL,
                                    phaseMode);
}

/*****************************************************************************
 * Function: ciCs_ConfigureOutputRange
 * Purpose:  This functions configures the AC source's output range.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureOutputRange (
    ViSession       vi,
    ViReal64        range
)
{
    ViStatus    error = VI_SUCCESS;

    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (Ivi_RangeChecking(vi))
    {
        ViReal64    max, min;
        
        checkErr( Ivi_GetAttributeViReal64(vi, "", CICS_ATTR_VOLTAGE_RANGE_MINIMUM, 0, &min));
        
        checkErr( ciCs_GetMaximumRange ( vi,
                                         CICS_VAL_RANGE_VOLTAGE,
                                         &max));

		if (!Ivi_Simulating(vi)){                /* call only when locked */
	        if (range == min){
	        }
	        else if(range == max){
	        }
	        else
	            viCheckParm(IVI_ERROR_INVALID_VALUE, 3, "range");
	  	}
    }

    checkErr( Ivi_SetAttributeViReal64( vi,
                                        "PHASE1",
                                        CICS_ATTR_INSTR_VOLTAGE_RANGE,
                                        0, range));
    
    Delay( 0.1 ); 
    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureVoltageLimit
 * Purpose:  This function configures the voltage limit. 
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureVoltageLimit (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         behavior,
    ViReal64        limit
)
{
    ViStatus    error = VI_SUCCESS;
    ViBoolean   behaviorFirst = VI_FALSE;
    
    checkErr( Ivi_LockSession (vi, VI_NULL));

    behaviorFirst = (behavior != CICS_VAL_VOLTAGE_TRIP);
    
    if (behaviorFirst)
    {
        viCheckParm( Ivi_SetAttributeViInt32 (vi, phaseName, 
                                              CICS_ATTR_VOLTAGE_LIMIT_BEHAVIOR,
                                              0, behavior), 3,
                                              "Voltage Limit Behavior");
    }
    viCheckParm( Ivi_SetAttributeViReal64 (vi, phaseName, 
                                               CICS_ATTR_VOLTAGE_LIMIT,
                                               0, limit), 4,
                                               "Voltage Limit");
    if (!behaviorFirst)        
    {
        viCheckParm( Ivi_SetAttributeViInt32 (vi, phaseName, 
                                              CICS_ATTR_VOLTAGE_LIMIT_BEHAVIOR,
                                              0, behavior), 3,
                                              "Voltage Limit Behavior");
    }
    
    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureOVP
 * Purpose:  This function configures the AC source's over-voltage
 *           protection. 
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureOVP (
    ViSession       vi,
    ViConstString   phaseName,
    ViBoolean       enabled,
    ViReal64        limit
)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Ivi_LockSession (vi, VI_NULL));

    /*
     * If you are enabling over-voltage protection, set the limit first
     * to avoid a premature over-voltage condition.
     */

    if (enabled)
    {
        viCheckParm( Ivi_SetAttributeViReal64 (vi,
                                               phaseName, 
                                               CICS_ATTR_OVP_LIMIT,
                                               0, limit),
                                               4, "OVP Limit");

    }

    viCheckParm( Ivi_SetAttributeViBoolean (vi,
                                            phaseName, 
                                            CICS_ATTR_OVP_ENABLED,
                                            0, enabled),
                                            3, "OVP Enabled");
    

    if (!enabled)
    {
        viCheckParm( Ivi_SetAttributeViReal64 (vi,
                                               phaseName, 
                                               CICS_ATTR_OVP_LIMIT,
                                               0, limit),
                                               4, "OVP Limit");

    }

    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureCurrentLevel
 * Purpose:  This function configures the AC RMS current level that the power
 *           supply attempts to generate.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureCurrentLevel (
    ViSession       vi,
    ViConstString   phaseName,
    ViReal64        level
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     phaseName,
                                     CICS_ATTR_CURRENT_LEVEL,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     level);
}

/*****************************************************************************
 * Function: ciCs_ConfigureFrequencyValue
 * Purpose:  This function configures the frequency value that the AC source
 			 attempts to generate.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureFrequencyValue (
    ViSession       vi,
    ViReal64        value
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     VI_NULL,
                                     CICS_ATTR_FREQUENCY,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     value);
}

/*****************************************************************************
 * Function: ciCs_ConfigurePhaseAngle
 * Purpose:  This function configures the phase of the output voltage waveform
 *           relative to an internal reference.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigurePhaseAngle (
    ViSession       vi,
    ViConstString   phaseName,
    ViReal64        angle
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     phaseName,
                                     CICS_ATTR_PHASE_ANGLE,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     angle);
}

/*****************************************************************************
 * Function: ciCs_ConfigureWaveform
 * Purpose:  This function configures the phase of the output voltage waveform
 *           relative to an internal reference.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureWaveform (
    ViSession       vi,
    ViConstString   function,
    ViReal64        frequency
)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Ivi_LockSession (vi, VI_NULL));

    viCheckParm( Ivi_SetAttributeViString( vi,
                                           VI_NULL,
                                           CICS_ATTR_FUNCTION,
                                           0,
                                           function),
                                           2, "Function");
    viCheckParm( Ivi_SetAttributeViReal64( vi,
                                           VI_NULL,
                                           CICS_ATTR_FREQUENCY,
                                           0,
                                           frequency),
                                           3, "Frequency");


    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureClippingLevel
 * Purpose:  This function configures the clipping level when a cliped
 *           sine output waveform is selected.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureClippingLevel (
    ViSession   vi,
    ViReal64    clippingLevel
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     VI_NULL,
                                     CICS_ATTR_CLIPPING_LEVEL,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     clippingLevel);
}

/*****************************************************************************
 * Function: ciCs_ConfigureSlewCurrentLevel
 * Purpose:  This function configures the slew rate for all programmed
 *           changes in the ac rms output voltage level of the AC source.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureSlewCurrentLevel (
    ViSession       vi,
    ViConstString   phaseName,
    ViReal64        slewVoltageRate
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     phaseName,
                                     CICS_ATTR_SLEW_CURRENT_RATE,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     slewVoltageRate);
}

/*****************************************************************************
 * Function: ciCs_ConfigureSlewFrequency
 * Purpose:  This function configures the rate at which frequency changes
 *           for all programmed changes in output frequency
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureSlewFrequency (
    ViSession   vi,
    ViReal64    slewFrequencyRate
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     VI_NULL,
                                     CICS_ATTR_SLEW_FREQUENCY_RATE,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     slewFrequencyRate);
}


/*****************************************************************************
 * Function: ciCs_GetPhaseName
 * Purpose:  This function returns the highest-level phase name
 *           that corresponds to the specific driver phase string that is
 *           in the phase table at an index you specify.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_GetPhaseName (
    ViSession   vi,
    ViInt32     index,
    ViInt32     bufferSize,
    ViChar _VI_FAR phaseName[]
)
{
    ViStatus        error = VI_SUCCESS;
    ViConstString   phase_name;
    ViInt32         length;

    checkErr( Ivi_LockSession (vi, VI_NULL));
    
    checkErr( Ivi_GetNthRepCapName (vi,
                                    PHASE_REP_CAP_NAME,
                                    index,
                                    &phase_name));

    if (bufferSize > 0 && phaseName != VI_NULL)
    {
        strncpy(phaseName, phase_name, bufferSize);
    }
    
    length = (phase_name) ? (ViInt32)strlen(phase_name) : 0;
    
    if ((length + 1) > bufferSize)
        error = length+1;

Error:
    Ivi_UnlockSession (vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_WriteArbWaveform
 * Purpose:  This function writes the aribtrary waveform to AC source.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_WriteArbWaveform (
    ViSession       vi,
    ViConstString   name,
    ViInt32         waveformSize,
    ViReal64*       waveformDataArray
)
{

    ViStatus    error = VI_SUCCESS;
    ViChar*     pdata = VI_NULL;
    ViChar*     pend = VI_NULL;
    ViInt32     name_size,
                index;
    ViBoolean   is_defined = VI_FALSE;
    ViBoolean   update_timeout = VI_FALSE;
    ViUInt32    old_time_out;
    ViSession   io = VI_NULL;

	ViChar	    tempCmd[BUFFER_SIZE];  
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (waveformSize < 0)
        viCheckParm(IVI_ERROR_INVALID_VALUE, 3, "Waveform Size");
    if (name)
    {
        name_size = (ViInt32)strlen(name);
    }
    else
        viCheckParm(IVI_ERROR_INVALID_VALUE, 2, "Name");

    checkErr( ciCs_IsUserWaveformDefined( vi, name, &is_defined));

    if (Ivi_RangeChecking(vi))
    {
        ViInt32     number_of_user_waveforms;

        if (waveformSize != 1024)
            viCheckParm(IVI_ERROR_INVALID_VALUE, 3, "Waveform Size");
        
        checkErr( ciCs_GetNumberUserDefinedWaveforms(vi, name, &number_of_user_waveforms));

        if (!is_defined)
            number_of_user_waveforms++;

        if (number_of_user_waveforms > MAX_USER_DEFINED_WAVEFORMS)
            viCheckErrElab( IVI_ERROR_INVALID_VALUE, "Only 12 user waveforms can be defined.");
    }

    /*
     * Allocate buffer:
     *  size of allocated bufer
     *  <Header> + name_size + 1 comma + waveformSize*(<size of formated double> + 1 comma)
     *     5
     */
    checkErr( Ivi_Alloc( vi, 5 + name_size + 1 + (DOUBLE_FORMAT_LENGTH + 1) * waveformSize, (ViAddr*)&pdata));
    pend = pdata;
    pend += sprintf(pend, "TRAC:DATA %s,", name);

    for (index = 0; index < waveformSize; index++) 
        pend += sprintf(pend, "%Lf,", waveformDataArray[index]);
        
    *--pend = '\0';	//Remove last comma

	if( busInterface == VI_INTF_TCPIP ) 
		pend += sprintf(pend, "\n");//added a LF for LAN
	
    if (!Ivi_Simulating(vi))                /* call only when locked */
    {
        io = Ivi_IOSession(vi); /* call only when locked */

        viCheckErr( viGetAttribute (io, VI_ATTR_TMO_VALUE, &old_time_out));
        update_timeout = VI_TRUE;
        viCheckErr( viSetAttribute (io, VI_ATTR_TMO_VALUE, 15000));

        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        /*
         * Invalidate attribute CICS_ATTR_INSTR_USER_TABLE_LIST
         */
        checkErr( Ivi_InvalidateAttribute(vi, VI_NULL, CICS_ATTR_INSTR_USER_TABLE_LIST));

        if (!is_defined)
            if( busInterface == VI_INTF_TCPIP )
				viCheckErr( viPrintf(io, "TRAC:DEF %s\n", name));
			else if(busInterface == VI_INTF_ASRL){
				strcpy(tempCmd, "TRAC:DEF");
				strcat(tempCmd, " ");
				strcat(tempCmd, name);
				viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
				strcpy(tempCmd, ""); 
			}
			else
				viCheckErr( viPrintf(io, "TRAC:DEF %s", name));

		if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
        
        viCheckErr( viWrite(io, pdata, ((ViUInt32)pend) - ((ViUInt32)pdata), VI_NULL));
    	Delay(0.5);
    }
    else
    {
        ViChar  table_list[BUFFER_SIZE];
        ViChar* pend;

        checkErr( Ivi_GetAttributeViString( vi,
                                            VI_NULL,
                                            CICS_ATTR_INSTR_USER_TABLE_LIST,
                                            0,
                                            IVI_VAL_SET_CACHE_ONLY,
                                            table_list));
        pend = table_list;
        pend += strlen(table_list);
        sprintf(pend, "\"%s\"", name);
        
        checkErr( Ivi_SetAttributeViString( vi,
                                            VI_NULL,
                                            CICS_ATTR_INSTR_USER_TABLE_LIST,
                                            IVI_VAL_SET_CACHE_ONLY,
                                            table_list));
    }

    checkErr( ciCs_CheckStatus (vi));

Error:
    if (update_timeout && io)
        viSetAttribute (io, VI_ATTR_TMO_VALUE, old_time_out);
    if (pdata) Ivi_Free(vi, pdata);
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_QueryArbWaveformCapabilities
 * Purpose:  This function returns the attributes of the function generator
 *           that are related to creating arbitrary sequences.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_QueryArbWaveformCapabilities (
    ViSession   vi,
    ViInt32*    maxNbOfWaveforms,
    ViInt32*    waveformQuantum,
    ViInt32*    minWaveformSize,
    ViInt32*    maxWaveformSize
)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Ivi_LockSession (vi, VI_NULL));

    checkErr( Ivi_GetAttributeViInt32( vi,
                                       VI_NULL,
                                       CICS_ATTR_MAX_NUM_WAVEFORMS,
                                       0, maxNbOfWaveforms));

    checkErr( Ivi_GetAttributeViInt32( vi,
                                       VI_NULL,
                                       CICS_ATTR_WAVEFORM_QUANTUM,
                                       0, waveformQuantum));

    checkErr( Ivi_GetAttributeViInt32( vi,
                                       VI_NULL,
                                       CICS_ATTR_MIN_WAVEFORM_SIZE,
                                       0, minWaveformSize));

    checkErr( Ivi_GetAttributeViInt32( vi,
                                       VI_NULL,
                                       CICS_ATTR_MAX_WAVEFORM_SIZE,
                                       0, maxWaveformSize));

    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ClearAllArbWaveforms
 * Purpose:  This function deletes all user-defined waveforms.
 *           This function invalidates attribute CICS_ATTR_FUNCTION, when any
 *           user defined wave is set for this attribute.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ClearAllArbWaveforms (
    ViSession   vi
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar      active_function[BUFFER_SIZE];
    ViChar      waveform_names[BUFFER_SIZE];
    ViBoolean   is_user_function = VI_FALSE;
    ViChar	    tempCmd[BUFFER_SIZE];   
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    checkErr( Ivi_GetAttributeViString( vi,
                                        VI_NULL,
                                        CICS_ATTR_INSTR_USER_TABLE_LIST,
                                        0,
                                        BUFFER_SIZE, waveform_names));
    checkErr( Ivi_GetAttributeViString( vi,
                                        VI_NULL,
                                        CICS_ATTR_FUNCTION,
                                        0,
                                        BUFFER_SIZE,
                                        active_function));
    
    error = ciCs_IsUserWaveformDefined( vi,
                                        active_function,
                                        &is_user_function);

    if (error == IVI_ERROR_INVALID_VALUE)
#if !defined(MULTI_THREAD_USED)
        error = Ivi_ClearErrorInfo (vi);
#else
    {
        ViStatus    primaryError,
                    secondaryError;
        ViChar      errorElaboration[256];
        error = Ivi_GetErrorInfo(vi, &primaryError, &secondaryError, errorElaboration);
    }
#endif
    
    checkErr( error);

    if (is_user_function)
    {
        checkErr( Ivi_SetAttributeViString( vi,
                                            VI_NULL,
                                            CICS_ATTR_FUNCTION,
                                            IVI_VAL_DONT_MARK_AS_SET_BY_USER,
                                            CICS_VAL_SINE));
    }

    if (!Ivi_Simulating(vi))                /* call only when locked */
    {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */
        ViChar*     pwave = VI_NULL;

        /*
         * Invalidate attribute CICS_ATTR_INSTR_USER_TABLE_LIST
         * Invalidate it, before write any command.
         * Commad is send in the loop and one can failure. In this case invalidation
         * should be after Error label.
         */
        checkErr( Ivi_InvalidateAttribute(vi, VI_NULL, CICS_ATTR_INSTR_USER_TABLE_LIST));
        
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        pwave = strtok(waveform_names, ",");
        while (pwave != VI_NULL)
        {
            RemoveQuotationMarks(pwave);

            error = ciCs_GetStr2FromStr1 (gs_FunctionStringTable, pwave, VI_NULL);
            if (error != VI_SUCCESS)
				if( busInterface == VI_INTF_TCPIP )  
					viCheckErr( viPrintf(io, ":TRAC:DEL %s\n", pwave));
				else if(busInterface == VI_INTF_ASRL){
					strcpy(tempCmd, "TRAC:DEL");
					strcat(tempCmd, " ");
					strcat(tempCmd, pwave);
					viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
					strcpy(tempCmd, ""); 	
				}
				else
					viCheckErr( viPrintf(io, ":TRAC:DEL %s", pwave));
				
			if(busInterface == VI_INTF_ASRL)
    			Delay(serialDelay);
            pwave = strtok(VI_NULL, ",");
        }
    }
    else
    {
        checkErr( Ivi_SetAttributeViString( vi,
                                            VI_NULL,
                                            CICS_ATTR_INSTR_USER_TABLE_LIST,
                                            IVI_VAL_SET_CACHE_ONLY,
                                            "\"SINUSOID\",\"SQUARE\",\"CSINUSOID\""));
    }
    

    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ClearArbWaveforms
 * Purpose:  This function deletes individual user-defined waveforms.
 *           This function invalidates attribute CICS_ATTR_FUNCTION, when any
 *           user defined wave is set for this attribute.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ClearArbWaveforms (
    ViSession   vi,
    ViString waveformName
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar	    tempCmd[BUFFER_SIZE]; 
	
	checkErr( Ivi_LockSession (vi, VI_NULL));

    if (!Ivi_Simulating(vi))                /* call only when locked */
    {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */
        ViChar*     pwave = VI_NULL;

        checkErr( Ivi_InvalidateAttribute(vi, VI_NULL, CICS_ATTR_INSTR_USER_TABLE_LIST));
        
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        pwave = strtok(waveformName, ",");
        
        error = ciCs_GetStr2FromStr1 (gs_FunctionStringTable, pwave, VI_NULL);
        if (error != VI_SUCCESS)
            if( busInterface == VI_INTF_TCPIP )
				viCheckErr( viPrintf(io, ":TRAC:DEL %s\n", pwave));
			else if(busInterface == VI_INTF_ASRL){
				strcpy(tempCmd, "TRAC:DEL");
				strcat(tempCmd, " ");
				strcat(tempCmd, pwave);
				viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
				strcpy(tempCmd, ""); 
			}
			else
				viCheckErr( viPrintf(io, ":TRAC:DEL %s", pwave));

		if(busInterface == VI_INTF_ASRL)
			Delay(serialDelay);
        pwave = strtok(VI_NULL, ",");
    }
    else
    {
        checkErr( Ivi_SetAttributeViString( vi,
                                            VI_NULL,
                                            CICS_ATTR_INSTR_USER_TABLE_LIST,
                                            IVI_VAL_SET_CACHE_ONLY,
                                            "\"SINUSOID\",\"SQUARE\",\"CSINUSOID\""));
    }
    
    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_QueryDefinedWaveforms                                                     
 * Purpose:  This function queries the AC source defined waveforms
 *                                                                
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_QueryDefinedWaveforms (
    ViSession       vi,
    ViChar          definedWaveforms[]
)
{
    ViStatus    error = VI_SUCCESS;
    checkErr( Ivi_LockSession (vi, VI_NULL));
    
    if (definedWaveforms == VI_NULL)
        viCheckParm( IVI_ERROR_INVALID_PARAMETER, 3, "Null address for Error Message");

    if (!Ivi_Simulating(vi))                /* call only when locked */
        {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */
    
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));
        
        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    } 
        
        checkErr( Ivi_GetAttributeViString( vi,
                                        VI_NULL,
                                        CICS_ATTR_INSTR_USER_TABLE_LIST,
                                        0,
                                        BUFFER_SIZE, definedWaveforms));
        }
    else
        {
            // Simulate Error Query
        strcpy (definedWaveforms, "SINUSOID, SQUARE, CSINUSOID");
        }

Error:
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureAcqTriggerSource
 * Purpose:  This function configures the trigger source for a triggered
 *           measurement sequence.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureAcqTriggerSource (
    ViSession   vi,
    ViInt32     source
)
{
    return Ivi_SetAttributeViInt32( vi,
                                    VI_NULL,
                                    CICS_ATTR_ACQUISITION_TRIGGER_SOURCE,
                                    IVI_VAL_DIRECT_USER_CALL,
                                    source);
}

/*****************************************************************************
 * Function: ciCs_ConfigureAcquisitionStartTime
 * Purpose:  This function configures the length of time from the acquire
 *           trigger event to the first point in the record.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureAcquisitionStartTime (
    ViSession   vi,
    ViReal64    startTime
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     VI_NULL,
                                     CICS_ATTR_ACQUISITION_START_TIME,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     startTime);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggerSource
 * Purpose:  This function configures the trigger source for the first
 *           sequence in generating a step, pulse, or list output.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggerSource (
    ViSession   vi,
    ViInt32     source
)
{
    return Ivi_SetAttributeViInt32( vi,
                                    VI_NULL,
                                    CICS_ATTR_TRIGGER_SOURCE,
                                    IVI_VAL_DIRECT_USER_CALL,
                                    source);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggerDelay
 * Purpose:  This function configures the time delay between the detection
 *           of a trigger signal and the start of any corresponding trigger
 *           action.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggerDelay (
    ViSession   vi,
    ViReal64    triggerDelay
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     VI_NULL,
                                     CICS_ATTR_TRIGGER_DELAY,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     triggerDelay);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTTLTriggerOutput
 * Purpose:  This function configures trigger out signal.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTTLTriggerOutput (
    ViSession   vi,
    ViBoolean   state,
    ViInt32     source
)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Ivi_LockSession (vi, VI_NULL));

    viCheckParm( Ivi_SetAttributeViBoolean( vi,
                                         VI_NULL,
                                         CICS_ATTR_OUTPUT_TRIGGER_ENABLED,
                                         0, state),
                                         2, "State");

    viCheckParm( Ivi_SetAttributeViInt32( vi,
                                       VI_NULL,
                                       CICS_ATTR_OUTPUT_TRIGGER_SOURCE,
                                       0, source),
                                       3, "Source");
    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggeredCurrentMode
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggeredCurrentMode (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         transientMode
)
{
    return Ivi_SetAttributeViInt32( vi,
                                    phaseName,
                                    CICS_ATTR_TRIGGERED_CURRENT_MODE,
                                    IVI_VAL_DIRECT_USER_CALL,
                                    transientMode);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggeredFrequencyMode
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggeredFrequencyMode (
    ViSession       vi,
    ViInt32         transientMode
)
{
    return Ivi_SetAttributeViInt32( vi,
                                    VI_NULL,
                                    CICS_ATTR_TRIGGERED_FREQUENCY_MODE,
                                    IVI_VAL_DIRECT_USER_CALL,
                                    transientMode);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggeredFunctionMode
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggeredFunctionMode (
    ViSession       vi,
    ViInt32         transientMode
)
{
    return Ivi_SetAttributeViInt32( vi,
                                    VI_NULL,
                                    CICS_ATTR_TRIGGERED_FUNCTION_MODE,
                                    IVI_VAL_DIRECT_USER_CALL,
                                    transientMode);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggeredAngleMode
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggeredAngleMode (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         transientMode
)
{
    return Ivi_SetAttributeViInt32( vi,
                                    phaseName,
                                    CICS_ATTR_TRIGGERED_PHASE_ANGLE_MODE,
                                    IVI_VAL_DIRECT_USER_CALL,
                                    transientMode);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTrigSlewVoltageMode
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTrigSlewVoltageMode (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         transientMode
)
{
    return Ivi_SetAttributeViInt32( vi,
                                    phaseName,
                                    CICS_ATTR_TRIGGERED_SLEW_CURRENT_RATE_MODE,
                                    IVI_VAL_DIRECT_USER_CALL,
                                    transientMode);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTrigSlewFrequencyMode
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTrigSlewFrequencyMode (
    ViSession       vi,
    ViInt32         transientMode
)
{
    return Ivi_SetAttributeViInt32( vi,
                                    VI_NULL,
                                    CICS_ATTR_TRIGGERED_SLEW_FREQUENCY_RATE_MODE,
                                    IVI_VAL_DIRECT_USER_CALL,
                                    transientMode);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggeredVoltageLevel
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggeredVoltageLevel (
    ViSession       vi,
    ViConstString   phaseName,
    ViReal64        triggeredLevel
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     phaseName,
                                     CICS_ATTR_TRIGGERED_CURRENT_LEVEL,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     triggeredLevel);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggeredFrequency
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggeredFrequency (
    ViSession       vi,
    ViReal64        triggeredFrequency
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     VI_NULL,
                                     CICS_ATTR_TRIGGERED_FREQUENCY,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     triggeredFrequency);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggeredFunction
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggeredFunction (
    ViSession       vi,
    ViConstString   triggeredFunction
)
{
    return Ivi_SetAttributeViString( vi,
                                     VI_NULL,
                                     CICS_ATTR_TRIGGERED_FUNCTION,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     triggeredFunction);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggeredAngle
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggeredAngle (
    ViSession       vi,
    ViConstString   phaseName,
    ViReal64        angle
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     phaseName,
                                     CICS_ATTR_TRIGGERED_PHASE_ANGLE,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     angle);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggeredSlewVoltage
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggeredSlewVoltage (
    ViSession       vi,
    ViConstString   phaseName,
    ViReal64        slewRateVoltage
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     phaseName,
                                     CICS_ATTR_TRIGGERED_SLEW_CURRENT_RATE,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     slewRateVoltage);
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggeredSlewFrequency
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggeredSlewFrequency (
    ViSession       vi,
    ViReal64        slewRateFrequency
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     VI_NULL,
                                     CICS_ATTR_TRIGGERED_SLEW_FREQUENCY_RATE,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     slewRateFrequency);
}

/*****************************************************************************
 * Function: ciCs_QueryTrnsListStatus
 * Purpose:  This function returns the status of the transient list state of
 *           the AC source.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_QueryTrnsListStatus (
    ViSession       vi,
    ViChar          trigStatus[]
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar	    tempCmd[BUFFER_SIZE];
 	ViChar		valueStr[100]; 
	ViChar      buffer[BUFFER_SIZE];
	ViUInt32  	read_count;
	
    checkErr( Ivi_LockSession (vi, VI_NULL));


 	if (!Ivi_Simulating(vi))                /* call only when locked */
        {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */
    
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));
        
        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    } 
        
        if( busInterface == VI_INTF_TCPIP )
			viCheckErr( viQueryf(io, ":TRIG:STAT?\n", "%s", trigStatus));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, ":TRIG:STAT?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
			buffer[read_count] = '\0';
		
			strcpy( trigStatus, buffer);
		}
		else
			viCheckErr( viQueryf(io, ":TRIG:STAT?", "%s", trigStatus));	
        }
    else
        {
            /* Simulate Error Query */
        strcpy (trigStatus, "IDLE");
        }
        
    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigurePulse
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigurePulse (
    ViSession   vi,
    ViInt32     count,
    ViReal64    width,
    ViReal64    period
)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Ivi_LockSession (vi, VI_NULL));

    viCheckParm( Ivi_SetAttributeViInt32( vi,
                                          VI_NULL,
                                          CICS_ATTR_TRIGGER_PULSE_COUNT,
                                          0, count),
                                          2, "Count");

    viCheckParm( Ivi_SetAttributeViReal64( vi,
                                           VI_NULL,
                                           CICS_ATTR_TRIGGER_PULSE_WIDTH,
                                           0, width),
                                           3, "Width");

    viCheckParm( Ivi_SetAttributeViReal64( vi,
                                           VI_NULL,
                                           CICS_ATTR_TRIGGER_PULSE_PERIOD,
                                           0, period),
                                           4, "Period");

    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggerList
 * Purpose:  This function configures how many times the AC source sequences
 *           through a list before that list is completed, specifies the time
 *           interval that each value (point) of a list is to remain in
 *           effect and detemines if a trigger causes a list to advance
 *           only to its next point or to sequence through all of its points.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggerList (
    ViSession       vi,
    ViInt32         repeatCount,
    ViInt32         stepMode,
    ViInt32         listSize,
    ViReal64 _VI_FAR dwellTimes[],
    ViInt32  _VI_FAR repeatTimes[]
)
{
    ViStatus    error = VI_SUCCESS;
    ViInt32     index;
    ViChar*     pdata = VI_NULL;
    ViChar*     pend;
        
    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (dwellTimes && listSize < 1)
        viCheckParm(IVI_ERROR_INVALID_VALUE, 5, "List Size");

    viCheckParm( Ivi_SetAttributeViInt32( vi,
                                       VI_NULL,
                                       CICS_ATTR_TRIGGER_LIST_COUNT,
                                       0, repeatCount),
                                       2, "Repeat Count");

    viCheckParm( Ivi_SetAttributeViInt32( vi,
                                       VI_NULL,
                                       CICS_ATTR_TRIGGER_LIST_MODE,
                                       0, stepMode),
                                       2, "Step Mode");

    if (dwellTimes)
    {
        /*
         * Allocate buffer:
         *  size of allocated bufer
         *  <Header> + listSize*(<size of formated double> + 1 comma)
         *     11
         */
        checkErr( Ivi_Alloc( vi, 11 + (DOUBLE_FORMAT_LENGTH + 1) * listSize, (ViAddr*)&pdata));
        pend = pdata;
        pend += sprintf(pend, ":LIST:DWEL ");
        for (index = 0; index < listSize; index++)
            pend += sprintf(pend, "%Le,", dwellTimes[index]);
        /*
        * Remove last comma
        */
        *--pend = '\0';
        
		if( busInterface == VI_INTF_TCPIP ) 
			pend += sprintf(pend, "\n");//added a LF for LAN
		
        if (!Ivi_Simulating(vi))                /* call only when locked */
        {
            ViSession   io = Ivi_IOSession(vi); /* call only when locked */

            checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

            if(busInterface == VI_INTF_ASRL)
    			Delay(serialDelay);
            viCheckErr( viWrite(io, pdata, ((ViUInt32)pend) - ((ViUInt32)(pdata)), VI_NULL));
        }
    }

	if (repeatTimes)
    {
        /*
         * Allocate buffer:
         *  size of allocated bufer
         *  <Header> + listSize*(<size of formated double> + 1 comma)
         *     11
         */
        checkErr( Ivi_Alloc( vi, 11 + (DOUBLE_FORMAT_LENGTH + 1) * listSize, (ViAddr*)&pdata));
        pend = pdata;
        pend += sprintf(pend, ":LIST:REP ");
        for (index = 0; index < listSize; index++)
            pend += sprintf(pend, "%ld,", repeatTimes[index]);
        /*
        * Remove last comma
        */
        *--pend = '\0';
        
		if( busInterface == VI_INTF_TCPIP ) 
			pend += sprintf(pend, "\n");//added a LF for LAN
		
        if (!Ivi_Simulating(vi))                /* call only when locked */
        {
            ViSession   io = Ivi_IOSession(vi); /* call only when locked */

            checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

            if(busInterface == VI_INTF_ASRL)
    			Delay(serialDelay);
            viCheckErr( viWrite(io, pdata, ((ViUInt32)pend) - ((ViUInt32)(pdata)), VI_NULL));
        }
    }
    
    checkErr( ciCs_CheckStatus (vi));

Error:    
    if (pdata) Ivi_Free(vi, pdata);
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggerVoltageList
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggerVoltageList (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         listSize,
    ViReal64 _VI_FAR voltage[]
)
{
    ViStatus    error = VI_SUCCESS;
    ViInt32     index;
    ViChar*     pdata = VI_NULL;
    ViChar*     pend;

    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (listSize < 1)
        viCheckParm(IVI_ERROR_INVALID_VALUE, 3, "List Size");

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    /*
     * Allocate buffer:
     *  size of allocated bufer
     *  <Header> + listSize*(<size of formated double> + 1 comma)
     *     11
     */
    checkErr( Ivi_Alloc( vi, 11 + (DOUBLE_FORMAT_LENGTH + 1) * listSize, (ViAddr*)&pdata));
    pend = pdata;
    pend += sprintf(pend, ":LIST:VOLT ");
    for (index = 0; index < listSize; index++)
        pend += sprintf(pend, "%Le,", voltage[index]);
    /*
     * Remove last comma
     */
    *--pend = '\0';

	if( busInterface == VI_INTF_TCPIP ) 
			pend += sprintf(pend, "\n");//added a LF for LAN
	
    if (!Ivi_Simulating(vi))                /* call only when locked */
    {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */

        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
        viCheckErr( viWrite(io, pdata, (ViUInt32)(pend - pdata), VI_NULL));
    }

    checkErr( ciCs_CheckStatus (vi));

Error:
    if (pdata) Ivi_Free(vi, pdata);
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggerFrequencyList
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggerFrequencyList (
    ViSession   vi,
    ViInt32     listSize,
    ViReal64 _VI_FAR frequency[]
)
{
    ViStatus    error = VI_SUCCESS;
    ViInt32     index;
    ViChar*     pdata = VI_NULL;
    ViChar*     pend;

    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (listSize < 1)
        viCheckParm(IVI_ERROR_INVALID_VALUE, 2, "List Size");

    /*
     * Allocate buffer:
     *  size of allocated bufer
     *  <Header> + listSize*(<size of formated double> + 1 comma)
     *     11
     */
    checkErr( Ivi_Alloc( vi, 11 + (DOUBLE_FORMAT_LENGTH + 1) * listSize, (ViAddr*)&pdata));
    pend = pdata;
    pend += sprintf(pend, ":LIST:FREQ ");
    for (index = 0; index < listSize; index++)
        pend += sprintf(pend, "%Le,", frequency[index]);
    /*
     * Remove last comma
     */
    *--pend = '\0';

	if( busInterface == VI_INTF_TCPIP ) 
			pend += sprintf(pend, "\n");//added a LF for LAN
	
    if (!Ivi_Simulating(vi))                /* call only when locked */
    {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */

        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
        viCheckErr( viWrite(io, pdata, (ViUInt32)(pend - pdata), VI_NULL));
    }

    checkErr( ciCs_CheckStatus (vi));

Error:
    if (pdata) Ivi_Free(vi, pdata);
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggerFunctionList
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggerFunctionList (
    ViSession       vi,
    ViConstString   functionsList
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar*     pdata = VI_NULL;
    ViChar*     pend;

    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (Ivi_RangeChecking(vi))
    {
        ViChar          buffer[BUFFER_SIZE];
        ViConstString   pbegin = functionsList;
        ViConstString   pend;
        size_t          count;

        while ((pend = strchr(pbegin, ',')) != VI_NULL)
        {
            count = pend-pbegin;
            memcpy(buffer, pbegin, count);
            buffer[count] = '\0';
            checkErr( ciCs_CheckWaveformName( vi, buffer));
            pbegin = pend+1;
        }
    }

    /*
     * Allocate buffer:
     *  size of allocated bufer
     *  <Header> + strlen(functionsList)
     *     11
     */
    checkErr( Ivi_Alloc( vi, 11 + (ViInt32)strlen(functionsList), (ViAddr*)&pdata));
    pend = pdata;
    pend += sprintf(pend, ":LIST:SHAP %s", functionsList);

    if (!Ivi_Simulating(vi))                /* call only when locked */
    {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */

        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
        viCheckErr( viWrite(io, pdata, (ViUInt32)(pend - pdata), VI_NULL));
    }

    checkErr( ciCs_CheckStatus (vi));

Error:
    if (pdata) Ivi_Free(vi, pdata);
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}


/*****************************************************************************
 * Function: ciCs_ConfigureTriggerAngleList
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggerAngleList (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         listSize,
    ViReal64 _VI_FAR angles[]
)
{
    ViStatus    error = VI_SUCCESS;
    ViInt32     index;
    ViChar*     pdata = VI_NULL;
    ViChar*     pend;

    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (listSize < 1)
        viCheckParm(IVI_ERROR_INVALID_VALUE, 3, "List Size");

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 3, "Phase Name");

    /*
     * Allocate buffer:
     *  size of allocated bufer
     *  <Header> + listSize*(<size of formated double> + 1 comma)
     *     11
     */
    checkErr( Ivi_Alloc( vi, 11 + (DOUBLE_FORMAT_LENGTH + 1) * listSize, (ViAddr*)&pdata));
    pend = pdata;
    pend += sprintf(pend, ":LIST:PHAS ");
    for (index = 0; index < listSize; index++)
        pend += sprintf(pend, "%Le,", angles[index]);
    /*
     * Remove last comma
     */
    *--pend = '\0';

	if( busInterface == VI_INTF_TCPIP ) 
			pend += sprintf(pend, "\n");//added a LF for LAN
	
    if (!Ivi_Simulating(vi))                /* call only when locked */
    {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */

        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
        viCheckErr( viWrite(io, pdata, (ViUInt32)(pend - pdata), VI_NULL));
    }

    checkErr( ciCs_CheckStatus (vi));

Error:
    if (pdata) Ivi_Free(vi, pdata);
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}


/*****************************************************************************
 * Function: ciCs_ConfigureTrigSlewVoltageList
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTrigSlewVoltageList (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         listSize,
    ViReal64 _VI_FAR slewVoltage[]
)
{
    ViStatus    error = VI_SUCCESS;
    ViInt32     index;
    ViChar*     pdata = VI_NULL;
    ViChar*     pend;

    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (listSize < 1)
        viCheckParm(IVI_ERROR_INVALID_VALUE, 3, "List Size");

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    /*
     * Allocate buffer:
     *  size of allocated bufer
     *  <Header> + listSize*(<size of formated double> + 1 comma)
     *     16
     */
    checkErr( Ivi_Alloc( vi, 16 + (DOUBLE_FORMAT_LENGTH + 1) * listSize, (ViAddr*)&pdata));
    pend = pdata;
    pend += sprintf(pend, ":LIST:VOLT:SLEW ");
    for (index = 0; index < listSize; index++)
        pend += sprintf(pend, "%Le,", slewVoltage[index]);
    /*
     * Remove last comma
     */
    *--pend = '\0';

	if( busInterface == VI_INTF_TCPIP ) 
			pend += sprintf(pend, "\n");//added a LF for LAN
	
    if (!Ivi_Simulating(vi))                /* call only when locked */
    {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */

        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
        viCheckErr( viWrite(io, pdata, (ViUInt32)(pend - pdata), VI_NULL));
    }

    checkErr( ciCs_CheckStatus (vi));

Error:
    if (pdata) Ivi_Free(vi, pdata);
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureTrigSlewVoltageList
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTrigSlewFrequencyList (
    ViSession       vi,
    ViInt32         listSize,
    ViReal64 _VI_FAR slewFrequency[]
)
{
    ViStatus    error = VI_SUCCESS;
    ViInt32     index;
    ViChar*     pdata = VI_NULL;
    ViChar*     pend;

    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (listSize < 1)
        viCheckParm(IVI_ERROR_INVALID_VALUE, 2, "List Size");

    /*
     * Allocate buffer:
     *  size of allocated bufer
     *  <Header> + listSize*(<size of formated double> + 1 comma)
     *     16
     */
    checkErr( Ivi_Alloc( vi, 16 + (DOUBLE_FORMAT_LENGTH + 1) * listSize, (ViAddr*)&pdata));
    pend = pdata;
    pend += sprintf(pend, ":LIST:FREQ:SLEW ");
    for (index = 0; index < listSize; index++)
        pend += sprintf(pend, "%Le,", slewFrequency[index]);
    /*
     * Remove last comma
     */
    *--pend = '\0';

	if( busInterface == VI_INTF_TCPIP ) 
			pend += sprintf(pend, "\n");//added a LF for LAN
	
    if (!Ivi_Simulating(vi))                /* call only when locked */
    {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */

        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
        viCheckErr( viWrite(io, pdata, (ViUInt32)(pend - pdata), VI_NULL));
    }

    checkErr( ciCs_CheckStatus (vi));

Error:
    if (pdata) Ivi_Free(vi, pdata);
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureTriggerTTLOutputList
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureTriggerTTLOutputList (
    ViSession       vi,
    ViInt32         listSize,
    ViBoolean _VI_FAR TTLOutput[]
)
{
    ViStatus    error = VI_SUCCESS;
    ViInt32     index;
    ViChar*     pdata = VI_NULL;
    ViChar*     pend;

    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (listSize < 1)
        viCheckParm(IVI_ERROR_INVALID_VALUE, 2, "List Size");

    /*
     * Allocate buffer:
     *  size of allocated bufer
     *  <Header> + listSize*(<size of formated double> + 1 comma)
     *     11
     */
    checkErr( Ivi_Alloc( vi, 16 + (DOUBLE_FORMAT_LENGTH + 1) * listSize, (ViAddr*)&pdata));
    pend = pdata;
    pend += sprintf(pend, ":LIST:TTLT ");
    for (index = 0; index < listSize; index++)
        pend += sprintf(pend, "%ld,", (TTLOutput[index]) ? 1 : 0);
    /*
     * Remove last comma
     */
    *--pend = '\0';

	if( busInterface == VI_INTF_TCPIP ) 
			pend += sprintf(pend, "\n");//added a LF for LAN
	
    if (!Ivi_Simulating(vi))                /* call only when locked */
    {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */

        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
        viCheckErr( viWrite(io, pdata, (ViUInt32)(pend - pdata), VI_NULL));
    }

    checkErr( ciCs_CheckStatus (vi));

Error:
    if (pdata) Ivi_Free(vi, pdata);
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ConfigureSynchronizationSource
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureSynchronizationSource (
    ViSession   vi,
    ViInt32     source
)
{
    return Ivi_SetAttributeViInt32( vi,
                                    VI_NULL,
                                    CICS_ATTR_TRIGGER_SYNCHRONIZATION_SOURCE,
                                    IVI_VAL_DIRECT_USER_CALL,
                                    source);
}

/*****************************************************************************
 * Function: ciCs_ConfigureSynchronizationPhase
 * Purpose:  This function configures
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ConfigureSynchronizationPhase (
    ViSession   vi,
    ViReal64    angle
)
{
    return Ivi_SetAttributeViReal64( vi,
                                     VI_NULL,
                                     CICS_ATTR_TRIGGER_SYNCHRONIZATION_PHASE,
                                     IVI_VAL_DIRECT_USER_CALL,
                                     angle);
}

/*****************************************************************************
 * Function: ciCs_SendSoftwareTrigger
 * Purpose:  This function sends a command to trigger the AC source.
 *           Call this function if you configure the AC source to respond
 *           to software triggers. 
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_SendSoftwareTrigger (
    ViSession   vi
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar	    tempCmd[BUFFER_SIZE]; 
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (!Ivi_Simulating(vi))                /* call only when locked */
    {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */
        
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));
        if( busInterface == VI_INTF_TCPIP )
			viCheckErr( viPrintf (io, "*TRG\n"));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, "*TRG");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 	
		}
		else
			viCheckErr( viPrintf (io, "*TRG"));
    }

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_QueryMaxCurrentLimit
 * Purpose:  This function returns the maximum programmable current limit of
 *           the AC source.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_QueryMaxCurrentLimit (
    ViSession       vi,
    ViReal64*       maxCurrentLimit
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    max;
     
    checkErr( Ivi_LockSession (vi, VI_NULL));

 	checkErr( ciCs_GetMaximumRange ( vi,
                                     CICS_VAL_RANGE_CURRENT,
                                     &max));
 
    checkErr( ciCs_CheckStatus (vi));

Error:    
    *maxCurrentLimit = max;
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_QueryMaxVoltageLevel
 * Purpose:  This function returns the maximum programmable voltage level
 *           of the AC source.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_QueryMaxVoltageLevel (
    ViSession       vi,
    ViReal64*       maxVoltageLvel
)   
{
    ViStatus    error = VI_SUCCESS;
    ViReal64   max;  
    
    checkErr( Ivi_LockSession (vi, VI_NULL));

    checkErr( ciCs_GetMaximumRange ( vi,
                                     CICS_VAL_RANGE_VOLTAGE,
                                     &max));

    checkErr( ciCs_CheckStatus (vi));

Error:
	*maxVoltageLvel = max;
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_QueryOutputState
 * Purpose:  This function returns whether the AC source is in
 *           a particular output state.
 *****************************************************************************/
static IviRangeTableEntry outputStateRangeTableEntries[] =
    {
        {CICS_VAL_OUTPUT_UNREGULATED,       0, 0, "", 3},
        {CICS_VAL_OUTPUT_OVER_VOLTAGE,      0, 0, "", 0},
        {CICS_VAL_OUTPUT_OVER_CURRENT,      0, 0, "", 1},
        {CICS_VAL_OUTPUT_OVER_TEMPERATURE,  0, 0, "", 4},
        {IVI_RANGE_TABLE_LAST_ENTRY}
    };
static IviRangeTable outputStateRangeTable =
    {
        IVI_VAL_DISCRETE,
        VI_FALSE,
        VI_FALSE,
        VI_NULL,
        outputStateRangeTableEntries,
    };

ViStatus _VI_FUNC ciCs_QueryOutputState (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         outputState,
    ViBoolean*      inState
)
{
    ViStatus    error = VI_SUCCESS;
    ViInt32     bit;
    ViChar	    tempCmd[BUFFER_SIZE];
 	ViChar		valueStr[100]; 
	ViChar      buffer[BUFFER_SIZE];
	ViUInt32  	read_count;
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (inState == VI_NULL)
        viCheckParm (IVI_ERROR_INVALID_PARAMETER, 4, "Null value for inState.");
    
    viCheckParm( Ivi_GetViInt32EntryFromValue (outputState,
                                               &outputStateRangeTable,
                                               VI_NULL, VI_NULL, VI_NULL, 
                                               VI_NULL, VI_NULL, &bit),
                                               3, "Output State");

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    if (!Ivi_Simulating (vi))
    {
        ViSession   io = Ivi_IOSession(vi);
        ViInt32     condition;
        
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

		if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    }
        	
        if( busInterface == VI_INTF_TCPIP )
        	viCheckErr( viQueryf(io, ":STAT:QUES:INST:ISUM:COND?\n", "%ld", &condition));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, ":STAT:QUES:INST:ISUM:COND?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
			buffer[read_count] = '\0';
		
			condition = atoi(buffer);
		}
		else
			viCheckErr( viQueryf(io, ":STAT:QUES:INST:ISUM:COND?", "%ld", &condition));
		
        *inState = (ViBoolean)((condition & (1L << bit)) ? VI_TRUE : VI_FALSE);
    }
    else
    {
        *inState = VI_FALSE;
    }

    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_ResetOutputProtection
 * Purpose:  This function returns whether the AC source is in
 *           a particular output state.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_ResetOutputProtection (
    ViSession       vi,
    ViConstString   phaseName
)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Ivi_LockSession (vi, VI_NULL));

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    if (!Ivi_Simulating (vi))
    {
        ViSession   io = Ivi_IOSession(vi);
        
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if( busInterface == VI_INTF_TCPIP ) 
			viCheckErr( viPrintf(io, ":OUTP:PROT:CLE\n"));
		else
			viCheckErr( viPrintf(io, ":OUTP:PROT:CLE"));
        
        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
    }

    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_Fetch
 * Purpose:  This function takes a single measurement on the phase you specify.
 *****************************************************************************/
static IviRangeTableEntry measureFunctionsRangeTableEntries[] =
    {
        {CICS_VAL_MEASURE_AC_VOLTAGE_RMS,           0, 0, "VOLT:AC",        0},
        {CICS_VAL_MEASURE_AC_CURRENT_RMS,           0, 0, "CURR:AC",        0},
        {CICS_VAL_MEASURE_VOLTAGE_MAXIMUM,          0, 0, "VOLT:AMPL:MAX",  0},
        {CICS_VAL_MEASURE_VOLTAGE_CRESTFACTOR,      0, 0, "VOLT:CRES",      0},
        {CICS_VAL_MEASURE_AC_POWER,                 0, 0, "POW:AC",         0},
        {CICS_VAL_MEASURE_AC_APPARENT_POWER,        0, 0, "POW:AC:APP",     0},
        {CICS_VAL_MEASURE_AC_REACTIVE_POWER,        0, 0, "POW:AC:REAC",    0},
        {CICS_VAL_MEASURE_AC_TOTAL_POWER,           0, 0, "POW:AC:TOT",     0},
        {CICS_VAL_MEASURE_AC_POWER_FACTOR,          0, 0, "POW:AC:PFAC",    0},
        {CICS_VAL_MEASURE_NEUTRAL_AC_CURRENT_RMS,   0, 0, "CURR:NEUT:AC",   0},
        {CICS_VAL_MEASURE_FREQUENCY,                0, 0, "FREQ",           0},
        {CICS_VAL_MEASURE_PHASE,                	0, 0, "PHAS",           0},
        {IVI_RANGE_TABLE_LAST_ENTRY}
    };

static IviRangeTable measureFunctionsRangeTable =
    {
        IVI_VAL_DISCRETE,
        VI_FALSE,
        VI_FALSE,
        VI_NULL,
        measureFunctionsRangeTableEntries,
    };

ViStatus _VI_FUNC ciCs_Fetch (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         measurementType,
    ViReal64*       measurement
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    value;
    ViString    cmd;
	ViChar	    tempCmd[BUFFER_SIZE];  
	ViChar      buffer[BUFFER_SIZE];
	ViUInt32  	read_count;
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (measurement == VI_NULL)
        viCheckParm (IVI_ERROR_INVALID_PARAMETER, 4, "Null value for inState.");
    
    viCheckParm( Ivi_GetViInt32EntryFromValue (measurementType,
                                               &measureFunctionsRangeTable,
                                               VI_NULL, VI_NULL, VI_NULL, 
                                               VI_NULL, &cmd, VI_NULL),
                                               3, "Measurement Type");

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    if (!Ivi_Simulating (vi))
    {
        ViSession   io = Ivi_IOSession(vi);
        
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    } 
        	
        if( busInterface == VI_INTF_TCPIP )
        	viCheckErr( viQueryf(io, ":FETC:%s?\n", "%Le", cmd, &value));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, ":FETC:");
			strcat(tempCmd, cmd);
			strcat(tempCmd, "?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
				
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
    		buffer[read_count] = '\0';
			
			value = atof(buffer);
		}
		else
			viCheckErr( viQueryf(io, ":FETC:%s?", "%Le", cmd, &value));
    }
    else
    {
        checkErr( ciCs_GetMeasurementSimulation(vi, phaseName, measurementType, &value));
    }

    checkErr( ciCs_CheckStatus (vi));

Error:
    *measurement = value;
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_Measure
 * Purpose:  This function takes a single measurement on the phase you specify.
 *****************************************************************************/
 ViStatus _VI_FUNC ciCs_Measure (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         maxTimeout,
    ViInt32         measurementType,
    ViReal64*       measurement
)
{
    ViStatus    error = VI_SUCCESS;
    ViSession   io = VI_NULL;
    ViUInt32    old_time_out;
    ViReal64    value;
    ViString    cmd;

	ViChar	    tempCmd[BUFFER_SIZE];  
	ViChar      buffer[BUFFER_SIZE];
	ViUInt32  	read_count;
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (measurement == VI_NULL)
        viCheckParm (IVI_ERROR_INVALID_PARAMETER, 4, "Null value for inState.");
    
    viCheckParm( Ivi_GetViInt32EntryFromValue (measurementType,
                                               &measureFunctionsRangeTable,
                                               VI_NULL, VI_NULL, VI_NULL, 
                                               VI_NULL, &cmd, VI_NULL),
                                               3, "Measurement Type");

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    if (!Ivi_Simulating (vi))
    {
        io = Ivi_IOSession(vi);
        
        viCheckErr( viGetAttribute (io, VI_ATTR_TMO_VALUE, &old_time_out));
        viCheckErr( viSetAttribute (io, VI_ATTR_TMO_VALUE, maxTimeout));
        
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    }
        
        if( busInterface == VI_INTF_TCPIP )
		{
			strcpy(tempCmd, ":MEAS:");
			strcat(tempCmd, cmd);
			strcat(tempCmd, "?");
			strcat(tempCmd, "\n"); 
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			
			Delay(0.100);
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
    		buffer[read_count] = '\0';
			 
			value = atof(buffer);	
		}
		else 
		{
			strcpy(tempCmd, ":MEAS:");
			strcat(tempCmd, cmd);
			strcat(tempCmd, "?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
				
			Delay(0.100);
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
    		buffer[read_count] = '\0';
			
			value = atof(buffer);
		}
    }
    else
         checkErr( ciCs_GetMeasurementSimulation(vi, phaseName, measurementType, &value));
    
    checkErr( ciCs_CheckStatus (vi));

Error:
    if (io)
        viSetAttribute (io, VI_ATTR_TMO_VALUE, old_time_out);

    *measurement = value;
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_FetchHarmonic
 * Purpose:  This function takes a harmonic measurement on the phase you 
 *           specify.
 *****************************************************************************/
static IviRangeTableEntry measureHarmonicFunctionsRangeTableEntries[] =
    {
        {CICS_VAL_HARMONIC_VOLTAGE_AMPLITUDE,           0, 0, "VOLT:HARM",      0},
        {CICS_VAL_HARMONIC_VOLTAGE_PHASE,               0, 0, "VOLT:HARM:PHAS", 0},
        {CICS_VAL_HARMONIC_CURRENT_AMPLITUDE,           0, 0, "CURR:HARM",      0},
        {CICS_VAL_HARMONIC_CURRENT_PHASE,               0, 0, "CURR:HARM:PHAS", 0},
        {CICS_VAL_HARMONIC_NEUTRAL_CURRENT_AMPLITUDE,   0, 0, "CURR:NEUT:HARM",     0},
        {CICS_VAL_HARMONIC_NEUTRAL_CURRENT_PHASE,       0, 0, "CURR:NEUT:HARM:PHAS",0},
        {IVI_RANGE_TABLE_LAST_ENTRY}
    };

static IviRangeTable measureHarmonicFunctionsRangeTable =
    {
        IVI_VAL_DISCRETE,
        VI_FALSE,
        VI_FALSE,
        VI_NULL,
        measureHarmonicFunctionsRangeTableEntries,
    };
ViStatus _VI_FUNC ciCs_FetchHarmonic (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         harmonic,
    ViInt32         measurementType,
    ViReal64*       measurement
)
{
    ViStatus    error = VI_SUCCESS;
    ViString    cmd;
    ViReal64    value;
    ViChar	    tempCmd[BUFFER_SIZE], valueStr[100]; 
	ViChar      buffer[BUFFER_SIZE];
	ViUInt32  	read_count;
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    if (measurement == VI_NULL)
        viCheckParm (IVI_ERROR_INVALID_PARAMETER, 4, "Null value for inState.");
    
    viCheckParm( Ivi_GetViInt32EntryFromValue (measurementType,
                                               &measureHarmonicFunctionsRangeTable,
                                               VI_NULL, VI_NULL, VI_NULL, 
                                               VI_NULL, &cmd, VI_NULL),
                                               3, "Measurement Type");

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    if (!Ivi_Simulating (vi))
    {
        ViSession   io = Ivi_IOSession(vi);
        
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

		if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    } 
        	
        if( busInterface == VI_INTF_TCPIP )
			viCheckErr( viQueryf(io, ":FETC:%s?\n %ld", "%Le", cmd, harmonic, &value));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, ":FETC:");
			strcat(tempCmd, cmd);
			strcat(tempCmd, "?");
			strcat(tempCmd, " "); 
			Fmt(valueStr, "%s<%i", harmonic);
			strcpy(tempCmd, valueStr);    
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
				
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
    		buffer[read_count] = '\0';
			
			value = atof(buffer);
		}
		else
			viCheckErr( viQueryf(io, ":FETC:%s? %ld", "%Le", cmd, harmonic, &value));
    }
    else
    {
         checkErr( ciCs_GetHarmonicMeasurementSimulation(vi, phaseName, measurementType, harmonic, &value));
    }


    checkErr( ciCs_CheckStatus (vi));

Error:    
    *measurement = value;
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_MeasureHarmonic
 * Purpose:  This function takes a single harmonic measurement on the
 *           phase you specify.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_MeasureHarmonic (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         timeout,
    ViInt32         harmonic,
    ViInt32         measurementType,
    ViReal64*       measurement
)
{
    ViStatus    error = VI_SUCCESS;
    ViSession   io = VI_NULL;
    ViUInt32    old_time_out;
    ViString    cmd;
    ViReal64    value;
    ViChar	    tempCmd[BUFFER_SIZE], valueStr[100]; 
	ViChar      buffer[BUFFER_SIZE];
	ViUInt32  	read_count;
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    if (measurement == VI_NULL)
        viCheckParm (IVI_ERROR_INVALID_PARAMETER, 4, "Null value for inState.");
    
    viCheckParm( Ivi_GetViInt32EntryFromValue (measurementType,
                                               &measureHarmonicFunctionsRangeTable,
                                               VI_NULL, VI_NULL, VI_NULL, 
                                               VI_NULL, &cmd, VI_NULL),
                                               3, "Measurement Type");

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    if (!Ivi_Simulating (vi))
    {
        io = Ivi_IOSession(vi);
        
        viCheckErr( viGetAttribute (io, VI_ATTR_TMO_VALUE, &old_time_out));
        viCheckErr( viSetAttribute (io, VI_ATTR_TMO_VALUE, timeout));
        
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    }
        	
        if( busInterface == VI_INTF_TCPIP )
        	viCheckErr( viQueryf(io, ":MEAS:%s?\n %ld", "%Le", cmd, harmonic, &value));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, ":MEAS:");
			strcat(tempCmd, cmd);
			strcat(tempCmd, "?");
			strcat(tempCmd, " "); 
			Fmt(valueStr, "%s<%i", harmonic);
			strcpy(tempCmd, valueStr);    
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
				
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
    		buffer[read_count] = '\0';
			
			value = atof(buffer);	
		}
		else
			viCheckErr( viQueryf(io, ":MEAS:%s? %ld", "%Le", cmd, harmonic, &value));
    }
    else
    {
        checkErr( ciCs_GetHarmonicMeasurementSimulation(vi, phaseName, measurementType, harmonic, &value));
    }


    checkErr( ciCs_CheckStatus (vi));

Error:
    if (io)
        viSetAttribute (io, VI_ATTR_TMO_VALUE, old_time_out);
    *measurement = value;
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_FetchArray
 * Purpose:  This function takes a single measurement on the phase you specify.
 *****************************************************************************/
static IviRangeTableEntry measureArrayFunctionsRangeTableEntries[] =
    {
        {CICS_VAL_HARMONIC_VOLTAGE_AMPLITUDE,           0, 0, "VOLT:HARM",      0},
        {CICS_VAL_HARMONIC_VOLTAGE_PHASE,               0, 0, "VOLT:HARM:PHAS", 0},
        {CICS_VAL_HARMONIC_CURRENT_AMPLITUDE,           0, 0, "CURR:HARM",      0},
        {CICS_VAL_HARMONIC_CURRENT_PHASE,               0, 0, "CURR:HARM:PHAS", 0},
        {CICS_VAL_HARMONIC_NEUTRAL_CURRENT_AMPLITUDE,   0, 0, "CURR:NEUT:HARM",     0},
        {CICS_VAL_HARMONIC_NEUTRAL_CURRENT_PHASE,       0, 0, "CURR:NEUT:HARM:PHAS",0},
        {IVI_RANGE_TABLE_LAST_ENTRY}
    };

static IviRangeTable measureArrayFunctionsRangeTable =
    {
        IVI_VAL_DISCRETE,
        VI_FALSE,
        VI_FALSE,
        VI_NULL,
        measureArrayFunctionsRangeTableEntries,
    };
ViStatus _VI_FUNC ciCs_FetchArray (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         measurementType,
    ViInt32         arraySize,
    ViReal64        measurement[],
    ViInt32*        numberOfMeasurements
)
{
    ViStatus    error = VI_SUCCESS;
    ViString    cmd;
    ViChar*     pdata = VI_NULL;
    ViInt32     number_of_values = 0;
    ViInt32     update_timeout = VI_FALSE;
    ViUInt32    old_time_out;
    ViSession   io = VI_NULL;
	ViChar	    tempCmd[BUFFER_SIZE];  
    
    checkErr( Ivi_LockSession (vi, VI_NULL));

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    if (measurement == VI_NULL)
        viCheckParm (IVI_ERROR_INVALID_PARAMETER, 5, "Null value for inState.");

    viCheckParm( Ivi_GetViInt32EntryFromValue (measurementType,
                                               &measureArrayFunctionsRangeTable,
                                               VI_NULL, VI_NULL, VI_NULL, 
                                               VI_NULL, &cmd, VI_NULL),
                                               3, "Measurement Type");

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    if (!Ivi_Simulating (vi))
    {
        ViChar*     pvalue;
        ViReal64*   parray = measurement;
        ViInt32     i;

        io = Ivi_IOSession(vi);

        viCheckErr( viGetAttribute (io, VI_ATTR_TMO_VALUE, &old_time_out));
        update_timeout = VI_TRUE;
        viCheckErr( viSetAttribute (io, VI_ATTR_TMO_VALUE, 20000));

        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    }
        	
        if( busInterface == VI_INTF_TCPIP )
        	viCheckErr( viPrintf(io, ":FETC:ARR:%s?\n", cmd));
		else if(busInterface == VI_INTF_ASRL){ 
			strcpy(tempCmd, ":FETC:ARR:");
			strcat(tempCmd, cmd);
			strcat(tempCmd, "?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		}
		else
			viCheckErr( viPrintf(io, ":FETC:ARR:%s?", cmd));
		
        viCheckErr( ciCs_ReadDataUnknownLength ( io,
                                                 &pdata,
                                                 VI_NULL));
        pvalue = strtok(pdata, ",");
        for (i = 0; i < arraySize && pvalue; i++)
        {
            *parray++ = atof(pvalue);
            pvalue = strtok(NULL, ",");
        }
        
        number_of_values = i-1;

    }
    else
    {
        checkErr( ciCs_GetArrayMeasurementSimulation(vi, phaseName,
                                                     measurementType,
                                                     arraySize, measurement,
                                                     &number_of_values));
    }
	  
    checkErr( ciCs_CheckStatus (vi));

Error:    
    if (update_timeout && io)
        viSetAttribute (io, VI_ATTR_TMO_VALUE, old_time_out);
    if (pdata) free(pdata);
    Ivi_UnlockSession(vi, VI_NULL);
    if (*numberOfMeasurements)
        *numberOfMeasurements = number_of_values;
    return error;
}

/*****************************************************************************
 * Function: ciCs_MeasureArray
 * Purpose:  This function takes a array measurement on the phase you specify.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_MeasureArray (
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         timeOut,
    ViInt32         measurementType,
    ViInt32         arraySize,
    ViReal64        measurement[],
    ViInt32*        numberOfMeasurements
)
{
    ViStatus    error = VI_SUCCESS;
    ViSession   io = VI_NULL;
    ViUInt32    old_time_out;
    ViString    cmd;
    ViChar*     pdata = VI_NULL;
    ViInt32     number_of_values = 0;
	ViChar	    tempCmd[BUFFER_SIZE]; 
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    if (measurement == VI_NULL)
        viCheckParm (IVI_ERROR_INVALID_PARAMETER, 5, "Null value for inState.");

    viCheckParm( Ivi_GetViInt32EntryFromValue (measurementType,
                                               &measureArrayFunctionsRangeTable,
                                               VI_NULL, VI_NULL, VI_NULL, 
                                               VI_NULL, &cmd, VI_NULL),
                                               3, "Measurement Type");

    viCheckParm( ciCs_SetActivePhase(vi, phaseName), 2, "Phase Name");

    if (!Ivi_Simulating (vi))
    {
        ViChar*     pvalue;
        ViReal64*   parray = measurement;
        ViInt32     i;
        
        io = Ivi_IOSession(vi);
        
        viCheckErr( viGetAttribute (io, VI_ATTR_TMO_VALUE, &old_time_out));
        viCheckErr( viSetAttribute (io, VI_ATTR_TMO_VALUE, timeOut));
        

        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    }
        	
        if( busInterface == VI_INTF_TCPIP )
			viCheckErr( viPrintf(io, ":MEAS:ARR:%s?\n", cmd));
		else if(busInterface == VI_INTF_ASRL){ 
			strcpy(tempCmd, ":MEAS:ARR:");
			strcat(tempCmd, cmd);
			strcat(tempCmd, "?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		}
		else
			viCheckErr( viPrintf(io, ":MEAS:ARR:%s?", cmd));
		
        viCheckErr( ciCs_ReadDataUnknownLength ( io,
                                                 &pdata,
                                                 VI_NULL));
        
        pvalue = strtok(pdata, ",");
        for (i = 0; i < arraySize && pvalue; i++)
        {
            *parray++ = atof(pvalue);
            pvalue = strtok(NULL, ",");
        }
      
        number_of_values = i-1;

    }
    else
    {
        checkErr( ciCs_GetArrayMeasurementSimulation(vi, phaseName,
                                                     measurementType,
                                                     arraySize, measurement,
                                                     &number_of_values));
    }
		 
    checkErr( ciCs_CheckStatus (vi));

Error:    
    if (io)
        viSetAttribute (io, VI_ATTR_TMO_VALUE, old_time_out);
    if (pdata) free(pdata);
    Ivi_UnlockSession(vi, VI_NULL);
    
    
    if (*numberOfMeasurements)
        *numberOfMeasurements = number_of_values;
    return error;
}

/*****************************************************************************
 * Function: ciCs_InitiateTransient
 * Purpose:  This function 
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_InitiateTransient (
    ViSession vi
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar	    tempCmd[BUFFER_SIZE];  
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (!Ivi_Simulating (vi))
    {
        ViSession   io = Ivi_IOSession(vi);
    
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if( busInterface == VI_INTF_TCPIP )  
			viCheckErr( viPrintf(io, "INIT:SEQ1\n"));
		else if(busInterface == VI_INTF_ASRL){ 
			strcpy(tempCmd, "INIT:SEQ1");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		}
		else
			viCheckErr( viPrintf(io, "INIT:SEQ1"));
		
        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
    }

    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_InitiateAcquisition
 * Purpose:  This function 
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_InitiateAcquisition (
    ViSession   vi
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar	    tempCmd[BUFFER_SIZE], valueStr[100];   
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (!Ivi_Simulating (vi))
    {
        ViSession   io = Ivi_IOSession(vi);
    
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if( busInterface == VI_INTF_TCPIP )  
			viCheckErr( viPrintf(io, "INIT:SEQ3\n"));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, "INIT:SEQ3");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		}
		else
			viCheckErr( viPrintf(io, "INIT:SEQ3"));
		
        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
    }

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: ciCs_InitiateAcquisition
 * Purpose:  This function 
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_Abort (
    ViSession   vi
)
{
    ViStatus    error = VI_SUCCESS;
	ViChar	    tempCmd[BUFFER_SIZE];   
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (!Ivi_Simulating (vi))
    {
        ViSession   io = Ivi_IOSession(vi);
    
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if( busInterface == VI_INTF_TCPIP ) 
			viCheckErr( viPrintf(io, "ABOR\n"));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, "ABOR");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 	
		}	
		else
        	viCheckErr( viPrintf(io, "ABOR"));
		
        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
    }

    checkErr( ciCs_CheckStatus (vi));

Error:    
    Ivi_UnlockSession(vi, VI_NULL);
    return error;
}

/*****************************************************************************
 *  Function: ciCs_StoreRecallRegister
 *  Purpose:  This function saves and recalls up to 8 settings of the AC
 *            source (register 0 through 7)
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_StoreRecallRegister (ViSession vi,
                                             ViBoolean command,
                                             ViInt16 registers)
{	 
	ViStatus    error = VI_SUCCESS;
	ViChar	    tempCmd[BUFFER_SIZE], valueStr[100];
	
    checkErr( Ivi_LockSession (vi, VI_NULL));

    if (!Ivi_Simulating (vi))
    {
        ViSession   io = Ivi_IOSession(vi);
    
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));

        if (command)
         	if( busInterface == VI_INTF_TCPIP )
         		viCheckErr( viPrintf (io, "*RCL %d\n", registers));                    
			else if(busInterface == VI_INTF_ASRL){
				Fmt(valueStr, "%s<%i", registers);
				strcpy(tempCmd, "*RCL");
				strcat(tempCmd, " "); 
				strcat(tempCmd, valueStr);
				viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
				strcpy(tempCmd, ""); 	
			}
			else
				viCheckErr( viPrintf (io, "*RCL %d", registers));                    
	   	else{
	     	if( busInterface == VI_INTF_TCPIP )
				viCheckErr( viPrintf (io, "*SAV %d\n", registers));  
	     	else if(busInterface == VI_INTF_ASRL){
				Fmt(valueStr, "%s<%i", registers);
				strcpy(tempCmd, "*SAV");
				strcat(tempCmd, " "); 
				strcat(tempCmd, valueStr);
				viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
				strcpy(tempCmd, ""); 	
			}
			else
				viCheckErr( viPrintf (io, "*SAV %d", registers));  
			
	     	Delay( 0.1 );//v1.1.0.2
	    }
        
        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
    }

    checkErr( ciCs_CheckStatus (vi));
    
Error:
	Ivi_UnlockSession (vi, VI_NULL);
	return error;
}

/*****************************************************************************
 * Function: ciCs_GetAttribute<type> Functions                                    
 * Purpose:  These functions enable the instrument driver user to get 
 *           attribute values directly.  There are typesafe versions for 
 *           ViInt32, ViReal64, ViString, ViBoolean, and ViSession attributes.                                         
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_GetAttributeViInt32 (ViSession vi, ViConstString channelName, 
                                                ViAttr attributeId, ViInt32 *value)
{                                                                                                           
    return Ivi_GetAttributeViInt32 (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                    value);
}                                                                                                           
ViStatus _VI_FUNC ciCs_GetAttributeViReal64 (ViSession vi, ViConstString channelName, 
                                                 ViAttr attributeId, ViReal64 *value)
{                                                                                                           
    return Ivi_GetAttributeViReal64 (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                     value);
}                                                                                                           
ViStatus _VI_FUNC ciCs_GetAttributeViString (ViSession vi, ViConstString channelName, 
                                                 ViAttr attributeId, ViInt32 bufSize, 
                                                 ViChar value[]) 
{   
    return Ivi_GetAttributeViString (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                     bufSize, value);
}   
ViStatus _VI_FUNC ciCs_GetAttributeViBoolean (ViSession vi, ViConstString channelName, 
                                                  ViAttr attributeId, ViBoolean *value)
{                                                                                                           
    return Ivi_GetAttributeViBoolean (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                      value);
}                                                                                                           
ViStatus _VI_FUNC ciCs_GetAttributeViSession (ViSession vi, ViConstString channelName, 
                                                  ViAttr attributeId, ViSession *value)
{                                                                                                           
    return Ivi_GetAttributeViSession (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                      value);
}                                                                                                           

/*****************************************************************************
 * Function: ciCs_SetAttribute<type> Functions                                    
 * Purpose:  These functions enable the instrument driver user to set 
 *           attribute values directly.  There are typesafe versions for 
 *           ViInt32, ViReal64, ViString, ViBoolean, and ViSession datatypes.                                         
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_SetAttributeViInt32 (ViSession vi, ViConstString channelName, 
                                                ViAttr attributeId, ViInt32 value)
{                                                                                                           
    return Ivi_SetAttributeViInt32 (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                    value);
}                                                                                                           
ViStatus _VI_FUNC ciCs_SetAttributeViReal64 (ViSession vi, ViConstString channelName, 
                                                 ViAttr attributeId, ViReal64 value)
{                                                                                                           
    return Ivi_SetAttributeViReal64 (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                     value);
}                                                                                                           
ViStatus _VI_FUNC ciCs_SetAttributeViString (ViSession vi, ViConstString channelName, 
                                                 ViAttr attributeId, ViConstString value) 
{   
    return Ivi_SetAttributeViString (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                     value);
}   
ViStatus _VI_FUNC ciCs_SetAttributeViBoolean (ViSession vi, ViConstString channelName, 
                                                  ViAttr attributeId, ViBoolean value)
{                                                                                                           
    return Ivi_SetAttributeViBoolean (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                      value);
}                                                                                                           
ViStatus _VI_FUNC ciCs_SetAttributeViSession (ViSession vi, ViConstString channelName, 
                                                  ViAttr attributeId, ViSession value)
{                                                                                                           
    return Ivi_SetAttributeViSession (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                      value);
}                                                                                                           

/*****************************************************************************
 * Function: ciCs_CheckAttribute<type> Functions                                  
 * Purpose:  These functions enable the instrument driver user to check  
 *           attribute values directly.  These functions check the value you
 *           specify even if you set the CICS_ATTR_RANGE_CHECK 
 *           attribute to VI_FALSE.  There are typesafe versions for ViInt32, 
 *           ViReal64, ViString, ViBoolean, and ViSession datatypes.                         
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_CheckAttributeViInt32 (ViSession vi, ViConstString channelName, 
                                                  ViAttr attributeId, ViInt32 value)
{                                                                                                           
    return Ivi_CheckAttributeViInt32 (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                      value);
}
ViStatus _VI_FUNC ciCs_CheckAttributeViReal64 (ViSession vi, ViConstString channelName, 
                                                   ViAttr attributeId, ViReal64 value)
{                                                                                                           
    return Ivi_CheckAttributeViReal64 (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                       value);
}                                                                                                           
ViStatus _VI_FUNC ciCs_CheckAttributeViString (ViSession vi, ViConstString channelName, 
                                                   ViAttr attributeId, ViConstString value)  
{   
    return Ivi_CheckAttributeViString (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                       value);
}   
ViStatus _VI_FUNC ciCs_CheckAttributeViBoolean (ViSession vi, ViConstString channelName, 
                                                    ViAttr attributeId, ViBoolean value)
{                                                                                                           
    return Ivi_CheckAttributeViBoolean (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                        value);
}                                                                                                           
ViStatus _VI_FUNC ciCs_CheckAttributeViSession (ViSession vi, ViConstString channelName, 
                                                    ViAttr attributeId, ViSession value)
{                                                                                                           
    return Ivi_CheckAttributeViSession (vi, channelName, attributeId, IVI_VAL_DIRECT_USER_CALL, 
                                        value);
}                                                                                                           

/*****************************************************************************
 * Function: ciCs_LockSession and ciCs_UnlockSession Functions                        
 * Purpose:  These functions enable the instrument driver user to lock the 
 *           session around a sequence of driver calls during which other
 *           execution threads must not disturb the instrument state.
 *                                                                          
 *           NOTE:  The callerHasLock parameter must be a local variable 
 *           initialized to VI_FALSE and passed by reference, or you can pass 
 *           VI_NULL.                     
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_LockSession (ViSession vi, ViBoolean *callerHasLock)  
{                                              
    return Ivi_LockSession(vi,callerHasLock);      
}                                              
ViStatus _VI_FUNC ciCs_UnlockSession (ViSession vi, ViBoolean *callerHasLock) 
{                                              
    return Ivi_UnlockSession(vi,callerHasLock);    
}

/*****************************************************************************
 * Function: WriteInstrData and ReadInstrData Functions                      
 * Purpose:  These functions enable the instrument driver user to  
 *           write and read commands directly to and from the instrument.            
 *                                                                           
 *           Note:  These functions bypass the IVI attribute state caching.  
 *                  WriteInstrData invalidates the cached values for all 
 *                  attributes.
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_WriteInstrData (ViSession vi, ViConstString writeBuffer)   
{   
	ViStatus    error = VI_SUCCESS;
    ViUInt32  	read_count;
	ViChar		strBuff[256];    
	
	checkErr( Ivi_LockSession (vi, VI_NULL));
	
Error:      
	Ivi_UnlockSession(vi, VI_NULL);    	
    							  
	
	if(busInterface == VI_INTF_ASRL){
		ViSession   io = Ivi_IOSession(vi); /* call only when locked */
    	viFlush (io, VI_ASRL_IN_BUF);
		
		viWrite (io, (ViBuf)writeBuffer, strlen(writeBuffer), VI_NULL);
	}
	else if( busInterface == VI_INTF_TCPIP ){
		ViSession   io = Ivi_IOSession(vi); /* call only when locked */    
		strcpy( strBuff, (char*)writeBuffer );
		strcat( strBuff, "\n" );
		viCheckErr( viWrite (io, strBuff, strlen(strBuff), &read_count));
	}
	else
		Ivi_WriteInstrData (vi, writeBuffer);    
		  
	return error;
}   

ViStatus _VI_FUNC ciCs_ReadInstrData (
    ViSession   vi,
    ViInt32     numBytes,
    ViChar      rdBuf[],
    ViInt32*    bytesRead
)
{   
    ViStatus    error = VI_SUCCESS;  
    ViUInt32    read_count;
	
	if(busInterface == VI_INTF_ASRL){
		ViSession   io = Ivi_IOSession(vi);
		viRead (io, rdBuf, numBytes-1, &read_count);
    	rdBuf[read_count] = '\0';
		
		*bytesRead = read_count;
	}
	else if( busInterface == VI_INTF_TCPIP ){
		ViSession   io = Ivi_IOSession(vi);      
		viCheckErr( viRead (io, rdBuf, numBytes-1, &read_count));    		
		rdBuf[read_count] = '\0';
	}
	else
		Ivi_ReadInstrData (vi, numBytes, rdBuf, bytesRead);   
	
Error:	
	return error; 
} 

/**************************************************************************** 
 *  Function: ciCs_GetNextInterchangeWarning,
 *            ciCs_ResetInterchangeCheck, and
 *            ciCs_ClearInterchangeWarnings
 *  Purpose:  These functions allow the user to retrieve interchangeability
 *            warnings, reset the driver's interchangeability checking 
 *            state, and clear all previously logged interchangeability warnings.
 ****************************************************************************/
ViStatus _VI_FUNC ciCs_GetNextInterchangeWarning (
    ViSession   vi, 
    ViInt32     bufferSize, 
    ViChar      warnString[]
)
{
    return Ivi_GetNextInterchangeCheckString (vi, bufferSize, warnString);
}

ViStatus _VI_FUNC ciCs_ResetInterchangeCheck (ViSession vi)
{
    return Ivi_ResetInterchangeCheck (vi);
}

ViStatus _VI_FUNC ciCs_ClearInterchangeWarnings (ViSession vi)
{
    return Ivi_ClearInterchangeWarnings (vi);
}

/*****************************************************************************
 * Function: ciCs_GetErrorInfo and ciCs_ClearErrorInfo Functions                       
 * Purpose:  These functions enable the instrument driver user to  
 *           get or clear the error information the driver associates with the
 *           IVI session.                                                        
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_GetErrorInfo (
    ViSession   vi,
    ViStatus*   primaryError,
    ViStatus*   secondaryError,
    ViChar      errorElaboration[256]
)
{                                                                                                           
    return Ivi_GetErrorInfo(vi, primaryError, secondaryError, errorElaboration);                                
}                                                                                                           
ViStatus _VI_FUNC ciCs_ClearErrorInfo (ViSession vi)                                                        
{                                                                                                           
    return Ivi_ClearErrorInfo (vi);                                                                             
}


/*****************************************************************************
 * Function: ciCs_GetError and ciCs_ClearError Functions                       
 * Purpose:  These functions enable the instrument driver user to  
 *           get or clear the error information the driver associates with the
 *           IVI session.                                                        
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_GetError (
    ViSession   vi, 
    ViStatus*   errorCode, 
    ViInt32     bufferSize,
    ViChar      description[]
)
{
    ViStatus    error       = VI_SUCCESS;
    ViStatus    primary     = VI_SUCCESS,
                secondary   = VI_SUCCESS;
    ViChar      elaboration[256]    = "";
    ViChar      errorMessage[1024]  = "";
    ViChar*     appendPoint = errorMessage;
    ViInt32     actualSize = 0;
    ViBoolean   locked = VI_FALSE;

    /* Lock Session */
    if (vi != 0)
        {
        checkErr( Ivi_LockSession(vi, &locked));
        }

    /* Test for nulls and acquire error data */
    if (bufferSize != 0)
    {
        if (errorCode == VI_NULL)
        {
            viCheckParm( IVI_ERROR_INVALID_PARAMETER, 3, "Null address for Error");
        }
        if (description == VI_NULL)
        {
            viCheckParm( IVI_ERROR_INVALID_PARAMETER, 4, "Null address for Description");
        }
        checkErr( Ivi_GetErrorInfo (vi, &primary, &secondary, elaboration));
    }
    else
    {
        checkErr( Ivi_GetAttributeViString( vi,
                                            VI_NULL,
                                            IVI_ATTR_ERROR_ELABORATION,
                                            0, 256, 
                                            elaboration));
        checkErr( Ivi_GetAttributeViInt32( vi,
                                           VI_NULL,
                                           IVI_ATTR_SECONDARY_ERROR,
                                           0, &secondary));
        checkErr( Ivi_GetAttributeViInt32( vi,
                                           VI_NULL,
                                           IVI_ATTR_PRIMARY_ERROR,
                                           0, &primary));
    }
        
    /* Format data */
    if (primary != VI_SUCCESS)
    {
        ViChar  msg[256] = "";

        checkErr( ciCs_error_message (vi, primary, msg));
        appendPoint += sprintf(appendPoint, "Primary Error: (Hex 0x%08X) %s", primary, msg);
    }
    
    if (secondary != VI_SUCCESS)
    {
        ViChar  msg[256] = "";
        checkErr( ciCs_error_message (vi, secondary, msg));
        appendPoint += sprintf(appendPoint, "\nSecondary Error: (Hex 0x%08X) %s", secondary, msg);
    }
    
    if (elaboration[0])
    {
        sprintf(appendPoint, "\nElaboration: %s", elaboration);
    }
    
    actualSize = ((ViInt32)strlen(errorMessage)) + 1;
    
    /* Prepare return values */
    if (bufferSize == 0)
    {
        error = actualSize;
    }
    else 
    {
        if (bufferSize > 0)
        {
            if (actualSize > bufferSize)
            {
                error = actualSize;
                actualSize = bufferSize;
            }
        }
        memcpy(description, errorMessage, actualSize-1);
        description[actualSize-1] = '\0';
    }
    
    if (errorCode) 
    {
        *errorCode = primary;
    }
    
Error:
    /* Unlock Session */
    Ivi_UnlockSession (vi, &locked);
    return error;
}   

ViStatus _VI_FUNC ciCs_ClearError (ViSession vi)                                                        
{                                                                                                           
    return Ivi_ClearErrorInfo (vi);
}

/*****************************************************************************
 * Function: ciCs_GetNextCoercionRecord                        
 * Purpose:  This function enables the instrument driver user to obtain
 *           the coercion information associated with the IVI session.                                                              
 *           This function retrieves and clears the oldest instance in which 
 *           the instrument driver coerced a value the instrument driver user
 *           specified to another value.                     
 *****************************************************************************/
ViStatus _VI_FUNC ciCs_GetNextCoercionRecord (
    ViSession   vi,
    ViInt32     bufferSize,
    ViChar      record[]
)
{
    return Ivi_GetNextCoercionString (vi, bufferSize, record);
}

/*****************************************************************************
 *-------------------- Utility Functions (Not Exported) ---------------------*
 *****************************************************************************/
/*****************************************************************************
 * Function: RemoveSurroundingWhiteSpace
 * Purpose:  This function removes surrounding white space around a string
 *****************************************************************************/
static void RemoveSurroundingWhiteSpace(ViChar string[])
{
    ViChar  *start = string,
            *end;
    size_t lengthWithoutLeadingWS;

    if (start)
    {
        while (*start && isspace((unsigned char)*start))
            start++;

        lengthWithoutLeadingWS = strlen( start);
        memmove( string, start, lengthWithoutLeadingWS);
        end = string + lengthWithoutLeadingWS;
        while (end > string && isspace((unsigned char)*(end-1)))
            end--;

        *end = '\0';
    }
}

/*****************************************************************************
 * Function: RemoveQuotationMarks
 * Purpose:  This function removes quotation marks around a string.
 *****************************************************************************/
static void RemoveQuotationMarks(ViChar string[])
{
    ViChar  *start = string,
            *end;
    
    size_t lengthWithoutLeadingQM;
    
    if (start)
    {
        if (*start == '\"')
            start++;

        lengthWithoutLeadingQM = strlen( start);
        memmove( string, start, lengthWithoutLeadingQM);
        end = string + lengthWithoutLeadingQM - 1;
        while (end > string && *end != '\"')
            end--;
        *end = '\0';
    }
}

/*****************************************************************************
 *  Function: ciCs_GetDriverSetupOption
 *  Purpose:  Get the value of a DriverSetup option. The driver setup string
 *            is part of the option string the user passes to the
 *            ciCs_InitWithOptions function. Specifically it is the part
 *            that follows "DriverSetup=". This function assumes the 
 *            DriverSetup string has the following format:
 *              "Model : OptionValue"
 *****************************************************************************/
static ViStatus ciCs_GetDriverSetupOption(
    ViSession   vi,         /* Session */
    ViInt32*    pModelType  /* Returns Model Type */
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar*     ptag;
    ViChar      setup[BUFFER_SIZE],
                model_buffer[256];

    /* Initialize output variables */
    *pModelType = NOT_SUPPORTED;

    checkErr( Ivi_GetAttributeViString (vi,
                                        VI_NULL,
                                        IVI_ATTR_DRIVER_SETUP,
                                        0, BUFFER_SIZE,
                                        setup));

    if ((ptag = strstr (setup, SETUP_MODEL_TAG)) != VI_NULL)
    {
        sscanf(ptag, "%*[^:]:%256[^;]", model_buffer);
        RemoveSurroundingWhiteSpace (model_buffer);
        if (Ivi_GetValueFromTable (gs_ciCsModelTable, model_buffer, pModelType) < VI_SUCCESS)
        {
            ViChar message[BUFFER_SIZE];

            sprintf (message, "Bad value for %s option in DriverSetup: %s",
                     SETUP_MODEL_TAG,
                     model_buffer);
            viCheckErrElab (IVI_ERROR_BAD_OPTION_VALUE, message);
        }
    }
    
Error:
    return error;
}

/*****************************************************************************
 *  Function: ciCs_GetIDNInformation
 *  Purpose:  This function scans ID Query and sets string attribute. This
 *            function is used for following attributes:
 *                 CICS_ATTR_INSTRUMENT_MODEL
 *                 CICS_ATTR_INSTRUMENT_MANUFACTURER
 *                 CICS_ATTR_INSTRUMENT_FIRMWARE_REVISION
 *****************************************************************************/
static ViStatus ciCs_GetIDNInformation(
    ViSession       vi,
    ViAttr          attributeId,
    ViConstString   formatString
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar      idQ[256];

    checkErr( Ivi_GetAttributeViString (vi, VI_NULL,CICS_ATTR_ID_QUERY_RESPONSE,
                                        0, 256, idQ));
    sscanf (idQ, formatString, idQ);
    checkErr( Ivi_SetValInStringCallback (vi, attributeId, idQ));

Error:
    return error;
}

/*****************************************************************************
 *  Function: ciCs_QueryID
 *  Purpose:  Query the instrument for its id string.
 *****************************************************************************/
static ViStatus ciCs_QueryID(
    ViSession   vi,         /* Driver Session */
    ViSession   io,         /* IO Session */
    ViInt32     bufferSize, /* Size of buffer */
    ViChar      buffer[]    /* buffer for ID string */
)
{
    ViStatus  error = VI_SUCCESS;
    ViUInt32  read_count;
	  
    if(busInterface == VI_INTF_ASRL){
    	viFlush (io, VI_ASRL_IN_BUF);  
    	Delay(serialDelay);
    }
    
	if( busInterface == VI_INTF_TCPIP )
		viCheckErr( viWrite (io, "*IDN?\n", strlen("*IDN?\n"), &read_count));    
	else
		viCheckErr( viWrite (io, "*IDN?", strlen("*IDN?"), VI_NULL));
	
	viCheckErr( viRead (io, buffer, bufferSize-1, &read_count));    		
	buffer[read_count] = '\0';
	
Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_GetModelFromBuffer
 * Purpose:  This function returns model number from buffer. If table 
 *           ciCs_ModelTable does not contains this model, function
 *           returns VI_ERROR_FAIL_ID_QUERY;
 *****************************************************************************/
static ViStatus ciCs_GetModelFromBuffer(
    ViSession       vi,                 /* Driver Session */
    ViConstString   buffer,             /* ID Query Buffer */
    ViInt32*        pModelType)         /* Model type see ESupportedInstruments */
{
    ViStatus    error = VI_SUCCESS;
    ViChar      instrument_buffer[BUFFER_SIZE];

    sscanf(buffer, "%*[^,],%[^,]", instrument_buffer);

    RemoveSurroundingWhiteSpace(instrument_buffer);

    if (Ivi_GetValueFromTable (gs_ciCsModelTable, instrument_buffer, pModelType) < VI_SUCCESS)
        viCheckErr( VI_ERROR_FAIL_ID_QUERY);

Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_GetPhaseIndex
 * Purpose:  This function check if the phaseName is valid. For single phase
 *           mode is valid only one phase. 
 *****************************************************************************/
static ViStatus ciCs_GetPhaseIndex(
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32*        pPhaseIndex,
    ViBoolean*      pIsSinglePhase
)
{
    ViStatus        error = VI_SUCCESS;
    ViInt32         phase_index = 0;
    ViInt32         phase_count, phaseState, limitPhase;
    ViConstString   coerced_phase_name;

	ViChar      	modeState[BUFFER_SIZE] = "";  
    /*
     * Change virtual phase name to driver specific phase name
     */
    checkErr( Ivi_CoerceRepCapName(vi, PHASE_REP_CAP_NAME, phaseName, &coerced_phase_name));

    sscanf(coerced_phase_name, "PHASE%ld", &phase_index);

    /*
     * If only one phase mode is enabled, phase_count must be 1.
     */
    checkErr(Ivi_GetAttributeViString( vi,
                                      VI_NULL,
                                      CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT,
                                      0,
									  BUFFER_SIZE,
									  modeState));

	phaseState = atoi(modeState);  
	
    if( phaseState == 0 ){
		checkErr(Ivi_GetAttributeViInt32( vi,
	                                      VI_NULL,
	                                      CICS_ATTR_INSTR_OUTPUT_PHASE_C,
	                                      0, &limitPhase));
		if( limitPhase == 0 ){
			phase_count = CICS_VAL_1_PHASE;
		}
		else if( limitPhase == 120 ){
			phase_count = CICS_VAL_3_PHASE;
		}
		else{
			phase_count = 2;				
		}
	
	}
	else{
		phase_count = phaseState;	
	}
	
	/*if ((phase_count == CICS_VAL_1_PHASE) &&		
        (phase_index != 1))
    {
        viCheckErr(IVI_ERROR_INVALID_VALUE);
    }*/			

Error:
    if (pPhaseIndex)    *pPhaseIndex = phase_index;
    if (pIsSinglePhase) *pIsSinglePhase = (phase_count == CICS_VAL_1_PHASE);
    return error;
}

/*****************************************************************************
 * Function: ciCs_SetActivePhase
 * Purpose:  This function sets apecified active phase.
 *****************************************************************************/
static ViStatus ciCs_SetActivePhase(
    ViSession       vi,
    ViConstString   phaseName
)
{
    ViStatus        error = VI_SUCCESS;
    ViInt32         phase_index = 0;
    ViBoolean       is_single_phase;

    checkErr( ciCs_GetPhaseIndex( vi,
                                  phaseName,
                                  &phase_index,
                                  &is_single_phase));
    
    //11010x
	if ( strcmp(phaseName, "PHASE1") == 0 || strcmp(phaseName, "PHASE2") == 0 || strcmp(phaseName, "PHASE3") == 0 )
    {
        checkErr( Ivi_SetAttributeViInt32( vi ,
                                       VI_NULL,
                                       CICS_ATTR_INSTR_PHASE_COUPLING,
                                       0,
                                       CICS_VAL_COUPLING_NONE));	
		
		checkErr( Ivi_SetAttributeViInt32( vi,
                                        VI_NULL,
                                        CICS_ATTR_INSTR_ACTIVE_PHASE,
                                        0,
                                        phase_index));
    }
	else if ( strcmp(phaseName, "ALLPHASES") == 0 )
	{
		checkErr( Ivi_SetAttributeViInt32( vi ,
                                       VI_NULL,
                                       CICS_ATTR_INSTR_PHASE_COUPLING,
                                       0,
                                       CICS_VAL_COUPLING_ALL));		
		
	}
		  
Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_CheckWaveformName
 * Purpose:  This function checks if the specified waveform is valid.
 *****************************************************************************/
static ViStatus ciCs_CheckWaveformName(
    ViSession       vi,
    ViConstString   waveformName
)
{
    ViStatus    error = VI_SUCCESS;

    error = ciCs_GetStr2FromStr1 (gs_FunctionStringTable, waveformName, VI_NULL);
    if (error < VI_SUCCESS)
    {
        ViBoolean   is_defined = VI_FALSE;

        checkErr( ciCs_IsUserWaveformDefined(vi, waveformName, &is_defined));

        if (!is_defined)
            viCheckErr(IVI_ERROR_INVALID_VALUE);
    }

Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_IsUserWaveformDefined
 * Purpose:  This function checks if the specified waveform is defined.
 *           If the specified waveform is one from predefined function names
 *           like "SINUSOID","SQUARE","CSINUSOID", function returns an error
 *           and isDefined returns VI_FALSE.
 *           WARNING: DON'T USE STRTOK FUNCTION INSIDE THIS FUNCTION !!!!
 *****************************************************************************/
static ViStatus ciCs_IsUserWaveformDefined(
    ViSession       vi,
    ViConstString   waveformName,
    ViBoolean*      isDefined
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar      buffer[BUFFER_SIZE];
    ViChar      tableList[BUFFER_SIZE];

    *isDefined = VI_FALSE;

    error = ciCs_GetStr2FromStr1 (gs_FunctionStringTable, waveformName, VI_NULL);
    if (error < VI_SUCCESS)
    {
        checkErr( Ivi_GetAttributeViString( vi,
                                            VI_NULL,
                                            CICS_ATTR_INSTR_USER_TABLE_LIST,
                                            0,
                                            BUFFER_SIZE, tableList));
        sprintf(buffer,"\"%s\"", waveformName);
    }
    else
    {
        viCheckErr( IVI_ERROR_INVALID_VALUE);
    }

    *isDefined = (strstr(tableList, buffer) != NULL);

Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_GetNumberUserDefinedWaveforms
 * Purpose:  This function returns number of defined waveforms
 *****************************************************************************/
static ViStatus ciCs_GetNumberUserDefinedWaveforms(
    ViSession       vi,
    ViConstString   waveformName,
    ViInt32*        numberOfWaves
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar      tableList[BUFFER_SIZE];
    ViChar*     pwave;
    ViInt32     count = 0;

    checkErr( Ivi_GetAttributeViString( vi,
                                        VI_NULL,
                                        CICS_ATTR_INSTR_USER_TABLE_LIST,
                                        0,
                                        BUFFER_SIZE, tableList));
    pwave = strtok(tableList, ",");
    while (pwave)
    {
        RemoveQuotationMarks(pwave);
        
        if (ciCs_GetStr2FromStr1 (gs_FunctionStringTable, pwave, VI_NULL) != VI_SUCCESS)
            count++;
        pwave = strtok(NULL, ",");
    }

Error:
    *numberOfWaves = count;
    return error;
}

/*****************************************************************************
 *  Function: ciCs_GetStr2FromStr1
 *  Purpose:  This function searches the string/string table the driver
 *            defines with the ciCsStringTable type. 
 *            Use this function when you know the string in the first table 
 *            column. Function returns pointer to the contents of the
 *            respective second column.
 *****************************************************************************/
static ViStatus ciCs_GetStr2FromStr1 (
    ciCsStringTable     strTable, 
    ViConstString       str1,
    ViConstString*      str2
)
{
    ViStatus    error = VI_SUCCESS;
    ViBoolean   found = VI_FALSE;
    ViInt32     i = 0;
    
    if (str2)
        *str2 = VI_NULL;
    while (!found && strTable[i].col1)
    {
        if (strcmp (strTable[i].col1, str1) == 0)
        {
            if (str2)
                *str2 = strTable[i].col2;
            found = VI_TRUE;
        }
        i++;
    }
    if (!found)
        checkErr( IVI_ERROR_INVALID_VALUE);
Error:
    return error;
}

/*****************************************************************************
 *  Function: ciCs_GetStr1FromStr2
 *  Purpose:  This function searches the string/string table the driver
 *            defines with the ciCsStringTable type. 
 *            Use this function when you know the string in the second table 
 *            column. Function returns pointer to the contents of the
 *            respective first column.
 *****************************************************************************/
static ViStatus ciCs_GetStr1FromStr2 (
    ciCsStringTable     strTable, 
    ViConstString*      str1,
    ViConstString       str2
)
{
    ViStatus    error = VI_SUCCESS;
    ViBoolean   found = VI_FALSE;
    ViInt32     i = 0;
    
    if (str1)
        *str1 = VI_NULL;
    while (!found && strTable[i].col1)
    {
        if (strcmp (strTable[i].col2, str2) == 0)
        {
            if (str1)
                *str1 = strTable[i].col1;
            found = VI_TRUE;
        }
        i++;
    }
    if (!found)
        checkErr( IVI_ERROR_INVALID_VALUE);
Error:
    return error;
}

/*****************************************************************************
 *  Function: ciCs_GetMaximumRange
 *  Purpose:  This function returns maximum possible value for specified range
 *****************************************************************************/
static ViStatus ciCs_GetMaximumRange (
    ViSession       vi,
    ViInt32         rangeType,
    ViReal64*       maximumValue
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    max;// = 200.0; /* Max Voltage Range */

    if (rangeType == CICS_VAL_RANGE_CURRENT)
    {
        ViInt32     model_type;
        ViInt32     phase_count;

		ViChar      modeState[BUFFER_SIZE] = ""; 
		
        checkErr( Ivi_GetAttributeViInt32( vi,
                                            VI_NULL,
                                            CICS_ATTR_INSTR_MODEL_TYPE,
                                            0, &model_type));
        checkErr( Ivi_GetAttributeViString( vi,
                                            VI_NULL,
                                            CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT,
                                            0,
											BUFFER_SIZE,
											modeState));
		
		phase_count = atoi(modeState);
        
        checkErr( Ivi_GetAttributeViReal64(vi, "", CICS_ATTR_CURRENT_MAXIMUM, 0, &max));
        
    }
	else{
		checkErr( Ivi_GetAttributeViReal64(vi, "", CICS_ATTR_VOLTAGE_RANGE_MAXIMUM, 0, &max));
	}

Error:
    *maximumValue = max;
    return error;
}

/*****************************************************************************
 * Function: ciCs_GetMeasurementSimulation                                                 
 * Purpose:  This function reaturs scalar measurement in simulation mode
 *****************************************************************************/
static ViStatus ciCs_GetMeasurementSimulation(
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         measurementType,
    ViReal64*       pMeasurement
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    voltage = 10.0;
    ViReal64    current = 1.0;
    ViInt32     num_phase;
    
    if (!pMeasurement)
        return error;

    checkErr( Ivi_GetAttributeViReal64(vi, phaseName, CICS_ATTR_CURRENT_LEVEL, 0, &current));
    checkErr( Ivi_GetAttributeViReal64(vi, phaseName, CICS_ATTR_VOLTAGE_LIMIT, 0, &voltage));
    
    switch (measurementType)
    {
    case CICS_VAL_MEASURE_AC_VOLTAGE_RMS:
        *pMeasurement = voltage / sqrt(2);
        break;
    case CICS_VAL_MEASURE_AC_CURRENT_RMS:
        *pMeasurement = current / sqrt(2);
        break;
    case CICS_VAL_MEASURE_VOLTAGE_MAXIMUM:
        *pMeasurement = current;
        break;
    case CICS_VAL_MEASURE_VOLTAGE_CRESTFACTOR:
        *pMeasurement = sqrt(2);  /* crest factor = current-peak / current_rms */
        /* modification */
        break;
    case CICS_VAL_MEASURE_AC_POWER:
        *pMeasurement = (voltage / sqrt(2)) * (current / sqrt(2)) * 0.85;
        break;
    case CICS_VAL_MEASURE_AC_APPARENT_POWER:
        *pMeasurement = (voltage / sqrt(2) + voltage * 0.02) * (current / sqrt(2) + current * 0.02); 
            /* apparent power = sqrt( true power^2 + reactive power^2) */
            /* or apparent power = impedance current^2 * impedance     */
        /* modification */
        break;
    case CICS_VAL_MEASURE_AC_REACTIVE_POWER:
        *pMeasurement = (voltage / sqrt(2) + voltage * 0.02) * (current / sqrt(2) + current * 0.02) * 0.53;
        /* modification */
        break;
    case CICS_VAL_MEASURE_AC_TOTAL_POWER:
        checkErr( Ivi_GetAttributeViInt32 (vi, VI_NULL,
                                           CICS_ATTR_OUTPUT_PHASE_COUNT, 0,
                                           &num_phase));
        if (num_phase == CICS_VAL_1_PHASE)
            *pMeasurement = 0.0;
        else
        {
            ViReal64    curr1,
                        curr2,
                        curr3,
                        volt1,
                        volt2,
                        volt3;
            
            checkErr( Ivi_GetAttributeViReal64(vi, "PHASE1", CICS_ATTR_CURRENT_LEVEL, 0, &curr1));
            checkErr( Ivi_GetAttributeViReal64(vi, "PHASE1", CICS_ATTR_VOLTAGE_LIMIT, 0, &volt1));
            checkErr( Ivi_GetAttributeViReal64(vi, "PHASE2", CICS_ATTR_CURRENT_LEVEL, 0, &curr2));
            checkErr( Ivi_GetAttributeViReal64(vi, "PHASE2", CICS_ATTR_VOLTAGE_LIMIT, 0, &volt2));
            checkErr( Ivi_GetAttributeViReal64(vi, "PHASE3", CICS_ATTR_CURRENT_LEVEL, 0, &curr3));
            checkErr( Ivi_GetAttributeViReal64(vi, "PHASE3", CICS_ATTR_VOLTAGE_LIMIT, 0, &volt3));
            
            *pMeasurement = (volt1 / sqrt(2)) * (curr1 / sqrt(2)) * 0.85 +
                            (volt2 / sqrt(2)) * (curr2 / sqrt(2)) * 0.85 +
                            (volt3 / sqrt(2)) * (curr3 / sqrt(2)) * 0.85;
        }
        /* modification */
        break;
    case CICS_VAL_MEASURE_AC_POWER_FACTOR:
        *pMeasurement = 0.85;
        /* modification */
        break;
    case CICS_VAL_MEASURE_NEUTRAL_AC_CURRENT_RMS:
        *pMeasurement = current * 0.08;
        break;
    case CICS_VAL_MEASURE_FREQUENCY:
        {
            ViReal64    frequency;
            
            checkErr( Ivi_GetAttributeViReal64(vi, VI_NULL, CICS_ATTR_FREQUENCY, 0, &frequency));
            *pMeasurement = frequency;
        }
        break;
    default:
        /*
         * TODO: add simulation data.
         */
        *pMeasurement = 0.0;
        break;
    }
    

Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_GetHarmonicMeasurementSimulation                                                 
 * Purpose:  This function reaturs scalar harmonic measurement in simulation mode
 *****************************************************************************/
static ViStatus ciCs_GetHarmonicMeasurementSimulation(
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         measurementType,
    ViInt32         harmonic,
    ViReal64*       pMeasurement
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    voltage,
                current,
                phase;
    
    if (!pMeasurement)
        return error;

    if (harmonic > 1)
    {
        *pMeasurement = 0.0;
        return error;
    }
    
    checkErr( Ivi_GetAttributeViReal64(vi, phaseName, CICS_ATTR_CURRENT_LEVEL, 0, &current));
    checkErr( Ivi_GetAttributeViReal64(vi, phaseName, CICS_ATTR_VOLTAGE_LIMIT, 0, &voltage));
    checkErr( Ivi_GetAttributeViReal64(vi, phaseName, CICS_ATTR_PHASE_ANGLE, 0, &phase));
    
    switch (measurementType)
    {
    case CICS_VAL_HARMONIC_VOLTAGE_AMPLITUDE:
        if (harmonic == 0)
            *pMeasurement = voltage * 0.02;
        else
            *pMeasurement = voltage / sqrt(2);
        break;
    case CICS_VAL_HARMONIC_VOLTAGE_PHASE:
        if (harmonic == 0)
            *pMeasurement = 0.0;
        else
            *pMeasurement = phase;
        break;
    case CICS_VAL_HARMONIC_CURRENT_AMPLITUDE:
        if (harmonic == 0)
            *pMeasurement = current * 0.02;
        else
            *pMeasurement = current / sqrt(2);
        break;
    case CICS_VAL_HARMONIC_CURRENT_PHASE:
        if (harmonic == 0)
            *pMeasurement = 0.0;
        else
            *pMeasurement = phase;
        break;
    case CICS_VAL_HARMONIC_NEUTRAL_CURRENT_AMPLITUDE:
        *pMeasurement = current * 0.02;
        break;
    case CICS_VAL_HARMONIC_NEUTRAL_CURRENT_PHASE:
        *pMeasurement = 0.0;
        break;
    default:
        /*
         * TODO: add simulation data.
         */
        *pMeasurement = 0.0;
        break;
    }
    

Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_GetArrayMeasurementSimulation                                                 
 * Purpose:  This function reaturs array of measurement in simulation mode
 *****************************************************************************/
static ViStatus ciCs_GetArrayMeasurementSimulation(
    ViSession       vi,
    ViConstString   phaseName,
    ViInt32         measurementType,
    ViInt32         arraySize,
    ViReal64        measurement[],
    ViInt32*        number_of_values
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    voltage,
                current,
                frequency,
                phase;
    ViInt32     i;
    
    if (!measurement)
        return error;
    
    if (arraySize < 2)
        return error;
    
    checkErr( Ivi_GetAttributeViReal64(vi, phaseName, CICS_ATTR_CURRENT_LEVEL, 0, &current));
    checkErr( Ivi_GetAttributeViReal64(vi, phaseName, CICS_ATTR_VOLTAGE_LIMIT, 0, &voltage));
    checkErr( Ivi_GetAttributeViReal64(vi, phaseName, CICS_ATTR_PHASE_ANGLE, 0, &phase));
    checkErr( Ivi_GetAttributeViReal64(vi, VI_NULL, CICS_ATTR_FREQUENCY, 0, &frequency));

    switch (measurementType)
    {
    case CICS_VAL_HARMONIC_VOLTAGE_AMPLITUDE:
    case CICS_VAL_HARMONIC_VOLTAGE_PHASE:
    case CICS_VAL_HARMONIC_CURRENT_AMPLITUDE:
    case CICS_VAL_HARMONIC_CURRENT_PHASE:
    case CICS_VAL_HARMONIC_NEUTRAL_CURRENT_AMPLITUDE:
    case CICS_VAL_HARMONIC_NEUTRAL_CURRENT_PHASE:
        *number_of_values = (arraySize < 50) ? arraySize : 50;
        for (i = 0; i < *number_of_values; i++)
            measurement[i] = 0.0;
        break;
    }
    
    switch (measurementType)
    {
    case CICS_VAL_HARMONIC_VOLTAGE_AMPLITUDE:
        measurement[0] = voltage * 0.02;
        measurement[1] = voltage / sqrt(2);
        break;
    case CICS_VAL_HARMONIC_VOLTAGE_PHASE:
        measurement[0] = phase;
        measurement[1] = phase;
        break;
    case CICS_VAL_HARMONIC_CURRENT_AMPLITUDE:
        measurement[0] = current * 0.02;
        measurement[1] = current / sqrt(2);
        break;
    case CICS_VAL_HARMONIC_CURRENT_PHASE:
        measurement[0] = phase;
        measurement[1] = phase;
        break;
    case CICS_VAL_HARMONIC_NEUTRAL_CURRENT_AMPLITUDE:
        measurement[0] = current * 0.02;
        measurement[1] = current * 0.02;
        break;
    case CICS_VAL_HARMONIC_NEUTRAL_CURRENT_PHASE:
        measurement[0] = 0.0;
        measurement[1] = 0.0;
        break;
    }
    

Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_ReadDataUnknownLength                                                 
 * Purpose:  This function reads all data from instrument without specified
 *           length.
 *
 *           Note:  Call this function only when the session is locked.
 *                  Use viCheckErr for error checking.
 *****************************************************************************/
static ViStatus ciCs_ReadDataUnknownLength (
    ViSession       io,
    ViChar**        pdata,
    ViInt32*        returnCount
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar*     pbuffer = VI_NULL;
    ViUInt32    allocated_size,
                readed_bytes = 0;
    ViChar*     pend_buffer;
    
    *pdata = VI_NULL;
    if (returnCount) *returnCount = 0;
    
    /*
     * Allocate Buffer
     */
    allocated_size = BUFFER_SIZE;
    pbuffer = (ViChar *)calloc(allocated_size, sizeof(ViChar));
    if (pbuffer == VI_NULL)
        checkErr( VI_ERROR_ALLOC);

    pend_buffer = pbuffer;
    do
    {
        error = viRead (io, pend_buffer, BUFFER_SIZE, &readed_bytes);
        if (error < VI_SUCCESS)
            checkErr(error);

         /*
          * Buffer was small, reallocate it
          */
        if (readed_bytes == BUFFER_SIZE)
        {
            allocated_size += BUFFER_SIZE;
            pbuffer = (ViChar *)realloc(pbuffer, allocated_size);
            if (pbuffer == VI_NULL)
                checkErr( VI_ERROR_ALLOC);
             /*
              * Set pointer to end of data in rallocated buffer
              */
            pend_buffer = pbuffer + (allocated_size - BUFFER_SIZE);
        }
        else
        {
            /*
             * For sure set end of string on the end of data
             */
            *(pend_buffer+readed_bytes) = '\0'; 
        }
    } while (error == 0x3FFF0006); /* if buffer was small, continue with next read */ 
    
    *pdata = pbuffer;
    
    if (returnCount)
        *returnCount = (allocated_size - BUFFER_SIZE) + readed_bytes;

Error:
    if (error < VI_SUCCESS && pbuffer != VI_NULL)
    {
        free(pbuffer);
        *pdata = VI_NULL;
    }
    return error;
}


/*****************************************************************************
 * Function: ciCsDisableOutputEnabledCache                                                 
 * Purpose:  This function disabled cache for attribute
 *           CICS_ATTR_OUTPUT_ENABLED, when current behavior is trip for any
 *           phase.
 *****************************************************************************/
static ViStatus ciCsDisableOutputEnabledCache (
    ViSession       vi
)
{
    ViStatus        error = VI_SUCCESS;
    ViConstString   phase_name;
    ViInt32         phase_count = CICS_VAL_1_PHASE;
    ViInt32         index = 1;
    ViInt32         voltage_behavior = CICS_VAL_VOLTAGE_REGULATE;
    ViInt32         flags;
    ViBoolean       is_ovp = VI_FALSE;      

	ViInt32         phaseState, limitPhase;
   
	ViChar      	modeState[BUFFER_SIZE] = ""; 
    /*
     * Query how many phases can be configured.
     */
    checkErr(Ivi_GetAttributeViString( vi,
                                      VI_NULL,
                                      CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT,
                                      0,
									  BUFFER_SIZE,
									  modeState));
    
	phaseState = atoi(modeState);
	
	if( phaseState == 0 ){
		checkErr(Ivi_GetAttributeViInt32( vi,
	                                      VI_NULL,
	                                      CICS_ATTR_INSTR_OUTPUT_PHASE_C,
	                                      0, &limitPhase));
		if( limitPhase == 0 ){
			phase_count = CICS_VAL_1_PHASE;
		}
		else if( limitPhase == 120 ){
			phase_count = CICS_VAL_3_PHASE;
		}
		else{
			phase_count = 2;				
		}
	
	}
	else{
		phase_count = phaseState;	
	}
	
	do
    {
        checkErr( Ivi_GetNthRepCapName (vi,
                                        PHASE_REP_CAP_NAME,
                                        index,
                                        &phase_name));
    
        if (phase_name)
        {
            checkErr( Ivi_GetAttributeViInt32( vi,
                                               phase_name,
                                               CICS_ATTR_INSTR_VOLTAGE_LIMIT_BEHAVIOR,
                                               0, &voltage_behavior));
            checkErr( Ivi_GetAttributeViBoolean( vi,
                                               phase_name,
                                               CICS_ATTR_OVP_ENABLED,
                                               0, &is_ovp));
        }
        index++;    
        
        /*
         * If all phases has been tested, exit loop.
         * If current behaviour is trip, exit loop.
         * If phase_count == CICS_VAL_1_PHASE, exit loop.
         */
    } while ( phase_name != VI_NULL &&
              voltage_behavior == CICS_VAL_VOLTAGE_REGULATE &&
              phase_count != CICS_VAL_1_PHASE &&
              is_ovp != VI_FALSE
            );

    /*
     * Disable cache if the voltage_behavior is trip or OCP is enabled for any phase
     */
    checkErr( Ivi_GetAttributeFlags (vi, CICS_ATTR_OUTPUT_ENABLED, &flags));

    if (voltage_behavior == CICS_VAL_VOLTAGE_REGULATE ||is_ovp)
        flags |= IVI_VAL_NEVER_CACHE;
    else
        flags &= ~IVI_VAL_NEVER_CACHE;

    checkErr( Ivi_SetAttributeFlags (vi, CICS_ATTR_OUTPUT_ENABLED, flags));

Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_CheckStatus                                                 
 * Purpose:  This function checks the status of the instrument to detect 
 *           whether the instrument has encountered an error.  This function  
 *           is called at the end of most exported driver functions.  If the    
 *           instrument reports an error, this function returns      
 *           IVI_ERROR_INSTR_SPECIFIC.  The user can set the 
 *           IVI_ATTR_QUERY_INSTR_STATUS attribute to VI_FALSE to disable this 
 *           check and increase execution speed.                                   
 *
 *           Note:  Call this function only when the session is locked.
 *****************************************************************************/
static ViStatus ciCs_CheckStatus (ViSession vi)
{
    ViStatus    error = VI_SUCCESS;

    if (Ivi_QueryInstrStatus (vi) && Ivi_NeedToCheckStatus (vi) && !Ivi_Simulating (vi))
        {
        checkErr( ciCs_CheckStatusCallback (vi, Ivi_IOSession(vi)));
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_FALSE));
        }
        
Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_WaitForOPC                                                  
 * Purpose:  This function waits for the instrument to complete the      
 *           execution of all pending operations.  This function is a        
 *           wrapper for the WaitForOPCCallback.  It can be called from 
 *           other instrument driver functions. 
 *
 *           The maxTime parameter specifies the maximum time to wait for
 *           operation complete in milliseconds.
 *
 *           Note:  Call this function only when the session is locked.
 *****************************************************************************/
static ViStatus ciCs_WaitForOPC (ViSession vi, ViInt32 maxTime) 
{
    ViStatus    error = VI_SUCCESS;

    if (!Ivi_Simulating(vi))
        {
        ViInt32  oldOPCTimeout; 
        
        checkErr( Ivi_GetAttributeViInt32 (vi, VI_NULL, CICS_ATTR_OPC_TIMEOUT, 
                                           0, &oldOPCTimeout));
        Ivi_SetAttributeViInt32 (vi, VI_NULL, CICS_ATTR_OPC_TIMEOUT,        
                                 0, maxTime);

        error = ciCs_WaitForOPCCallback (vi, Ivi_IOSession(vi));

        Ivi_SetAttributeViInt32 (vi, VI_NULL, CICS_ATTR_OPC_TIMEOUT, 
                                 0, oldOPCTimeout);
        viCheckErr( error);
        }
Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_DefaultInstrSetup                                               
 * Purpose:  This function sends a default setup to the instrument.  The    
 *           ciCs_reset function calls this function.  The 
 *           ciCs_IviInit function calls this function when the
 *           user passes VI_FALSE for the reset parameter.  This function is 
 *           useful for configuring settings that other instrument driver 
 *           functions require.    
 *
 *           Note:  Call this function only when the session is locked.
 *****************************************************************************/
static ViStatus ciCs_DefaultInstrSetup (ViSession vi)
{
    ViStatus    error = VI_SUCCESS;
        
    /* Invalidate all attributes */
    checkErr( Ivi_InvalidateAllAttributes (vi));

    if (!Ivi_Simulating(vi))
    {
        ViSession   io = Ivi_IOSession(vi); /* call only when locked */
        checkErr( Ivi_SetNeedToCheckStatus (vi, VI_TRUE));
            
        if( busInterface == VI_INTF_TCPIP )     
			viCheckErr( viPrintf (io, "*CLS;*ESE 1;*SRE 32\n"));
		else if(busInterface == VI_INTF_ASRL)
			viPrintf (io, "*CLS;*ESE 1;*SRE 32"); //ok
		else
			viCheckErr( viPrintf (io, "*CLS;*ESE 1;*SRE 32"));
        
        if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
    }
    else
    {
        ViInt32     phase_count;
		
		ViChar      modeState[BUFFER_SIZE] = "";
		
        /*
         * Set default setting for simulation
         */
        checkErr( Ivi_GetAttributeViString( vi,
                                           VI_NULL,
                                           CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT,
                                           0,
										   BUFFER_SIZE,
										   modeState));
		
		phase_count = atoi(modeState);
		
        if (phase_count != CICS_VAL_1_PHASE) 
        {
            checkErr( Ivi_SetAttributeViReal64( vi ,
                                                "PHASE1",
                                                CICS_ATTR_PHASE_ANGLE,
                                                IVI_VAL_DONT_MARK_AS_SET_BY_USER,
                                                0.0));
            checkErr( Ivi_SetAttributeViReal64( vi ,
                                                "PHASE1",
                                                CICS_ATTR_TRIGGERED_PHASE_ANGLE,
                                                IVI_VAL_DONT_MARK_AS_SET_BY_USER,
                                                0.0));
            
            checkErr( Ivi_SetAttributeViReal64( vi ,
                                                "PHASE2",
                                                CICS_ATTR_PHASE_ANGLE,
                                                IVI_VAL_DONT_MARK_AS_SET_BY_USER,
                                                240.0));
            checkErr( Ivi_SetAttributeViReal64( vi ,
                                                "PHASE2",
                                                CICS_ATTR_TRIGGERED_PHASE_ANGLE,
                                                IVI_VAL_DONT_MARK_AS_SET_BY_USER,
                                                240.0));
            
            checkErr( Ivi_SetAttributeViReal64( vi ,
                                                "PHASE3",
                                                CICS_ATTR_PHASE_ANGLE,
                                                IVI_VAL_DONT_MARK_AS_SET_BY_USER,
                                                120.0));
            checkErr( Ivi_SetAttributeViReal64( vi ,
                                                "PHASE3",
                                                CICS_ATTR_TRIGGERED_PHASE_ANGLE,
                                                IVI_VAL_DONT_MARK_AS_SET_BY_USER,
                                                120.0));
        }
    }
    checkErr( Ivi_SetAttributeViInt32( vi ,
                                       VI_NULL,
                                       CICS_ATTR_INSTR_PHASE_COUPLING,
                                       IVI_VAL_DONT_MARK_AS_SET_BY_USER,
                                       CICS_VAL_COUPLING_NONE));
    /*
     * Magic delay required.
     * Tested with Serial Interface, AMD Duron 800MHz
     * Driver generates TMO error for *ESR? query without this delay.
     */

    Delay(0.1);

    checkErr( ciCsDisableOutputEnabledCache (vi));

Error:
    return error;
}

/*****************************************************************************
 *------------------------ Global Session Callbacks -------------------------*
 *****************************************************************************/
/*****************************************************************************
 * Function: ciCs_CheckStatusCallback                                               
 * Purpose:  This function queries the instrument to determine if it has 
 *           encountered an error.  If the instrument has encountered an 
 *           error, this function returns the IVI_ERROR_INSTRUMENT_SPECIFIC 
 *           error code.  This function is called by the 
 *           ciCs_CheckStatus utility function.  The 
 *           Ivi_SetAttribute and Ivi_GetAttribute functions invoke this 
 *           function when the optionFlags parameter includes the
 *           IVI_VAL_DIRECT_USER_CALL flag.
 *           
 *           The user can disable calls to this function by setting the 
 *           IVI_ATTR_QUERY_INSTR_STATUS attribute to VI_FALSE.  The driver can 
 *           disable the check status callback for a particular attribute by 
 *           setting the IVI_VAL_DONT_CHECK_STATUS flag.
 *****************************************************************************/
static ViStatus _VI_FUNC ciCs_CheckStatusCallback (
    ViSession   vi,
    ViSession   io
)
{
    ViStatus    error = VI_SUCCESS;
    ViInt32     esr = 0; 
    
	ViChar	    tempCmd[BUFFER_SIZE];  
	ViChar      buffer[BUFFER_SIZE];
	ViUInt32  	read_count;
	
    if(busInterface == VI_INTF_ASRL){
    	viFlush (io, VI_ASRL_IN_BUF);  
    	Delay(serialDelay);
    }
    	
    if( busInterface == VI_INTF_TCPIP )
	{
		strcpy(tempCmd, "*ESR?\n");
		viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
		
		Delay( 0.1 );
		viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
		buffer[read_count] = '\0';
		
		esr = atoi(buffer);	
		
	}
	else if(busInterface == VI_INTF_ASRL){
		strcpy(tempCmd, "*ESR?");
		viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
		strcpy(tempCmd, ""); 
			
		viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
		buffer[read_count] = '\0';
		
		esr = atoi(buffer);
	}
	else
		viCheckErr( viQueryf (io, "*ESR?", "%hd", &esr));
        
    if (esr & IEEE_488_2_ERROR_BITS)
    {     
        viCheckErr( IVI_ERROR_INSTR_SPECIFIC);
    }

Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_WaitForOPCCallback                                               
 * Purpose:  This function waits until the instrument has finished processing 
 *           all pending operations.  This function is called by the 
 *           ciCs_WaitForOPC utility function.  The IVI engine invokes
 *           this function in the following two cases:
 *           - Before invoking the read callback for attributes for which the 
 *             IVI_VAL_WAIT_FOR_OPC_BEFORE_READS flag is set.
 *           - After invoking the write callback for attributes for which the 
 *             IVI_VAL_WAIT_FOR_OPC_AFTER_WRITES flag is set.
 *****************************************************************************/
static ViStatus _VI_FUNC ciCs_WaitForOPCCallback (ViSession vi, ViSession io)
{
    ViStatus    error = VI_SUCCESS;
    ViInt32     opcTimeout;
    ViUInt16    statusByte;

	ViChar	    tempCmd[BUFFER_SIZE];    
	
    checkErr( Ivi_GetAttributeViInt32 (vi, VI_NULL, CICS_ATTR_OPC_TIMEOUT, 
                                       0, &opcTimeout));

    viCheckErr( viEnableEvent (io, VI_EVENT_SERVICE_REQ, VI_QUEUE, VI_NULL));

    if( busInterface == VI_INTF_TCPIP )
    	viCheckErr( viPrintf (io, "*OPC\n"));
	else if(busInterface == VI_INTF_ASRL){
		strcpy(tempCmd, "*OPC");
		viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
		strcpy(tempCmd, ""); 
	}
	else
		viCheckErr( viPrintf (io, "*OPC"));

        // wait for SRQ 
    if(busInterface == VI_INTF_ASRL){
		viWaitOnEvent (io, VI_EVENT_SERVICE_REQ, opcTimeout, VI_NULL, VI_NULL);
    	viDisableEvent (io, VI_EVENT_SERVICE_REQ, VI_QUEUE);
	}
	else{
		viCheckErr( viWaitOnEvent (io, VI_EVENT_SERVICE_REQ, opcTimeout, VI_NULL, VI_NULL));
	    viCheckErr( viDisableEvent (io, VI_EVENT_SERVICE_REQ, VI_QUEUE));
	}
	
        // clean up after SRQ 
    if( busInterface == VI_INTF_TCPIP )
    	viCheckErr( viBufWrite (io, "*CLS\n", 4, VI_NULL));  
	else if(busInterface == VI_INTF_ASRL){
		strcpy(tempCmd, "*CLS");
		viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
		strcpy(tempCmd, ""); 
	}
	else
		viCheckErr( viBufWrite (io, "*CLS", 4, VI_NULL));  
	
    viCheckErr( viReadSTB (io, &statusByte));
    
Error:
    viDiscardEvents (io, VI_EVENT_SERVICE_REQ, VI_QUEUE);
    return error;
}

/*****************************************************************************
 *---------------------------- Shared Callbacks -----------------------------*
 *****************************************************************************/
/*****************************************************************************
 * Function: ciCsViInt32_ReadCallback                                               
 * Purpose:  This function queries the integer value of the ViInt32 attribute.
 *****************************************************************************/
static ViStatus _VI_FUNC ciCsViInt32_ReadCallback (
    ViSession       vi,
    ViSession       io,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViInt32*        value
)
{
    ViStatus    error = VI_SUCCESS;
    ViString    cmd;
	ViChar	    tempCmd[BUFFER_SIZE];  
	ViChar      buffer[BUFFER_SIZE];
	ViUInt32  	read_count;
	
    viCheckErr( Ivi_GetStringFromTable (gs_ViInt32Commands, attributeId, &cmd)); 

    if (!Ivi_Simulating (vi)){
        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    }
        
        if( busInterface == VI_INTF_TCPIP )
        	viCheckErr( viQueryf (io, "%s?\n", "%ld", cmd, value));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, cmd);
			strcat(tempCmd, "?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
				
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
    		buffer[read_count] = '\0';
			
			*value = atoi(buffer);
		}
		else
			viCheckErr( viQueryf (io, "%s?", "%ld", cmd, value));
    }    
Error:
    return error;
}

/*****************************************************************************
 * Function: ciCsViInt32_WriteCallback                                               
 * Purpose:  This function sets the integer value of the ViInt32 attribute.
 *****************************************************************************/
static ViStatus _VI_FUNC ciCsViInt32_WriteCallback (
    ViSession       vi,
    ViSession       io,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViInt32         value
)
{
    ViStatus    error = VI_SUCCESS;
    ViString    cmd;
	ViChar	    tempCmd[BUFFER_SIZE], valueStr[100];   
	
    viCheckErr( Ivi_GetStringFromTable (gs_ViInt32Commands, attributeId, &cmd)); 
    
    if (!Ivi_Simulating (vi))
        if( busInterface == VI_INTF_TCPIP )
			viCheckErr( viPrintf (io, "%s %ld\n", cmd, value));
		else if(busInterface == VI_INTF_ASRL){
			Fmt(valueStr, "%s<%i", value);
			strcpy(tempCmd, cmd);
			strcat(tempCmd, " "); 
			strcat(tempCmd, valueStr);
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		}
		else
			viCheckErr( viPrintf (io, "%s %ld", cmd, value));

	if(busInterface == VI_INTF_ASRL)
    	Delay(serialDelay);
    	
Error:
    return error;
}

/*****************************************************************************
 * Function: ciCsViReal64_ReadCallback                                               
 * Purpose:  This function queries the real value of the ViReal64 attribute.
 *****************************************************************************/
static ViStatus _VI_FUNC ciCsViReal64_ReadCallback (ViSession vi, ViSession io,
                                                     ViConstString channelName,
                                                     ViAttr attributeId,
                                                     ViReal64 *value)
{
    ViStatus    error = VI_SUCCESS;
    ViString    cmd;
	ViChar	    tempCmd[BUFFER_SIZE], valueStr[100];
	ViChar      buffer[BUFFER_SIZE];
	ViUInt32  	read_count;
	
    viCheckErr( Ivi_GetStringFromTable (gs_ViReal64Commands, attributeId, &cmd)); 
    /*
     * If phaseName is not NUL, we must first set active phase
     */
    if (channelName != VI_NULL)
        checkErr( ciCs_SetActivePhase(vi, channelName));

    if (!Ivi_Simulating (vi)){
        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    }
        
        if( busInterface == VI_INTF_TCPIP ) 
		{  
			if( strstr( cmd, "?" ) )
				viCheckErr( viQueryf (io, "%s\n", "%Le", cmd, value));
			else
				viCheckErr( viQueryf (io, "%s?\n", "%Le", cmd, value));
		}
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, cmd);
			strcat(tempCmd, "?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
				
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
    		buffer[read_count] = '\0';
			
			*value = atof(buffer);
		}
		else
			viCheckErr( viQueryf (io, "%s?", "%Le", cmd, value));
    }    
Error:
    return error;
}

/*****************************************************************************
 * Function: ciCsViReal64_WriteCallback                                               
 * Purpose:  This function sets the real value of the ViReal64 attribute.
 *****************************************************************************/
static ViStatus _VI_FUNC ciCsViReal64_WriteCallback (ViSession vi, ViSession io,
                                                      ViConstString channelName,
                                                      ViAttr attributeId,
                                                      ViReal64 value)
{
    ViStatus    error = VI_SUCCESS;
    ViString    cmd;
	ViChar	    tempCmd[BUFFER_SIZE], valueStr[100];    
	
    viCheckErr( Ivi_GetStringFromTable (gs_ViReal64Commands, attributeId, &cmd)); 
    /*
     * If phaseName is not NULL, we must first set active phase
     */
    
    if (channelName != VI_NULL)
        checkErr( ciCs_SetActivePhase(vi, channelName));

    if (!Ivi_Simulating (vi))
        if( busInterface == VI_INTF_TCPIP )
			viCheckErr( viPrintf (io, "%s %Le\n", cmd, value));
		else if(busInterface == VI_INTF_ASRL){
			Fmt(valueStr, "%s<%f", value);     
			strcpy(tempCmd, cmd);
			strcat(tempCmd, " ");
			strcat(tempCmd, valueStr);
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		}
		else
			viCheckErr( viPrintf (io, "%s %Le", cmd, value));

	if(busInterface == VI_INTF_ASRL)
    	Delay(serialDelay);
    	
Error:
    return error;
}

/*****************************************************************************
 * Function: ciCsEnum_WriteCallback                                               
 * Purpose:  This function sets the enum value of the ViInt32 attribute.
 *****************************************************************************/
static ViStatus _VI_FUNC ciCsEnum_WriteCallback (ViSession vi, ViSession io,
                                                  ViConstString channelName,
                                                  ViAttr attributeId,
                                                  ViInt32 value)
{
    ViStatus            error = VI_SUCCESS;
    ViString            command,
                        commandValue;
    IviRangeTablePtr    table;
	ViChar	    		tempCmd[BUFFER_SIZE];  
	
    viCheckErr( Ivi_GetStringFromTable (gs_EnumCommands, attributeId, &command)); 
    checkErr( Ivi_GetAttrRangeTable (vi, channelName, attributeId, &table));
    viCheckErr( ciCs_GetCmdFromIntValue(value, table, &commandValue));

    /*
     * If phaseName is not NULL, we must first set active phase
     */
    if (channelName != VI_NULL)
        checkErr( ciCs_SetActivePhase(vi, channelName));

    if (!Ivi_Simulating (vi))
        if( busInterface == VI_INTF_TCPIP )
			viCheckErr( viPrintf (io, "%s %s\n", command, commandValue));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, command);
			strcat(tempCmd, " ");
			strcat(tempCmd, commandValue);
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		}
		else
			viCheckErr( viPrintf (io, "%s %s", command, commandValue));
        
    if(busInterface == VI_INTF_ASRL)
    	Delay(serialDelay);

Error:
    return error;
}

/*****************************************************************************
 * Function: ciCsEnum_ReadCallback                                               
 * Purpose:  This function reads the enum value of the ViInt32 attribute.
 *****************************************************************************/
static ViStatus _VI_FUNC ciCsEnum_ReadCallback (ViSession vi, ViSession io,
                                                 ViConstString channelName,
                                                 ViAttr attributeId,
                                                 ViInt32 *value)
{
    ViStatus            error = VI_SUCCESS;
    ViString            command;
    IviRangeTablePtr    table;
	ViChar	    		tempCmd[BUFFER_SIZE];  
	ViUInt32  			read_count;
	
    viCheckErr( Ivi_GetStringFromTable (gs_EnumCommands, attributeId, &command)); 
    checkErr( Ivi_GetAttrRangeTable (vi, channelName, attributeId, &table));

    /*
     * If phaseName is not NULL, we must first set active phase
     */
    if (channelName != VI_NULL)
        checkErr( ciCs_SetActivePhase(vi, channelName));

    if (!Ivi_Simulating (vi))
    {
        ViChar  buffer[BUFFER_SIZE];
        ViInt32 buffer_size = sizeof(buffer);

		if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    }
			
        if( busInterface == VI_INTF_TCPIP )
			viCheckErr( viPrintf (io, "%s?\n", command));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, command);
			strcat(tempCmd, "?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		}
		else
			viCheckErr( viPrintf (io, "%s?", command));
		
        if(busInterface == VI_INTF_ASRL){ 
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
    		buffer[read_count] = '\0';
		}
		else
			viCheckErr( viScanf(io, "%#s", &buffer_size, buffer));

        viCheckErr( ciCs_GetValueFromCmd(buffer, table, value));
    }
Error:
    return error;
}

/*****************************************************************************
 * Function: ciCsViBoolean_ReadCallback                                               
 * Purpose:  This function queries the bool value of the ViBoolean attribute.
 *****************************************************************************/
static ViStatus _VI_FUNC ciCsViBoolean_ReadCallback (ViSession vi, ViSession io,
                                                      ViConstString channelName,
                                                      ViAttr attributeId,
                                                      ViBoolean *value)
{
    ViStatus    error = VI_SUCCESS;
    ViString    cmd;
	ViChar	    tempCmd[BUFFER_SIZE];  
	ViChar      buffer[BUFFER_SIZE];
	ViUInt32  	read_count;
	
    viCheckErr( Ivi_GetStringFromTable (gs_ViBooleanCommands, attributeId, &cmd)); 
    /*
     * If phaseName is not NULL, we must first set active phase
     */
    if (channelName != VI_NULL)
        checkErr( ciCs_SetActivePhase(vi, channelName));

    if (!Ivi_Simulating (vi)){
        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    } 
        	
        if( busInterface == VI_INTF_TCPIP )	
        	viCheckErr( viQueryf (io, "%s?\n", "%hd", cmd, value));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, cmd);
			strcat(tempCmd, "?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		
			viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
			buffer[read_count] = '\0';
		
			*value = atoi(buffer);
		}
		else
			viCheckErr( viQueryf (io, "%s?", "%hd", cmd, value));
	}
	
Error:
    return error;
}

/*****************************************************************************
 * Function: ciCsViBoolean_WriteCallback                                               
 * Purpose:  This function sets the bool value of the ViBoolean attribute.
 *****************************************************************************/
static ViStatus _VI_FUNC ciCsViBoolean_WriteCallback (ViSession vi,
                                                       ViSession io,
                                                       ViConstString channelName,
                                                       ViAttr attributeId,
                                                       ViBoolean value)
{
    ViStatus    error = VI_SUCCESS;
    ViString    cmd;
	ViChar	    tempCmd[BUFFER_SIZE], valueStr[100];   
	
    viCheckErr( Ivi_GetStringFromTable (gs_ViBooleanCommands, attributeId, &cmd)); 
    
    /*
     * If phaseName is not NULL, we must first set active phase
     */
    if (channelName != VI_NULL)
        checkErr( ciCs_SetActivePhase(vi, channelName));

    if (!Ivi_Simulating (vi))
        if( busInterface == VI_INTF_TCPIP )	
			viCheckErr( viPrintf (io, "%s %ld\n", cmd, (value) ? 1 : 0));
		else if(busInterface == VI_INTF_ASRL){ 
			Fmt(valueStr, "%s<%i", value);
			strcpy(tempCmd, cmd);
			strcat(tempCmd, " "); 
			strcat(tempCmd, valueStr);
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		}
		else
			viCheckErr( viPrintf (io, "%s %ld", cmd, (value) ? 1 : 0));
        
    if(busInterface == VI_INTF_ASRL)
    	Delay(serialDelay);

Error:
    return error;
}

/*****************************************************************************
 * Function: ciCsViString_WriteCallback                                               
 * Purpose:  This function sets the string value of the ViString attribute.
 *****************************************************************************/
static ViStatus _VI_FUNC ciCsViString_WriteCallback (
    ViSession       vi,
    ViSession       io,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViConstString   value)
{
    ViStatus    error = VI_SUCCESS;
    ViString    command;
	ViChar	    tempCmd[BUFFER_SIZE]; 
	
    viCheckErr( Ivi_GetStringFromTable (gs_ViStringCommands, attributeId, &command)); 
    if (!Ivi_Simulating (vi))
        if( busInterface == VI_INTF_TCPIP )
			viCheckErr( viPrintf (io, "%s %s\n", command, value));
		else if(busInterface == VI_INTF_ASRL){ 
			strcpy(tempCmd, command);
			strcat(tempCmd, " "); 
			strcat(tempCmd, value);
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		}
		else
			viCheckErr( viPrintf (io, "%s %s", command, value));

	if(busInterface == VI_INTF_ASRL)
    	Delay(serialDelay);
    	
Error:
    return error;
}

/*****************************************************************************
 * Function: ciCsViString_ReadCallback                                               
 * Purpose:  This function reads the string value of the ViString attribute.
 *****************************************************************************/
static ViStatus _VI_FUNC ciCsViString_ReadCallback (
    ViSession           vi,
    ViSession           io,
    ViConstString       channelName,
    ViAttr              attributeId,
    const ViConstString cacheValue)
{
    ViStatus    error = VI_SUCCESS;
    ViString    command;
    ViChar*     pdata = VI_NULL;
    ViChar*     pend  = VI_NULL;
    
    viCheckErr( Ivi_GetStringFromTable (gs_ViStringCommands, attributeId, &command)); 

    if (!Ivi_Simulating (vi))
    {
        ViChar      buffer[BUFFER_SIZE];
        ViChar*     pbuffer = VI_NULL;
        ViUInt32    allocated_count = 0;
        ViUInt32    count = 0;
        ViUInt32    readed_count;
		ViChar	    tempCmd[BUFFER_SIZE];  
		
        if(busInterface == VI_INTF_ASRL){
	    	viFlush (io, VI_ASRL_IN_BUF);  
	    	Delay(serialDelay);
	    }
        	
        if( busInterface == VI_INTF_TCPIP )
			viCheckErr( viPrintf (io, "%s?\n", command));
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, command);
			strcat(tempCmd, "?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 
		}
		else
			viCheckErr( viPrintf (io, "%s?", command));
        
        while ((error = viRead (io, buffer, BUFFER_SIZE, &readed_count)) == VI_SUCCESS_MAX_CNT)
        {
            if (allocated_count)
            {
                if (allocated_count <= count + readed_count)
                {
                    allocated_count += (allocated_count - count) + BUFFER_SIZE; 
                    pdata = realloc (pdata, allocated_count);
                }

                pend = pdata + count;
            }
            else
            {
                allocated_count = 2 * BUFFER_SIZE;
                pdata = malloc(allocated_count);
                pend  = pdata;
            }
            
            memcpy(pend, buffer, readed_count);
            count += readed_count;
        }
        
        if (readed_count > 0)
        {
            /*
             * Check if the last character is alnum.
             * if this character is not alnum we know that it's '\n'.
             * Remove it.
             */
            if (!isalnum(buffer[readed_count-1]))
                readed_count--;
            buffer[readed_count] = '\0';
        }
        
        /*
         * If function is read more data as BUFFER_SIZE, we must
         * copy last data from static buffer to dynamic buffer
         */
        if (pdata)
        {
            /*
             * For sure, check if the  buffer has enought element.
             * This condition should be FALSE every time.
             */
            if (allocated_count <= count + readed_count)
            {
                allocated_count += (allocated_count - count) + BUFFER_SIZE; 
                pdata = realloc (pdata, allocated_count);
            }
            pend = pdata + count;
            memcpy(pend, buffer, readed_count+1);           
            pbuffer = pdata;
        }
        else
            pbuffer = buffer;
        
        viCheckErr( error);
        checkErr( Ivi_SetValInStringCallback(vi, attributeId, pbuffer));
    }

Error:
    if (pdata)
        free( pdata);
    return error;
}

/*****************************************************************************
 *----------------- Attribute Range Tables and Callbacks --------------------*
 *****************************************************************************/
/*- CICS_ATTR_ID_QUERY_RESPONSE -*/
static ViStatus _VI_FUNC ciCsAttrIdQueryResponse_ReadCallback (
    ViSession           vi, 
    ViSession           io,
    ViConstString       channelName, 
    ViAttr              attributeId,
    const ViConstString cacheValue
)
{
    ViStatus    error = VI_SUCCESS;
    ViChar      read_buffer[BUFFER_SIZE];
    
    checkErr( ciCs_QueryID( vi, io, BUFFER_SIZE, read_buffer));
    checkErr( Ivi_SetValInStringCallback (vi, attributeId, read_buffer));
    
Error:
    return error;
}
    
/*- CICS_ATTR_SPECIFIC_DRIVER_REVISION -*/
static ViStatus _VI_FUNC ciCsAttrDriverRevision_ReadCallback (ViSession vi, 
                                                                  ViSession io,
                                                                  ViConstString channelName, 
                                                                  ViAttr attributeId,
                                                                  const ViConstString cacheValue)
{
    ViStatus    error = VI_SUCCESS;
    ViChar      driverRevision[256];
    
    
    sprintf (driverRevision, 
             "Driver: ciCs %.2f, Compiler: %s %3.2f, "
             "Components: IVIEngine %.2f, VISA-Spec %.2f",
             CICS_MAJOR_VERSION + CICS_MINOR_VERSION/1000.0, 
             IVI_COMPILER_NAME, IVI_COMPILER_VER_NUM, 
             IVI_ENGINE_MAJOR_VERSION + IVI_ENGINE_MINOR_VERSION/1000.0, 
             Ivi_ConvertVISAVer(VI_SPEC_VERSION));

    checkErr( Ivi_SetValInStringCallback (vi, attributeId, driverRevision));    
Error:
    return error;
}

/*- CICS_ATTR_INSTRUMENT_FIRMWARE_REVISION -*/
static ViStatus _VI_FUNC ciCsAttrFirmwareRevision_ReadCallback (
    ViSession           vi, 
    ViSession           io,
    ViConstString       channelName, 
    ViAttr              attributeId,
    const ViConstString cacheValue)
{
    return ciCs_GetIDNInformation( vi,
                                   CICS_ATTR_INSTRUMENT_FIRMWARE_REVISION,
                                   "%*[^,],%*[^,],%*[^,],%256[^\n]");
}

/*- CICS_ATTR_INSTRUMENT_MANUFACTURER -*/
static ViStatus _VI_FUNC ciCsAttrInstrumentManufacturer_ReadCallback (
    ViSession           vi, 
    ViSession           io,
    ViConstString       channelName, 
    ViAttr              attributeId,
    const ViConstString cacheValue
)
{
    return ciCs_GetIDNInformation( vi,
                                   CICS_ATTR_INSTRUMENT_MANUFACTURER,
                                   "%256[^,]");
}
    

/*- CICS_ATTR_INSTRUMENT_MODEL -*/
static ViStatus _VI_FUNC ciCsAttrInstrumentModel_ReadCallback (
    ViSession           vi, 
    ViSession           io,
    ViConstString       channelName, 
    ViAttr              attributeId,
    const ViConstString cacheValue
)
{
    return ciCs_GetIDNInformation( vi,
                                   CICS_ATTR_INSTRUMENT_MODEL,
                                   "%*[^,],%256[^,]");
}

/*- CICS_ATTR_OUTPUT_ALC_STATE -*/  
static IviRangeTableEntry attrOutputALCStateRangeTableEntries[] =
	{
		{CICS_VAL_ALC_ON, 		0, 0, "ON", 0},
		{CICS_VAL_ALC_OFF, 		0, 0, "OFF", 0},
		{CICS_VAL_ALC_REGULATE, 0, 0, "REG", 0},
		{IVI_RANGE_TABLE_LAST_ENTRY			}
	};

static IviRangeTable attrOutputALCStateRangeTable =
	{
		IVI_VAL_DISCRETE,
        VI_FALSE,
        VI_FALSE,
        "VI_FALSE",
        attrOutputALCStateRangeTableEntries,
	};
	
/*- CICS_ATTR_OUTPUT_PHASE_COUNT -*/
static IviRangeTableEntry attrOutputPhaseCountRangeTableEntries[] =
{
    {CICS_VAL_1_PHASE,   0,  0,  "1",  0}, 
    {CICS_VAL_3_PHASE,   0,  0,  "3",  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrOutputPhaseCountRangeTable =
{
    IVI_VAL_DISCRETE, 
    VI_FALSE, 
    VI_FALSE, 
    "", 
    attrOutputPhaseCountRangeTableEntries, 
};

static ViStatus _VI_FUNC ciCsAttrOutputPhaseCount_ReadCallback (
    ViSession       vi, 
    ViSession       io,
    ViConstString   channelName, 
    ViAttr          attributeId,
    ViInt32*        value
)
{
    return Ivi_GetAttributeViInt32( vi,
                                    channelName,
                                    CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT,
                                    0,
                                    value);
}

static ViStatus _VI_FUNC ciCsAttrOutputPhaseCount_WriteCallback (
    ViSession       vi, 
    ViSession       io,
    ViConstString   channelName, 
    ViAttr          attributeId,
    ViInt32         value
)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Ivi_SetAttributeViInt32( vi,
                                       channelName,
                                       CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT,
                                       0,
                                       value));
    
    Delay( 0.1 );
    checkErr( ciCs_DefaultInstrSetup(vi));
    
Error:
    return error;
}

static ViStatus _VI_FUNC ciCsAttrOutputPhaseMode_ReadCallback (
    ViSession       vi, 
    ViSession       io,
    ViConstString   channelName, 
    ViAttr          attributeId,
    ViInt32*        value
)
{
    return Ivi_GetAttributeViInt32( vi,
                                    channelName,
                                    CICS_ATTR_OUTPUT_PHASE_MODE,
                                    0,
                                    value);
}

static ViStatus _VI_FUNC ciCsAttrOutputPhaseMode_WriteCallback (
    ViSession       vi, 
    ViSession       io,
    ViConstString   channelName, 
    ViAttr          attributeId,
    ViInt32         value
)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Ivi_SetAttributeViInt32( vi,
                                       channelName,
                                       CICS_ATTR_OUTPUT_PHASE_MODE,
                                       0,
                                       value));
                                       
	Delay( 0.1 );                                       
    
Error:
    return error;
}

/*- Shared Read Callback ciCsViInt32_ReadCallback  -*/

/*- CICS_ATTR_OUTPUT_ENABLED -*/
/*- Shared Write Callback ciCsViBoolean_WriteCallback  -*/
/*- Shared Read Callback ciCsViBoolean_ReadCallback  -*/

/*- CICS_ATTR_OVP_ENABLED -*/
static ViStatus _VI_FUNC ciCsAttrOvpEnabled_CheckCallback(
    ViSession       vi,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViBoolean       value
)
{
#if defined(OVERVOLTAGE_PROTECTION_STATE_SUPPORTED) 
    return ciCs_GetPhaseIndex(vi, phaseName, VI_NULL, VI_NULL); 
#else
    ViStatus    error = VI_SUCCESS;
    viCheckErr((!value) ? IVI_ERROR_INVALID_VALUE : error);
Error:
    return error;
#endif
}

static ViStatus _VI_FUNC ciCsAttrOvpEnabled_WriteCallback (
    ViSession       vi, 
    ViSession       io,
    ViConstString   phaseName, 
    ViAttr          attributeId,
    ViBoolean       value
)
{
    ViStatus    error = VI_SUCCESS;

    checkErr( Ivi_SetAttributeViBoolean(vi, phaseName, CICS_ATTR_INSTR_OVP_ENABLED, 0, value));
    checkErr( ciCsDisableOutputEnabledCache(vi));

Error:
    return error;
}

static ViStatus _VI_FUNC ciCsAttrOvpEnabled_ReadCallback (
    ViSession       vi, 
    ViSession       io,
    ViConstString   phaseName, 
    ViAttr          attributeId,
    ViBoolean*      value
)
{
    return Ivi_GetAttributeViBoolean(vi, phaseName, CICS_ATTR_INSTR_OVP_ENABLED, 0, value);
}

/*- CICS_ATTR_INSTR_PHASE_COUPLING -*/
static IviRangeTableEntry attrInstrPhaseCouplingRangeTableEntries[] =
{
    {CICS_VAL_COUPLING_NONE,   0,  0,  "NONE",  0}, 
    {CICS_VAL_COUPLING_ALL,    0,  0,  "ALL",   0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrInstrPhaseCouplingRangeTable =
{
    IVI_VAL_DISCRETE, 
    VI_FALSE, 
    VI_FALSE, 
    "", 
    attrInstrPhaseCouplingRangeTableEntries, 
};
/*- Shared Write Callback ciCsEnum_WriteCallback  -*/
/*- Shared Read Callback ciCsEnum_ReadCallback  -*/

/*- CICS_ATTR_OVP_LIMIT -*/
static IviRangeTableEntry attrOvpLimitRangeTableEntries[] =
{
    {0.0,   500.0,  0,  NULL,  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrOvpLimitRangeTable =
{
    IVI_VAL_RANGED, 
    VI_TRUE, 
    VI_TRUE, 
    "", 
    attrOvpLimitRangeTableEntries, 
};

static ViStatus _VI_FUNC ciCsPhaseNameViReal64_CheckCallback(
    ViSession       vi,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViReal64        value
)
{
    ViStatus    error = VI_SUCCESS;

    checkErr( Ivi_DefaultCheckCallbackViReal64 ( vi,
                                                 phaseName,
                                                 attributeId,
                                                 value));
    checkErr( ciCs_GetPhaseIndex(vi, phaseName, VI_NULL, VI_NULL)); 
Error:
    return error;
}

static ViStatus _VI_FUNC ciCsAttrOVPLimit_CoerceCallback(
    ViSession       vi,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViReal64        value,
    ViReal64*       coercedValue
)
{
    ViReal64    nb_of_digits = floor(log10(fabs(value)));
    ViReal64    resolution = 10000.0 / ((nb_of_digits > 0.0) ? nb_of_digits * 10.0 : 1.0);
    
    *coercedValue = floor(value*resolution + 0.5)/resolution;

    return VI_SUCCESS;
}
/*- Shared Write Callback ciCsViReal64_WriteCallback  -*/
/*- Shared Read Callback ciCsViReal64_ReadCallback  -*/

/*- CICS_ATTR_VOLTAGE_LIMIT_BEHAVIOR -*/
static IviRangeTableEntry attrVoltageLimitBehaviorRangeTableEntries[] =
{
    {CICS_VAL_VOLTAGE_REGULATE,   0,  0,  "0",  0}, 
    {CICS_VAL_VOLTAGE_TRIP,       0,  0,  "1",  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrVoltageLimitBehaviorRangeTable =
{
    IVI_VAL_DISCRETE, 
    VI_FALSE, 
    VI_FALSE, 
    "", 
    attrVoltageLimitBehaviorRangeTableEntries, 
};

static ViStatus _VI_FUNC ciCsAttrVoltageLimitBehavior_ReadCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViInt32*        value
)
{
    return Ivi_GetAttributeViInt32( vi,
                                    phaseName,
                                    CICS_ATTR_INSTR_VOLTAGE_LIMIT_BEHAVIOR,
                                    0, value);
}

static ViStatus _VI_FUNC ciCsAttrVoltageLimitBehavior_WriteCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViInt32         value
)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Ivi_SetAttributeViInt32( vi,
                                       phaseName,
                                       CICS_ATTR_INSTR_VOLTAGE_LIMIT_BEHAVIOR,
                                       0, value));
    checkErr( ciCsDisableOutputEnabledCache(vi));

Error:
    return error;
}

/*- CICS_ATTR_INSTR_CURRENT_LIMIT_BEHAVIOR (hidden) -*/
/*- Range table attrCurrentLimitBehaviorRangeTable defined above -*/
/*- Shared Write Callback ciCsEnum_WriteCallback  -*/
/*- Shared Read Callback ciCsEnum_ReadCallback  -*/

/*- CICS_ATTR_CURRENT_LIMIT -*/
/*static ViStatus _VI_FUNC ciCsAttrCurrentLimit_CheckCallback(
    ViSession       vi,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViReal64        value
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    max_current;

    checkErr( Ivi_GetAttributeViReal64( vi,
                                        phaseName,
                                        CICS_ATTR_INSTR_CURRENT_RANGE,
                                        0, &max_current));

    if (value > max_current || value < 0.0)
        viCheckErr(IVI_ERROR_INVALID_VALUE);

Error:
    return error;
} */
/*- Shared Write Callback ciCsViReal64_WriteCallback  -*/
/*- Shared Read Callback ciCsViReal64_ReadCallback  -*/

/*- CILXS_ATTR_VOLTAGE_LIMIT -*/
static ViStatus _VI_FUNC ciCsAttrVoltageLimit_CheckCallback(
    ViSession       vi,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViReal64        value
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    max_voltage;

    checkErr( Ivi_GetAttributeViReal64( vi,
                                        phaseName,
                                        CICS_ATTR_INSTR_VOLTAGE_RANGE,
                                        0, &max_voltage));
    
    if (value > max_voltage || value < 0.0)
        viCheckErr(IVI_ERROR_INVALID_VALUE);

Error:
    return error;
}

/*- CICS_ATTR_CURRENT_LEVEL -*/
static ViStatus _VI_FUNC ciCsAttrCurrentLevel_CheckCallback(
    ViSession       vi,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViReal64        value
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    max_current;

    checkErr( ciCs_GetPhaseIndex(vi, phaseName, VI_NULL, VI_NULL));
    
    checkErr( Ivi_GetAttributeViReal64( vi,
                                        "",
                                        CICS_ATTR_CURRENT_MAXIMUM,
                                        0, &max_current));
    
    if (value > max_current || value < 0.0)
        viCheckErr(IVI_ERROR_INVALID_VALUE);

Error:
    return error;
}

/*- CICS_ATTR_FREQUENCY -*/
static ViStatus _VI_FUNC ciCsAttrFrequencyValue_CheckCallback (ViSession vi,
                                                                ViConstString channelName,
                                                                ViAttr attributeId,
                                                                ViReal64 value)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    min_frequency, max_frequency;

    checkErr( Ivi_GetAttributeViReal64( vi,
                                        "",
                                        CICS_ATTR_FREQUENCY_MINIMUM,
                                        0, &min_frequency));
    
    checkErr( Ivi_GetAttributeViReal64( vi,
                                        "",
                                        CICS_ATTR_FREQUENCY_MAXIMUM,
                                        0, &max_frequency));
    
    if (value > max_frequency || value < min_frequency)
        viCheckErr(IVI_ERROR_INVALID_VALUE);

Error:
    return error;
}

/*- Shared Write Callback ciCsViReal64_WriteCallback  -*/
/*- Shared Read Callback ciCsViReal64_ReadCallback  -*/

/*- CICS_ATTR_SLEW_CURRENT_RATE   -*/
/*- CICS_ATTR_TRIGGERED_SLEW_CURRENT_RATE -*/
static ViStatus _VI_FUNC ciCsAttrSlewCurrentRate_CheckCallback(
    ViSession       vi,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViReal64        value
)
{
    ViStatus    error = VI_SUCCESS;

    checkErr( ciCs_GetPhaseIndex(vi, phaseName, VI_NULL, VI_NULL));
    checkErr( ciCsAttrSlewFrequencyRate_CheckCallback( vi,
                                                       phaseName,
                                                       attributeId,
                                                       value));
Error:
    return error;
}


/*- CICS_ATTR_SLEW_FREQUENCY_RATE -*/
/*- CICS_ATTR_TRIGGERED_SLEW_FREQUENCY_RATE -*/
static ViStatus _VI_FUNC ciCsAttrSlewFrequencyRate_CheckCallback (ViSession vi,
                                                                   ViConstString channelName,
                                                                   ViAttr attributeId,
                                                                   ViReal64 value)
{
    ViStatus    error = VI_SUCCESS;
    ViInt32     result;
    
    viCheckErr( Ivi_CompareWithPrecision(0, value, CICS_VAL_INFINTY_SLEW_RATE, &result));
    if (result)
    {
        if (value < 0.0 || value > 9.9e37)
            viCheckErr( IVI_ERROR_INVALID_VALUE);
    }

Error:
    return error;
}

static ViStatus _VI_FUNC ciCsAttrSlewRate_ReadCallback (ViSession vi,
                                                         ViSession io,
                                                         ViConstString channelName,
                                                         ViAttr attributeId,
                                                         ViReal64 *value)
{
    ViStatus    error = VI_SUCCESS;
    ViChar      read_buffer[BUFFER_SIZE];
    ViInt32     read_buffer_size = sizeof(read_buffer);
    ViString    cmd;
	ViChar	    tempCmd[BUFFER_SIZE]; 
	
    viCheckErr( Ivi_GetStringFromTable (gs_ViReal64Commands, attributeId, &cmd)); 

    if (channelName != NULL)
        checkErr( ciCs_SetActivePhase(vi, channelName));

    if(busInterface == VI_INTF_ASRL){
    	viFlush (io, VI_ASRL_IN_BUF);  
    	Delay(serialDelay);
    }
    	
    if( busInterface == VI_INTF_TCPIP )
		viCheckErr( viQueryf(io, "%s?\n", "%#s", cmd, &read_buffer_size, read_buffer));
	else if(busInterface == VI_INTF_ASRL){
		strcpy(tempCmd, cmd);
		strcat(tempCmd, "?");
		viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
		strcpy(tempCmd, "");
		
		viCheckErr( viRead (io, read_buffer, BUFFER_SIZE-1, &read_buffer_size));
	}
	else
		viCheckErr( viQueryf(io, "%s?", "%#s", cmd, &read_buffer_size, read_buffer));
	
    read_buffer[read_buffer_size] = '\0';
    if (strstr(read_buffer, "INF"))
        *value = CICS_VAL_INFINTY_SLEW_RATE;
    else
        *value = atof(read_buffer);
Error:
    return error;
}

static ViStatus _VI_FUNC ciCsAttrSlewRate_WriteCallback (ViSession vi,
                                                          ViSession io,
                                                          ViConstString channelName,
                                                          ViAttr attributeId,
                                                          ViReal64 value)
{
    ViStatus    error = VI_SUCCESS;
    ViString    cmd;
	ViChar	    tempCmd[BUFFER_SIZE];
	ViChar		valueStr[100]; 
	
    viCheckErr( Ivi_GetStringFromTable (gs_ViReal64Commands, attributeId, &cmd)); 

    if (channelName != NULL)
        checkErr( ciCs_SetActivePhase(vi, channelName));

    if (value < 0.0)
        if( busInterface == VI_INTF_TCPIP ) 
			error = viPrintf(io, "%s INF\n", cmd);
		else if(busInterface == VI_INTF_ASRL){
			strcpy(tempCmd, cmd);
			strcat(tempCmd, " ");
			strcat(tempCmd, "INF");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, "");
		}
		else
			error = viPrintf(io, "%s INF", cmd);	
    else
        if( busInterface == VI_INTF_TCPIP )   
        	error = viPrintf(io, "%s %Le\n", cmd, value);
		else if(busInterface == VI_INTF_ASRL){
			Fmt(valueStr, "%s<%f", value); 
			strcpy(tempCmd, cmd);
			strcat(tempCmd, " ");
			strcat(tempCmd, valueStr);
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, "");
		}
		else
			error = viPrintf(io, "%s %Le", cmd, value);
    
    if(busInterface == VI_INTF_ASRL)
    		Delay(serialDelay);
    viCheckErr( error);

Error:
    return error;
}

/*- CICS_ATTR_PHASE_ANGLE -*/
/*- CICS_ATTR_TRIGGERED_PHASE_ANGLE -*/ 
/*- CICS_ATTR_TRIGGER_SYNCHRONIZATION_PHASE -*/
static IviRangeTableEntry attrPhaseAngeleRangeTableEntries[] =
{
    {-360.0,   360.0,  0,  NULL,  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrPhaseAngeleRangeTable =
{
    IVI_VAL_RANGED, 
    VI_TRUE, 
    VI_TRUE, 
    "", 
    attrPhaseAngeleRangeTableEntries, 
};
/*- Shared Write Callback ciCsViReal64_WriteCallback  -*/
/*- Shared Read Callback ciCsViReal64_ReadCallback  -*/

/*- CICS_ATTR_FUNCTION -*/
/*- CICS_ATTR_TRIGGERED_FUNCTION -*/
static ViStatus _VI_FUNC ciCsAttrFunction_CheckCallback(
    ViSession       vi,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViConstString   value
)
{
    return ciCs_CheckWaveformName(  vi,
                                    value);
}

static ViStatus _VI_FUNC ciCsAttrFunction_ReadCallback (
    ViSession           vi,
    ViSession           io, 
    ViConstString       channelName,
    ViAttr              attributeId, 
    const ViConstString cacheValue
)
{
    ViStatus        error = VI_SUCCESS;
    ViChar*         string = NULL;
    ViConstString   driver_function_name = cacheValue;
    ViInt32         readed_count;
	ViChar	    tempCmd[BUFFER_SIZE];  
	
    if(busInterface == VI_INTF_ASRL){
    	viFlush (io, VI_ASRL_IN_BUF);  
    	Delay(serialDelay);
    }
    	
    if( busInterface == VI_INTF_TCPIP ) 
	    viCheckErr( viPrintf(io, (attributeId == CICS_ATTR_FUNCTION) ?
	                                ":FUNC?\n" :
	                                ":FUNC:TRIG?\n"
	                         )
	              );
	else if(busInterface == VI_INTF_ASRL){
		if( attributeId == CICS_ATTR_FUNCTION ){
			strcpy(tempCmd, ":FUNC?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 	
		}
		else{
			strcpy(tempCmd, ":FUNC:TRIG?");
			viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
			strcpy(tempCmd, ""); 	
		}
	}
	else
		viCheckErr( viPrintf(io, (attributeId == CICS_ATTR_FUNCTION) ?
	                                ":FUNC?" :
	                                ":FUNC:TRIG?"
	                         )
	              );
	
    viCheckErr( ciCs_ReadDataUnknownLength ( io, &string, &readed_count));
    if (string)
        string[readed_count-1] = '\0';

    error = ciCs_GetStr1FromStr2 (gs_FunctionStringTable, &driver_function_name, string);
    if (error < 0)
        driver_function_name = string;

    checkErr( Ivi_SetValInStringCallback(vi, attributeId, driver_function_name));

Error:
    if (string)
        free(string);

    return error;
}
/*- Shared Write Callback ciCsViString_WriteCallback  -*/

/*- CICS_ATTR_FREQUENCY -*/
/*- CICS_ATTR_TRIGGERED_FREQUENCY -*/
static IviRangeTableEntry attrFrequencyRangeTableEntries[] =
{
    {45.0,   5000.0,  0,  NULL,  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrFrequencyRangeTable =
{
    IVI_VAL_RANGED, 
    VI_TRUE, 
    VI_TRUE, 
    "", 
    attrFrequencyRangeTableEntries, 
};
/*- Shared Write Callback ciCsViReal64_WriteCallback  -*/
/*- Shared Read Callback ciCsViReal64_ReadCallback  -*/

/*- CICS_ATTR_CLIPPING_LEVEL -*/
static IviRangeTableEntry attrClippingLevelRangeTableEntries[] =
{
    {0.0,   100.0,  0,  NULL,  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrClippingLevelRangeTable =
{
    IVI_VAL_RANGED, 
    VI_TRUE, 
    VI_TRUE, 
    "", 
    attrClippingLevelRangeTableEntries, 
};
/*- Shared Write Callback ciCsViReal64_WriteCallback  -*/
/*- Shared Read Callback ciCsViReal64_ReadCallback  -*/

/*- CICS_ATTR_TRIGGER_SOURCE -*/
static IviRangeTableEntry attrTriggerSourceRangeTableEntries[] =
{
    {CICS_VAL_TRIG_IMMEDIATE,   0,  0,  "IMM",  0}, 
    {CICS_VAL_TRIG_EXTERNAL,    0,  0,  "EXT",  0}, 
    {CICS_VAL_SOFTWARE_TRIG,    0,  0,  "BUS",  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrTriggerSourceRangeTable =
{
    IVI_VAL_DISCRETE, 
    VI_FALSE, 
    VI_FALSE, 
    "", 
    attrTriggerSourceRangeTableEntries, 
};
/*- Shared Write Callback ciCsEnum_WriteCallback  -*/
/*- Shared Read Callback ciCsEnum_ReadCallback  -*/

/*- CICS_ATTR_TRIGGER_DELAY -*/
/*- CICS_ATTR_TRIGGER_PULSE_WIDTH -*/
/*- CICS_ATTR_TRIGGER_PULSE_PERIOD -*/
static ViStatus _VI_FUNC  ciCsAttrTriggerDelay_CheckCallback(
    ViSession       vi,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViReal64        value
)
{
    ViStatus    error = VI_SUCCESS;
    ViInt32     number_of_phases = 1;
    ViReal64    max;

	ViChar      	modeState[BUFFER_SIZE] = "";    
	
    checkErr( Ivi_GetAttributeViString(vi, 
                                      VI_NULL,
                                      CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT,
                                      0,
									  BUFFER_SIZE,
									  modeState));
	
	number_of_phases = atoi(modeState);

    max = (number_of_phases == CICS_VAL_1_PHASE) ?
        4.30133e5 : /* 1 Phase */
        1.07533e6;  /* 3 phase */

    if (value < 0.0 || value > max)
        viCheckErr( IVI_ERROR_INVALID_VALUE);

    
Error:
    return error;
}
/*- Shared Write Callback ciCsViReal64_WriteCallback  -*/
/*- Shared Read Callback ciCsViReal64_ReadCallback  -*/

/*- CICS_ATTR_TRIGGERED_VOLTAGE_MODE -*/
/*- CICS_ATTR_TRIGGERED_FREQUENCY_MODE -*/
/*- CICS_ATTR_TRIGGERED_FUNCTION_MODE -*/
/*- CICS_ATTR_TRIGGERED_PHASE_ANGLE_MODE -*/
/*- CICS_ATTR_TRIGGERED_SLEW_VOLTAGE_RATE_MODE -*/
/*- CICS_ATTR_TRIGGERED_SLEW_FREQUENCY_RATE_MODE -*/
static IviRangeTableEntry attrTriggerModeRangeTableEntries[] =
{
    {CICS_VAL_TRIGGER_MODE_FIX,   0,  0,  "FIX",   0}, 
    {CICS_VAL_TRIGGER_MODE_STEP,  0,  0,  "STEP",  0}, 
    {CICS_VAL_TRIGGER_MODE_PULSE, 0,  0,  "PULS",  0}, 
    {CICS_VAL_TRIGGER_MODE_LIST,  0,  0,  "LIST",  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrTriggerModeRangeTable =
{
    IVI_VAL_DISCRETE, 
    VI_FALSE, 
    VI_FALSE, 
    "", 
    attrTriggerModeRangeTableEntries, 
};
/*- Shared Write Callback ciCsEnum_WriteCallback  -*/
/*- Shared Read Callback ciCsEnum_ReadCallback  -*/

/*- CICS_ATTR_TRIGGERED_VOLTAGE_LEVEL -*/
/*- Shared Write Callback ciCsViReal64_WriteCallback  -*/
/*- Shared Read Callback ciCsViReal64_ReadCallback  -*/
/*- Shared Check ciCsAttrVoltageLevel_CheckCallback -*/

/*- CICS_ATTR_TRIGGER_PULSE_COUNT -*/
/*- CICS_ATTR_TRIGGER_LIST_COUNT -*/
static IviRangeTableEntry attrTriggerPulseCountRangeTableEntries[] =
{
    {1,   LONG_MAX,  0,  NULL,  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrTriggerPulseCountRangeTable =
{
    IVI_VAL_RANGED, 
    VI_TRUE, 
    VI_TRUE, 
    "", 
    attrTriggerPulseCountRangeTableEntries, 
};
/*- Shared Write Callback ciCsViInt32_WriteCallback  -*/
/*- Shared Read Callback ciCsViInt32_ReadCallback  -*/

/*- CICS_ATTR_TRIGGER_LIST_MODE -*/
static IviRangeTableEntry attrTriggerListModeRangeTableEntries[] =
{
    {CICS_VAL_TRIGGER_LIST_STEP_ONCE,   0,  0,  "ONCE",   0}, 
    {CICS_VAL_TRIGGER_LIST_STEP_AUTO,   0,  0,  "AUTO",  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrTriggerListModeRangeTable =
{
    IVI_VAL_DISCRETE, 
    VI_FALSE, 
    VI_FALSE, 
    "", 
    attrTriggerListModeRangeTableEntries, 
};
/*- Shared Write Callback ciCsEnum_WriteCallback  -*/
/*- Shared Read Callback ciCsEnum_ReadCallback  -*/

/*- CICS_ATTR_ACQUISITION_TRIGGER_SOURCE -*/
static IviRangeTableEntry attrAcquisitionTriggerSourceRangeTableEntries[] =
{
    {CICS_VAL_TRIG_EXTERNAL,    0,  0,  "EXT",  0}, 
    {CICS_VAL_SOFTWARE_TRIG,    0,  0,  "BUS",  0}, 
    {CICS_VAL_TRIG_TTLT,        0,  0,  "TTLT", 0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrAcquisitionTriggerSourceRangeTable =
{
    IVI_VAL_DISCRETE, 
    VI_FALSE, 
    VI_FALSE, 
    "", 
    attrAcquisitionTriggerSourceRangeTableEntries, 
};
/*- Shared Write Callback ciCsEnum_WriteCallback  -*/
/*- Shared Read Callback ciCsEnum_ReadCallback  -*/

/*- CICS_ATTR_ACQUISITION_TIME_INTERVAL -*/
static IviRangeTableEntry attrAcquisitionStartTimeRangeTableEntries[] =
{
    {25.037e-6,   250.037e-6,  0,  NULL,  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrAcquisitionStartTimeRangeTable =
{
    IVI_VAL_RANGED, 
    VI_TRUE, 
    VI_TRUE, 
    "", 
    attrAcquisitionStartTimeRangeTableEntries, 
};
/*
 * TODO: 020816 Add corece callback
 * Value is coerced with step 25.0e-6
 */
/*- Shared Write Callback ciCsViReal64_WriteCallback  -*/
/*- Shared Read Callback ciCsViReal64_ReadCallback  -*/

/*- CICS_ATTR_ACQUISITION_START_TIME -*/
static ViStatus _VI_FUNC  ciCsAttrAcquisitionStartTime_CheckCallback(
    ViSession       vi,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViReal64        value
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    time_interval;
    ViReal64    points;

    checkErr( Ivi_GetAttributeViReal64(vi, 
                                       VI_NULL,
                                       CICS_ATTR_ACQUISITION_TIME_INTERVAL,
                                       0, &time_interval));
    
    points = value / time_interval;
    if (points < 0)
        points =  ceil(points);
    else
        points =  floor(points);

    if (points < -4095 || points > 2E9)
        viCheckErr( IVI_ERROR_INVALID_VALUE);

Error:
    return  error;
}

static ViStatus _VI_FUNC  ciCsAttrAcquisitionStartTime_CoerceCallback(
    ViSession       vi,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViReal64        value,
    ViReal64*       coercedValue
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    time_interval;
    ViReal64    points;
    ViReal64    new_value = value;

    checkErr( Ivi_GetAttributeViReal64(vi, 
                                       VI_NULL,
                                       CICS_ATTR_ACQUISITION_TIME_INTERVAL,
                                       0, &time_interval));
    points = value / time_interval;
    if (points < 0)
        points =  ceil(points);
    else
        points =  floor(points);
    new_value = points * time_interval;
    
Error:
    *coercedValue = new_value;  
    return  error;
}

static ViStatus _VI_FUNC  ciCsAttrAcquisitionStartTime_ReadCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViReal64*       value
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    time_interval;
    ViInt32     points;
	ViChar	    tempCmd[BUFFER_SIZE];  
	ViChar      buffer[BUFFER_SIZE];
	ViUInt32  	read_count;
	
    checkErr( Ivi_GetAttributeViReal64(vi, 
                                       VI_NULL,
                                       CICS_ATTR_ACQUISITION_TIME_INTERVAL,
                                       0, &time_interval));
    
    if(busInterface == VI_INTF_ASRL){
    	viFlush (io, VI_ASRL_IN_BUF);  
    	Delay(serialDelay);
    }
    	
    if( busInterface == VI_INTF_TCPIP )
    	viCheckErr( viQueryf(io, ":SENS:SWE:OFFS:POIN?\n","%ld", &points));
	else if(busInterface == VI_INTF_ASRL){
		strcpy(tempCmd, ":SENS:SWE:OFFS:POIN?");
		viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
		strcpy(tempCmd, ""); 
		
		viCheckErr( viRead (io, buffer, BUFFER_SIZE-1, &read_count));
		buffer[read_count] = '\0';
		
		points = atoi(buffer);
	}
	else
		viCheckErr( viQueryf(io, ":SENS:SWE:OFFS:POIN?","%ld", &points));
	
    *value = points * time_interval;
Error:
    return  error;
}
static ViStatus _VI_FUNC  ciCsAttrAcquisitionStartTime_WriteCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViReal64        value
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    time_interval,
                points;
	ViChar	    tempCmd[BUFFER_SIZE], valueStr[100];   
	
    checkErr( Ivi_GetAttributeViReal64(vi, 
                                       VI_NULL,
                                       CICS_ATTR_ACQUISITION_TIME_INTERVAL,
                                       0, &time_interval));
    points = value / time_interval;
    if (points < 0)
        points =  ceil(points);
    else
        points =  floor(points);

    if( busInterface == VI_INTF_TCPIP )
		viCheckErr( viPrintf(io, ":SENS:SWE:OFFS:POIN %ld\n", (ViInt32)points));
	else if(busInterface == VI_INTF_ASRL){
		Fmt(valueStr, "%s<%i", points); 
		strcpy(tempCmd, ":SENS:SWE:OFFS:POIN");
		strcat(tempCmd, " ");
		strcat(tempCmd, valueStr);
		viCheckErr( viWrite (io, tempCmd, strlen(tempCmd), VI_NULL));
		strcpy(tempCmd, ""); 
	}
	else
		viCheckErr( viPrintf(io, ":SENS:SWE:OFFS:POIN %ld", (ViInt32)points));	
	
	if(busInterface == VI_INTF_ASRL)
    	Delay(serialDelay);
Error:
    return  error;
}

/*- CICS_ATTR_WINDOW_TYPE -*/
static IviRangeTableEntry attrWindowTypeRangeTableEntries[] =
{
    {CICS_VAL_WINDOW_TYPE_RECTANGULAR,  0,  0,  "RECT",  0}, 
    {CICS_VAL_WINDOW_TYPE_KBESSEL,      0,  0,  "KBES",  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrWindowTypeRangeTable =
{
    IVI_VAL_DISCRETE, 
    VI_FALSE, 
    VI_FALSE, 
    "", 
    attrWindowTypeRangeTableEntries, 
};
/*- Shared Write Callback ciCsEnum_WriteCallback  -*/
/*- Shared Read Callback ciCsEnum_ReadCallback  -*/

/*- CICS_ATTR_INSTR_VOLTAGE_RANGE -*/   /*- CICS_ATTR_VOLTAGE_RANGE_MAXIMUM -*/
static IviRangeTableEntry attrInstrVoltageRangeRangeTableEntries[] =
{
    {0.0,   135.0,  135.0,  "",  0}, 
    {135.0, 270.0,  270.0,  "",  0},
    {0.0,   150.0,  150.0,  "",  0}, 
    {150.0, 300.0,  300.0,  "",  0},
    {0.0,   156.0,  156.0,  "",  0}, 
    {156.0, 312.0,  312.0,  "",  0},
    {0.0,   200.0,  200.0,  "",  0}, 
    {200.0, 400.0,  400.0,  "",  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrInstrVoltageRangeRangeTable =
{
    IVI_VAL_COERCED, 
    VI_FALSE, 
    VI_FALSE, 
    "", 
    attrInstrVoltageRangeRangeTableEntries, 
};
/*- Shared Write Callback ciCsEnum_WriteCallback  -*/
/*- Shared Read Callback ciCsEnum_ReadCallback  -*/

/*- CICS_ATTR_INSTR_VOLTAGE_RANGE -*/
static ViStatus _VI_FUNC  ciCsAttrInstrVoltageRange_ReadCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   pahseName,
    ViAttr          attributeId,
    ViReal64*       value
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    max;
    
	checkErr( ciCs_GetMaximumRange ( vi,
                                     CICS_VAL_RANGE_VOLTAGE,
                                     &max));                                                  

Error:
	*value = max;
    return error;
}

static ViStatus _VI_FUNC  ciCsAttrInstrVoltageRange_WriteCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   phaseName,
    ViAttr          attributeId,
    ViReal64        value
)
{
    ViStatus    error = VI_SUCCESS;
    ViReal64    voltage_range;
    
	checkErr( ciCs_GetMaximumRange ( vi,
                                     CICS_VAL_RANGE_VOLTAGE,
                                     &voltage_range)); 
                                     
    checkErr( Ivi_SetAttributeViReal64(vi,
                                       phaseName,
                                       CICS_ATTR_INSTR_VOLTAGE_RANGE,
                                       0, voltage_range));
Error:
    return error;
}

/*- CICS_ATTR_OUTPUT_TRIGGER_SOURCE -*/
static IviRangeTableEntry attrOutputTriggerSourceRangeTableEntries[] =
{
    {CICS_VAL_OUTPUT_TRIGGER_SOURCE_BOT,    0,  0,  "BOT",  0}, 
    {CICS_VAL_OUTPUT_TRIGGER_SOURCE_EOT,    0,  0,  "EOT",  0}, 
    {CICS_VAL_OUTPUT_TRIGGER_SOURCE_LIST,   0,  0,  "LIST", 0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrOutputTriggerSourceRangeTable =
{
    IVI_VAL_DISCRETE, 
    VI_FALSE, 
    VI_FALSE, 
    "", 
    attrOutputTriggerSourceRangeTableEntries, 
};
/*- Shared Write Callback ciCsEnum_WriteCallback  -*/
/*- Shared Read Callback ciCsEnum_ReadCallback  -*/


/*- CICS_ATTR_TRIGGER_SYNCHRONIZATION_SOURCE -*/
static IviRangeTableEntry attrTriggerSynchronizationSourceRangeTableEntries[] =
{
    {CICS_VAL_SYNCHRONIZATION_SOURCE_IMMEDIATE, 0,  0,  "IMM",  0}, 
    {CICS_VAL_SYNCHRONIZATION_SOURCE_PHASE,     0,  0,  "PHAS", 0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrTriggerSynchronizationSourceRangeTable =
{
    IVI_VAL_DISCRETE, 
    VI_FALSE, 
    VI_FALSE, 
    "", 
    attrTriggerSynchronizationSourceRangeTableEntries, 
};
/*- Shared Write Callback ciCsEnum_WriteCallback  -*/
/*- Shared Read Callback ciCsEnum_ReadCallback  -*/

/*- CICS_ATTR_OUTPUT_PROTECTION_DELAY -*/
static IviRangeTableEntry attrOutputProtectionDelayEntries[] =
{
    {0.0,   100.0,  0,  NULL,  0}, 
    {IVI_RANGE_TABLE_LAST_ENTRY}
};

static IviRangeTable attrOutputProtectionDelay =
{
    IVI_VAL_RANGED, 
    VI_TRUE, 
    VI_TRUE, 
    "", 
    attrOutputProtectionDelayEntries, 
};
/*- Shared Write Callback ciCsViReal64_WriteCallback  -*/
/*- Shared Read Callback ciCsViReal64_ReadCallback  -*/

/*- Shared Write Callback ciCsViString_WriteCallback  -*/

/*- CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT (hidden) -*/
static ViStatus _VI_FUNC ciCsAttrInstrOutputPhaseCount_WriteCallback(
    ViSession       vi,
    ViSession       io,
    ViConstString   channelName,
    ViAttr          attributeId,
    ViInt32         value
)
{
    ViStatus    error = VI_SUCCESS;

    checkErr( ciCsViInt32_WriteCallback( vi, io, channelName, attributeId, value));

    Delay(SWITCH_PHASE_DELAY_MILISECONDS / 1000.0);

Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_InitAttributes                                                  
 * Purpose:  This function adds attributes to the IVI session, initializes   
 *           instrument attributes, and sets attribute invalidation          
 *           dependencies.                                                   
 *****************************************************************************/
static ViStatus ciCs_InitAttributes (
    ViSession   vi,
    ViInt32     modelType
)
{
    ViStatus    error = VI_SUCCESS;
    ViInt32     flags = 0;
    ViChar      id_buffer[256];

	ViInt32     actualModel;
	
	// to seperate Lx and Ls
	if( modelType == 3001 ){
		actualModel = 3000;
		sprintf(id_buffer, "CALIFORNIA INSTRUMENTS,%ldLs,12345,Rev1.00", actualModel);
	}
	else if( modelType == 4501){
		actualModel = 4500;
		sprintf(id_buffer, "CALIFORNIA INSTRUMENTS,%ldLs,12345,Rev1.00", actualModel);
	}
	else if( modelType == 6001){
		actualModel = 6000;
		sprintf(id_buffer, "CALIFORNIA INSTRUMENTS,%ldLs,12345,Rev1.00", actualModel);
	}
	else if( modelType == 9001){
		actualModel = 9000;
		sprintf(id_buffer, "CALIFORNIA INSTRUMENTS,%ldLs,12345,Rev1.00", actualModel);
	}
	else if( modelType == 12001){
		actualModel = 12000;
		sprintf(id_buffer, "CALIFORNIA INSTRUMENTS,%ldLs,12345,Rev1.00", actualModel);
	}
	else if( modelType == 13501){
		actualModel = 13500;
		sprintf(id_buffer, "CALIFORNIA INSTRUMENTS,%ldLs,12345,Rev1.00", actualModel);
	}
	else if( modelType == 18001){
		actualModel = 18000;
		sprintf(id_buffer, "CALIFORNIA INSTRUMENTS,%ldLs,12345,Rev1.00", actualModel);
	}
	else{
    	sprintf(id_buffer, "CALIFORNIA INSTRUMENTS,%ldLx,12345,Rev1.00", modelType);
    }

        /*- Initialize instrument attributes --------------------------------*/            

    checkErr( Ivi_SetAttributeViInt32 (vi, "", CICS_ATTR_SPECIFIC_DRIVER_MAJOR_VERSION,
                                       0, CICS_MAJOR_VERSION));
    checkErr( Ivi_SetAttributeViInt32 (vi, "", CICS_ATTR_SPECIFIC_DRIVER_MINOR_VERSION,
                                       0, CICS_MINOR_VERSION));
    checkErr( Ivi_SetAttrReadCallbackViString (vi, CICS_ATTR_SPECIFIC_DRIVER_REVISION,
                                               ciCsAttrDriverRevision_ReadCallback));
    checkErr( Ivi_SetAttributeViInt32 (vi, "", CICS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MAJOR_VERSION,
                                       0, CICS_CLASS_SPEC_MAJOR_VERSION));
    checkErr( Ivi_SetAttributeViInt32 (vi, "", CICS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MINOR_VERSION,
                                       0, CICS_CLASS_SPEC_MINOR_VERSION));
    checkErr( Ivi_SetAttributeViString (vi, "", CICS_ATTR_IO_SESSION_TYPE,
                                        0, CICS_IO_SESSION_TYPE));
    checkErr( Ivi_SetAttributeViString (vi, "", CICS_ATTR_SUPPORTED_INSTRUMENT_MODELS,
                                        0, CICS_SUPPORTED_INSTRUMENT_MODELS));


    checkErr( Ivi_GetAttributeFlags (vi, CICS_ATTR_INSTRUMENT_FIRMWARE_REVISION, &flags));
    checkErr( Ivi_SetAttributeFlags (vi, CICS_ATTR_INSTRUMENT_FIRMWARE_REVISION, 
                                     flags | IVI_VAL_USE_CALLBACKS_FOR_SIMULATION));
    checkErr( Ivi_SetAttrReadCallbackViString (vi, CICS_ATTR_INSTRUMENT_FIRMWARE_REVISION,
                                               ciCsAttrFirmwareRevision_ReadCallback));

    checkErr( Ivi_GetAttributeFlags (vi, CICS_ATTR_INSTRUMENT_MANUFACTURER, &flags));
    checkErr( Ivi_SetAttributeFlags (vi, CICS_ATTR_INSTRUMENT_MANUFACTURER, 
                                     flags | IVI_VAL_USE_CALLBACKS_FOR_SIMULATION));
    checkErr( Ivi_SetAttrReadCallbackViString (vi, CICS_ATTR_INSTRUMENT_MANUFACTURER,
                                               ciCsAttrInstrumentManufacturer_ReadCallback));

    checkErr( Ivi_GetAttributeFlags (vi, CICS_ATTR_INSTRUMENT_MODEL, &flags));
    checkErr( Ivi_SetAttributeFlags (vi, CICS_ATTR_INSTRUMENT_MODEL, 
                                     flags | IVI_VAL_USE_CALLBACKS_FOR_SIMULATION));
    checkErr( Ivi_SetAttrReadCallbackViString (vi, CICS_ATTR_INSTRUMENT_MODEL,
                                               ciCsAttrInstrumentModel_ReadCallback));

    checkErr( Ivi_SetAttributeViString (vi, "", CICS_ATTR_SPECIFIC_DRIVER_VENDOR,
                                        0, CICS_DRIVER_VENDOR));
    checkErr( Ivi_SetAttributeViString (vi, "", CICS_ATTR_SPECIFIC_DRIVER_DESCRIPTION,
                                        0, CICS_DRIVER_DESCRIPTION));
    checkErr( Ivi_SetAttributeViAddr (vi, VI_NULL, IVI_ATTR_OPC_CALLBACK, 0,
                                      ciCs_WaitForOPCCallback));
    checkErr( Ivi_SetAttributeViAddr (vi, VI_NULL, IVI_ATTR_CHECK_STATUS_CALLBACK, 0,
                                      ciCs_CheckStatusCallback));
    checkErr( Ivi_SetAttributeViBoolean (vi, VI_NULL, IVI_ATTR_SUPPORTS_WR_BUF_OPER_MODE, 
                                         0, VI_TRUE));

    checkErr( Ivi_SetAttributeViString (vi, VI_NULL, CICS_ATTR_GROUP_CAPABILITIES, 0,
                                        "IviACPwrArbWaveform,"
                                        "IviACPwrTrigger,"
                                        "IviACPwrPulseTrigger,"
                                        "IviACPwrListTrigger,"
                                        "IviACPwrSynchronization,"
                                        "IviACPwrAcquisition,"
                                        "IviACPwrSoftwareTrigger,"
                                        "IviACPwrCalibration"));
    
    /*- Add instrument-specific attributes ------------------------------*/            
    /*- CICS_ATTR_OUTPUT_PHASE_COUNT -*/
    /*
     * WARNING: Do not set NEVER_CACHE for this attribute,
     *          It invalidates all attributes in write callback.
     */
    checkErr( Ivi_AddAttributeViInt32 (vi, CICS_ATTR_OUTPUT_PHASE_COUNT,
                                       "CICS_ATTR_OUTPUT_PHASE_COUNT", 
                                       CICS_VAL_1_PHASE,
                                       IVI_VAL_USE_CALLBACKS_FOR_SIMULATION | IVI_VAL_DONT_CHECK_STATUS,
                                       ciCsAttrOutputPhaseCount_ReadCallback,
                                       ciCsAttrOutputPhaseCount_WriteCallback,
                                       &attrOutputPhaseCountRangeTable));
    
    /*- CICS_ATTR_OUTPUT_ENABLED -*/
    checkErr( Ivi_AddAttributeViBoolean (vi, CICS_ATTR_OUTPUT_ENABLED,
                                         "CICS_ATTR_OUTPUT_ENABLED", 
                                         VI_FALSE,
                                         0,
                                         ciCsViBoolean_ReadCallback,
                                         ciCsViBoolean_WriteCallback));
                                         
	/*- CICS_ATTR_OUTPUT_ALC_STATE -*/
    checkErr( Ivi_AddAttributeViInt32 (vi, CICS_ATTR_OUTPUT_ALC_STATE,
                                       "CICS_ATTR_OUTPUT_ALC_STATE", 
                                       0,
                                       0,
                                       ciCsEnum_ReadCallback,
                                       ciCsEnum_WriteCallback,
                                       &attrOutputALCStateRangeTable));
                                       
    /*- CICS_ATTR_OUTPUT_PHASE_MODE -*/                                     
	checkErr( Ivi_AddAttributeViInt32 (vi, CICS_ATTR_OUTPUT_PHASE_MODE,
                                       "CICS_ATTR_OUTPUT_PHASE_MODE", 
                                       CICS_VAL_3_PHASE,
                                       0,
                                       ciCsEnum_ReadCallback,
                                       ciCsEnum_WriteCallback,
                                       &attrOutputPhaseCountRangeTable));                                         
                                         
    /*- CICS_ATTR_OVP_ENABLED -*/
#if defined(OVERVOLTAGE_PROTECTION_STATE_SUPPORTED)     
    checkErr( Ivi_AddRepeatedAttributeViBoolean (vi,
                                                 PHASE_REP_CAP_NAME,
                                                 CICS_ATTR_OVP_ENABLED,
                                                 "CICS_ATTR_OVP_ENABLED",
                                                 VI_FALSE,
                                                 IVI_VAL_USE_CALLBACKS_FOR_SIMULATION | IVI_VAL_DONT_CHECK_STATUS | IVI_VAL_NEVER_CACHE,
                                                 ciCsAttrOvpEnabled_ReadCallback,
                                                 ciCsAttrOvpEnabled_WriteCallback));
    checkErr( Ivi_AddRepeatedAttributeViBoolean (vi,
                                                 PHASE_REP_CAP_NAME,
                                                 CICS_ATTR_INSTR_OVP_ENABLED,
                                                 "CICS_ATTR_INSTR_OVP_ENABLED",
                                                 VI_FALSE,
                                                 0,
                                                 ciCsAttrOvpEnabled_ReadCallback,
                                                 ciCsAttrOvpEnabled_WriteCallback));
#else
        checkErr( Ivi_AddRepeatedAttributeViBoolean (vi,
                                                 PHASE_REP_CAP_NAME,
                                                 CICS_ATTR_OVP_ENABLED,
                                                 "CICS_ATTR_OVP_ENABLED",
                                                 VI_TRUE,
                                                 0,
                                                 NULL,
                                                 NULL));
#endif
    /*- CICS_ATTR_OVP_LIMIT -*/
    checkErr( Ivi_AddRepeatedAttributeViReal64 (vi,
                                                 PHASE_REP_CAP_NAME,
                                                 CICS_ATTR_OVP_LIMIT,
                                                 "CICS_ATTR_OVP_LIMIT",
                                                 500.0,
                                                 0,
                                                 ciCsViReal64_ReadCallback,
                                                 ciCsViReal64_WriteCallback,
                                                 &attrOvpLimitRangeTable, 0));
    checkErr( Ivi_SetAttrCoerceCallbackViReal64(vi,
                                                CICS_ATTR_OVP_LIMIT,
                                                ciCsAttrOVPLimit_CoerceCallback));

    /*- Shared Check Callback ciCsPhaseNameViReal64_CheckCallback 
     *- for models Ci3000il and Ci4500iL ---------------------------*/

    /*- CICS_ATTR_VOLTAGE_LIMIT_BEHAVIOR -*/
    checkErr( Ivi_AddRepeatedAttributeViInt32 (vi,
                                               PHASE_REP_CAP_NAME,
                                               CICS_ATTR_VOLTAGE_LIMIT_BEHAVIOR,
                                               "CICS_ATTR_VOLTAGE_LIMIT_BEHAVIOR",
                                               CICS_VAL_VOLTAGE_REGULATE,
                                               IVI_VAL_USE_CALLBACKS_FOR_SIMULATION | IVI_VAL_DONT_CHECK_STATUS,
                                               ciCsAttrVoltageLimitBehavior_ReadCallback,
                                               ciCsAttrVoltageLimitBehavior_WriteCallback,
                                               &attrVoltageLimitBehaviorRangeTable));
    
    /*- CICS_ATTR_VOLTAGE_LIMIT -*/
    checkErr( Ivi_AddRepeatedAttributeViReal64 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_VOLTAGE_LIMIT,
                                                "CICS_ATTR_VOLTAGE_LIMIT",
                                                1.0,
                                                0,
                                                ciCsViReal64_ReadCallback,
                                                ciCsViReal64_WriteCallback,
                                                VI_NULL, 0));
    
	checkErr( Ivi_SetAttrCheckCallbackViReal64( vi,
                                                CICS_ATTR_VOLTAGE_LIMIT,
                                                ciCsAttrVoltageLimit_CheckCallback));

    /*- CICS_ATTR_CURRENT_LEVEL -*/
    checkErr( Ivi_AddRepeatedAttributeViReal64 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_CURRENT_LEVEL,
                                                "CICS_ATTR_CURRENT_LEVEL",
                                                1.0,
                                                IVI_VAL_NEVER_CACHE,
                                                ciCsViReal64_ReadCallback,
                                                ciCsViReal64_WriteCallback,
                                                VI_NULL, 0));
    checkErr( Ivi_SetAttrCheckCallbackViReal64( vi,
                                                CICS_ATTR_CURRENT_LEVEL,
                                                ciCsAttrCurrentLevel_CheckCallback));

    /*- CICS_ATTR_SLEW_CURRENT_RATE -*/
    checkErr( Ivi_AddRepeatedAttributeViReal64 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_SLEW_CURRENT_RATE,
                                                "CICS_ATTR_SLEW_CURRENT_RATE",
                                                CICS_VAL_INFINTY_SLEW_RATE,
                                                IVI_VAL_NEVER_CACHE,
                                                ciCsViReal64_ReadCallback,
                                                ciCsViReal64_WriteCallback,
                                                VI_NULL, 0));
    checkErr( Ivi_SetAttrCheckCallbackViReal64(vi,
                                               CICS_ATTR_SLEW_CURRENT_RATE,
                                               ciCsAttrSlewCurrentRate_CheckCallback));

    /*- CICS_ATTR_SLEW_FREQUENCY_RATE -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                        CICS_ATTR_SLEW_FREQUENCY_RATE,
                                        "CICS_ATTR_SLEW_FREQUENCY_RATE",
                                        CICS_VAL_INFINTY_SLEW_RATE,
                                        IVI_VAL_NEVER_CACHE,
                                        ciCsAttrSlewRate_ReadCallback,
                                        ciCsAttrSlewRate_WriteCallback,
                                        VI_NULL, 0));
    checkErr( Ivi_SetAttrCheckCallbackViReal64(vi,
                                               CICS_ATTR_SLEW_FREQUENCY_RATE,
                                               ciCsAttrSlewFrequencyRate_CheckCallback));
    /*- CICS_ATTR_PHASE_ANGLE -*/
    checkErr( Ivi_AddRepeatedAttributeViReal64 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_PHASE_ANGLE,
                                                "CICS_ATTR_PHASE_ANGLE",
                                                0.0,
                                                IVI_VAL_NEVER_CACHE,
                                                ciCsViReal64_ReadCallback,
                                                ciCsViReal64_WriteCallback,
                                                &attrPhaseAngeleRangeTable, 0));
    
	/*- CICS_ATTR_FUNCTION -*/
    checkErr( Ivi_AddAttributeViString (vi,
                                        CICS_ATTR_FUNCTION,
                                        "CICS_ATTR_FUNCTION",
                                        "SIN",
                                        IVI_VAL_NEVER_CACHE,
                                        ciCsAttrFunction_ReadCallback,
                                        ciCsViString_WriteCallback));
	
	/*- CICS_ATTR_OUTPUT_MODE -*/
  /*  checkErr( Ivi_AddAttributeViString (vi,
                                        CICS_ATTR_OUTPUT_MODE,
                                        "CICS_ATTR_OUTPUT_MODE",
                                        "AC",
                                        IVI_VAL_NEVER_CACHE,
                                        ciCsAttrOutputMode_ReadCallback,
                                        ciCsViString_WriteCallback));*/
	
    checkErr( Ivi_SetAttrCheckCallbackViString(vi,
                                               CICS_ATTR_FUNCTION,
                                               ciCsAttrFunction_CheckCallback));

    /*- CICS_ATTR_FREQUENCY -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                        CICS_ATTR_FREQUENCY,
                                        "CICS_ATTR_FREQUENCY",
                                        60.0,
                                        IVI_VAL_NEVER_CACHE,
                                        ciCsViReal64_ReadCallback,
                                        ciCsViReal64_WriteCallback,
                                        &attrFrequencyRangeTable, 0));
	checkErr( Ivi_SetAttrCheckCallbackViReal64( vi,
                                                CICS_ATTR_FREQUENCY,
                                                ciCsAttrFrequencyValue_CheckCallback));

    /*- CICS_ATTR_INSTR_USER_TABLE_LIST -*/
    checkErr( Ivi_AddAttributeViString (vi,
                                        CICS_ATTR_INSTR_USER_TABLE_LIST,
                                        "CICS_ATTR_INSTR_USER_TABLE_LIST",
                                        "SQU,SIN,CSIN",
                                        0,
                                        ciCsViString_ReadCallback,
                                        VI_NULL));
	
	/*- CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT -*/
    checkErr( Ivi_AddAttributeViString (vi,
                                        CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT,
                                        "CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT",
                                        "3",
                                        0,
                                        ciCsViString_ReadCallback,
                                        VI_NULL)); 
	
    /*- CICS_ATTR_CLIPPING_LEVEL -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                        CICS_ATTR_CLIPPING_LEVEL,
                                        "CICS_ATTR_CLIPPING_LEVEL",
                                        100.0,
                                        0,
                                        ciCsViReal64_ReadCallback,
                                        ciCsViReal64_WriteCallback,
                                        &attrClippingLevelRangeTable, 0));

    /*- CICS_ATTR_MAX_NUM_WAVEFORMS -*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_MAX_NUM_WAVEFORMS,
                                       "CICS_ATTR_MAX_NUM_WAVEFORMS",
                                       12,
                                       IVI_VAL_NOT_USER_WRITABLE,
                                       VI_NULL, VI_NULL, VI_NULL));
    /*- CICS_ATTR_WAVEFORM_QUANTUM -*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_WAVEFORM_QUANTUM,
                                       "CICS_ATTR_WAVEFORM_QUANTUM",
                                       1024,
                                       IVI_VAL_NOT_USER_WRITABLE,
                                       VI_NULL, VI_NULL, VI_NULL));
    /*- CICS_ATTR_MIN_WAVEFORM_SIZE -*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_MIN_WAVEFORM_SIZE,
                                       "CICS_ATTR_MIN_WAVEFORM_SIZE",
                                       1024,
                                       IVI_VAL_NOT_USER_WRITABLE,
                                       VI_NULL, VI_NULL, VI_NULL));
    /*- CICS_ATTR_MAX_WAVEFORM_SIZE -*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_MAX_WAVEFORM_SIZE,
                                       "CICS_ATTR_MAX_WAVEFORM_SIZE",
                                       1024,
                                       IVI_VAL_NOT_USER_WRITABLE,
                                       VI_NULL, VI_NULL, VI_NULL));
    
    /*- CICS_ATTR_TRIGGER_SOURCE -*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_TRIGGER_SOURCE,
                                       "CICS_ATTR_TRIGGER_SOURCE",
                                       CICS_VAL_SOFTWARE_TRIG,
                                       0,
                                       ciCsEnum_ReadCallback,
                                       ciCsEnum_WriteCallback,
                                       &attrTriggerSourceRangeTable));

    /*- CICS_ATTR_TRIGGER_DELAY -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                       CICS_ATTR_TRIGGER_DELAY,
                                       "CICS_ATTR_TRIGGER_DELAY",
                                       0.0,
                                       0,
                                       ciCsViReal64_ReadCallback,
                                       ciCsViReal64_WriteCallback,
                                       VI_NULL, 0));
    checkErr( Ivi_SetAttrCheckCallbackViReal64(vi,
                                               CICS_ATTR_TRIGGER_DELAY,
                                               ciCsAttrTriggerDelay_CheckCallback));
    
    /*- CICS_ATTR_TRIGGERED_CURRENT_MODE -*/
    checkErr( Ivi_AddRepeatedAttributeViInt32 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_TRIGGERED_CURRENT_MODE,
                                                "CICS_ATTR_TRIGGERED_CURRENT_MODE",
                                                CICS_VAL_TRIGGER_MODE_FIX,
                                                0,
                                                ciCsEnum_ReadCallback,
                                                ciCsEnum_WriteCallback,
                                                &attrTriggerModeRangeTable));
	
    /*- CICS_ATTR_TRIGGERED_FREQUENCY_MODE -*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_TRIGGERED_FREQUENCY_MODE,
                                       "CICS_ATTR_TRIGGERED_FREQUENCY_MODE",
                                       CICS_VAL_TRIGGER_MODE_FIX,
                                       0,
                                       ciCsEnum_ReadCallback,
                                       ciCsEnum_WriteCallback,
                                       &attrTriggerModeRangeTable));
    
    /*- CICS_ATTR_TRIGGERED_FUNCTION_MODE -*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_TRIGGERED_FUNCTION_MODE,
                                       "CICS_ATTR_TRIGGERED_FUNCTION_MODE",
                                       CICS_VAL_TRIGGER_MODE_FIX,
                                       0,
                                       ciCsEnum_ReadCallback,
                                       ciCsEnum_WriteCallback,
                                       &attrTriggerModeRangeTable));
    
    /*- CICS_ATTR_TRIGGERED_PHASE_ANGLE_MODE -*/
    checkErr( Ivi_AddRepeatedAttributeViInt32 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_TRIGGERED_PHASE_ANGLE_MODE,
                                                "CICS_ATTR_TRIGGERED_PHASE_ANGLE_MODE",
                                                CICS_VAL_TRIGGER_MODE_FIX,
                                                0,
                                                ciCsEnum_ReadCallback,
                                                ciCsEnum_WriteCallback,
                                                &attrTriggerModeRangeTable));

    /*- CICS_ATTR_TRIGGERED_SLEW_CURRENT_RATE_MODE -*/
    checkErr( Ivi_AddRepeatedAttributeViInt32 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_TRIGGERED_SLEW_CURRENT_RATE_MODE,
                                                "CICS_ATTR_TRIGGERED_SLEW_CURRENT_RATE_MODE",
                                                CICS_VAL_TRIGGER_MODE_FIX,
                                                0,
                                                ciCsEnum_ReadCallback,
                                                ciCsEnum_WriteCallback,
                                                &attrTriggerModeRangeTable));
    /*- CICS_ATTR_TRIGGERED_SLEW_FREQUENCY_RATE_MODE-*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_TRIGGERED_SLEW_FREQUENCY_RATE_MODE,
                                       "CICS_ATTR_TRIGGERED_SLEW_FREQUENCY_RATE_MODE",
                                       CICS_VAL_TRIGGER_MODE_FIX,
                                       0,
                                       ciCsEnum_ReadCallback,
                                       ciCsEnum_WriteCallback,
                                       &attrTriggerModeRangeTable));

    /*- CICS_ATTR_TRIGGERED_CURRENT_LEVEL -*/
    checkErr( Ivi_AddRepeatedAttributeViReal64 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_TRIGGERED_CURRENT_LEVEL,
                                                "CICS_ATTR_TRIGGERED_CURRENT_LEVEL",
                                                0.0,
                                                0,
                                                ciCsViReal64_ReadCallback,
                                                ciCsViReal64_WriteCallback,
                                                VI_NULL, 0));
    checkErr( Ivi_SetAttrCheckCallbackViReal64( vi,
                                                CICS_ATTR_TRIGGERED_CURRENT_LEVEL,
                                                ciCsAttrCurrentLevel_CheckCallback));

    /*- CICS_ATTR_TRIGGERED_FREQUENCY -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                        CICS_ATTR_TRIGGERED_FREQUENCY,
                                        "CICS_ATTR_TRIGGERED_FREQUENCY",
                                        60.0,
                                        0,
                                        ciCsViReal64_ReadCallback,
                                        ciCsViReal64_WriteCallback,
                                        &attrFrequencyRangeTable, 0));

    /*- CICS_ATTR_TRIGGERED_FUNCTION -*/
    checkErr( Ivi_AddAttributeViString (vi,
                                        CICS_ATTR_TRIGGERED_FUNCTION,
                                        "CICS_ATTR_TRIGGERED_FUNCTION",
                                        CICS_VAL_SINE,
                                        0,
                                        ciCsAttrFunction_ReadCallback,
                                        ciCsViString_WriteCallback));
    checkErr( Ivi_SetAttrCheckCallbackViString(vi,
                                               CICS_ATTR_TRIGGERED_FUNCTION,
                                               ciCsAttrFunction_CheckCallback));
    
    /*- CICS_ATTR_TRIGGERED_PHASE_ANGLE -*/
    checkErr( Ivi_AddRepeatedAttributeViReal64 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_TRIGGERED_PHASE_ANGLE,
                                                "CICS_ATTR_TRIGGERED_PHASE_ANGLE",
                                                0.0,
                                                0,
                                                ciCsViReal64_ReadCallback,
                                                ciCsViReal64_WriteCallback,
                                                &attrPhaseAngeleRangeTable, 0));

    /*- CICS_ATTR_TRIGGERED_SLEW_CURRENT_RATE -*/
    checkErr( Ivi_AddRepeatedAttributeViReal64 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_TRIGGERED_SLEW_CURRENT_RATE,
                                                "CICS_ATTR_TRIGGERED_SLEW_CURRENT_RATE",
                                                9.9e37,
                                                0,
                                                ciCsAttrSlewRate_ReadCallback,
                                                ciCsAttrSlewRate_WriteCallback,
                                                VI_NULL, 0));
    checkErr( Ivi_SetAttrCheckCallbackViReal64(vi,
                                               CICS_ATTR_TRIGGERED_SLEW_CURRENT_RATE,
                                               ciCsAttrSlewCurrentRate_CheckCallback));

    /*- CICS_ATTR_TRIGGERED_SLEW_FREQUENCY_RATE -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                        CICS_ATTR_TRIGGERED_SLEW_FREQUENCY_RATE,
                                        "CICS_ATTR_TRIGGERED_SLEW_FREQUENCY_RATE",
                                        9.9e37,
                                        0,
                                        ciCsAttrSlewRate_ReadCallback,
                                        ciCsAttrSlewRate_WriteCallback,
                                        VI_NULL, 0));
    checkErr( Ivi_SetAttrCheckCallbackViReal64(vi,
                                               CICS_ATTR_TRIGGERED_SLEW_FREQUENCY_RATE,
                                               ciCsAttrSlewFrequencyRate_CheckCallback));

    /*- CICS_ATTR_TRIGGER_PULSE_COUNT -*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_TRIGGER_PULSE_COUNT,
                                       "CICS_ATTR_TRIGGER_PULSE_COUNT",
                                       1,
                                       0,
                                       ciCsViInt32_ReadCallback,
                                       ciCsViInt32_WriteCallback,
                                       &attrTriggerPulseCountRangeTable));
    
    /*- CICS_ATTR_TRIGGER_PULSE_WIDTH -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                       CICS_ATTR_TRIGGER_PULSE_WIDTH,
                                       "CICS_ATTR_TRIGGER_PULSE_WIDTH",
                                       0.01667,
                                       0,
                                       ciCsViReal64_ReadCallback,
                                       ciCsViReal64_WriteCallback,
                                       VI_NULL, 0));
    checkErr( Ivi_SetAttrCheckCallbackViReal64(vi,
                                               CICS_ATTR_TRIGGER_PULSE_WIDTH,
                                               ciCsAttrTriggerDelay_CheckCallback));

    /*- CICS_ATTR_TRIGGER_PULSE_PERIOD -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                       CICS_ATTR_TRIGGER_PULSE_PERIOD,
                                       "CICS_ATTR_TRIGGER_PULSE_PERIOD",
                                       0.03333,
                                       0,
                                       ciCsViReal64_ReadCallback,
                                       ciCsViReal64_WriteCallback,
                                       VI_NULL, 0));
    checkErr( Ivi_SetAttrCheckCallbackViReal64(vi,
                                               CICS_ATTR_TRIGGER_PULSE_PERIOD,
                                               ciCsAttrTriggerDelay_CheckCallback));

    /*- CICS_ATTR_TRIGGER_LIST_COUNT -*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_TRIGGER_LIST_COUNT,
                                       "CICS_ATTR_TRIGGER_LIST_COUNT",
                                       1,
                                       0,
                                       ciCsViInt32_ReadCallback,
                                       ciCsViInt32_WriteCallback,
                                       &attrTriggerPulseCountRangeTable));
    /*- CICS_ATTR_TRIGGER_LIST_MODE -*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_TRIGGER_LIST_MODE,
                                       "CICS_ATTR_TRIGGER_LIST_MODE",
                                       1,
                                       0,
                                       ciCsEnum_ReadCallback,
                                       ciCsEnum_WriteCallback,
                                       &attrTriggerListModeRangeTable));
    
    /*- CICS_ATTR_ACQUISITION_TRIGGER_SOURCE -*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_ACQUISITION_TRIGGER_SOURCE,
                                       "CICS_ATTR_ACQUISITION_TRIGGER_SOURCE",
                                       CICS_VAL_SOFTWARE_TRIG,
                                       0,
                                       ciCsEnum_ReadCallback,
                                       ciCsEnum_WriteCallback,
                                       &attrAcquisitionTriggerSourceRangeTable));

    /*- CICS_ATTR_ACQUISITION_START_TIME -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                       CICS_ATTR_ACQUISITION_START_TIME,
                                       "CICS_ATTR_ACQUISITION_START_TIME",
                                       0.03333,
                                       0,
                                       ciCsAttrAcquisitionStartTime_ReadCallback,
                                       ciCsAttrAcquisitionStartTime_WriteCallback,
                                       VI_NULL, 0));
    checkErr( Ivi_SetAttrCheckCallbackViReal64(vi,
                                               CICS_ATTR_ACQUISITION_START_TIME,
                                               ciCsAttrAcquisitionStartTime_CheckCallback));
    checkErr( Ivi_SetAttrCoerceCallbackViReal64(vi,
                                                CICS_ATTR_ACQUISITION_START_TIME,
                                                ciCsAttrAcquisitionStartTime_CoerceCallback));

    /*- CICS_ATTR_ACQUISITION_TIME_INTERVAL -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                       CICS_ATTR_ACQUISITION_TIME_INTERVAL,
                                       "CICS_ATTR_ACQUISITION_TIME_INTERVAL",
                                       25.037e-6,
                                       0,
                                       ciCsViReal64_ReadCallback,
                                       ciCsViReal64_WriteCallback,
                                       &attrAcquisitionStartTimeRangeTable, 0));
    
    /*- CICS_ATTR_WINDOW_TYPE -*/
    checkErr( Ivi_AddRepeatedAttributeViInt32 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_WINDOW_TYPE,
                                                "CICS_ATTR_WINDOW_TYPE",
                                                CICS_VAL_WINDOW_TYPE_KBESSEL,
                                                0,
                                                ciCsEnum_ReadCallback,
                                                ciCsEnum_WriteCallback,
                                                &attrWindowTypeRangeTable));

    /*- CICS_ATTR_OUTPUT_TRIGGER_ENABLED -*/
    checkErr( Ivi_AddAttributeViBoolean (vi,
                                         CICS_ATTR_OUTPUT_TRIGGER_ENABLED,
                                         "CICS_ATTR_OUTPUT_TRIGGER_ENABLED",
                                         VI_FALSE,
                                         0,
                                         ciCsViBoolean_ReadCallback,
                                         ciCsViBoolean_WriteCallback));

    /*- CICS_ATTR_OUTPUT_TRIGGER_SOURCE -*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_OUTPUT_TRIGGER_SOURCE,
                                       "CICS_ATTR_OUTPUT_TRIGGER_SOURCE",
                                       CICS_VAL_OUTPUT_TRIGGER_SOURCE_BOT,
                                       0,
                                       ciCsEnum_ReadCallback,
                                       ciCsEnum_WriteCallback,
                                       &attrOutputTriggerSourceRangeTable));

    /*- CICS_ATTR_TRIGGER_SYNCHRONIZATION_SOURCE -*/
    checkErr( Ivi_AddAttributeViInt32 (vi,
                                       CICS_ATTR_TRIGGER_SYNCHRONIZATION_SOURCE,
                                       "CICS_ATTR_TRIGGER_SYNCHRONIZATION_SOURCE",
                                       CICS_VAL_SYNCHRONIZATION_SOURCE_IMMEDIATE,
                                       0,
                                       ciCsEnum_ReadCallback,
                                       ciCsEnum_WriteCallback,
                                       &attrTriggerSynchronizationSourceRangeTable));

    /*- CICS_ATTR_TRIGGER_SYNCHRONIZATION_PHASE -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                        CICS_ATTR_TRIGGER_SYNCHRONIZATION_PHASE,
                                        "CICS_ATTR_TRIGGER_SYNCHRONIZATION_PHASE",
                                       0.0,
                                       0,
                                       ciCsViReal64_ReadCallback,
                                       ciCsViReal64_WriteCallback,
                                       &attrPhaseAngeleRangeTable, 0));

    /*- CICS_ATTR_OUTPUT_PROTECTION_DELAY -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                        CICS_ATTR_OUTPUT_PROTECTION_DELAY,
                                        "CICS_ATTR_OUTPUT_PROTECTION_DELAY",
                                       100.0e-3,
                                       0,
                                       ciCsViReal64_ReadCallback,
                                       ciCsViReal64_WriteCallback,
                                       &attrOutputProtectionDelay, 0));
    
    /*- CICS_ATTR_VOLTAGE_RANGE_MINIMUM -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                        CICS_ATTR_VOLTAGE_RANGE_MINIMUM,
                                        "CICS_ATTR_VOLTAGE_RANGE_MINIMUM",
                                       135.0,
                                       IVI_VAL_NEVER_CACHE,
                                       ciCsViReal64_ReadCallback,
                                       ciCsViReal64_WriteCallback,
                                       VI_NULL, 0));  
                                       
    /*- CICS_ATTR_VOLTAGE_RANGE_MAXIMUM -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                        CICS_ATTR_VOLTAGE_RANGE_MAXIMUM,
                                        "CICS_ATTR_VOLTAGE_RANGE_MAXIMUM",
                                       270.0,
                                       IVI_VAL_NEVER_CACHE,
                                       ciCsViReal64_ReadCallback,
                                       ciCsViReal64_WriteCallback,
                                       VI_NULL, 0));
    
	/*- CICS_ATTR_CURRENT_MAXIMUM -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                        CICS_ATTR_CURRENT_MAXIMUM,
                                        "CICS_ATTR_CURRENT_MAXIMUM",
                                       15.0,
                                       IVI_VAL_NEVER_CACHE,
                                       ciCsViReal64_ReadCallback,
                                       ciCsViReal64_WriteCallback,
                                       VI_NULL, 0));
                                       
	/*- CICS_ATTR_VOLTAGE_MAXIMUM -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                        CICS_ATTR_VOLTAGE_MAXIMUM,
                                        "CICS_ATTR_VOLTAGE_MAXIMUM",
                                       300.0,
                                       IVI_VAL_NEVER_CACHE,
                                       ciCsViReal64_ReadCallback,
                                       ciCsViReal64_WriteCallback,
                                       VI_NULL, 0));                                       
    
	/*- CICS_ATTR_FREQUENCY_MINIMUM -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                        CICS_ATTR_FREQUENCY_MINIMUM,
                                        "CICS_ATTR_FREQUENCY_MINIMUM",
                                       16.0,
                                       IVI_VAL_NEVER_CACHE,
                                       ciCsViReal64_ReadCallback,
                                       ciCsViReal64_WriteCallback,
                                       VI_NULL, 0));
                                       
	/*- CICS_ATTR_FREQUENCY_MAXIMUM -*/
    checkErr( Ivi_AddAttributeViReal64 (vi,
                                        CICS_ATTR_FREQUENCY_MAXIMUM,
                                        "CICS_ATTR_FREQUENCY_MAXIMUM",
                                       1000.0,
                                       IVI_VAL_NEVER_CACHE,
                                       ciCsViReal64_ReadCallback,
                                       ciCsViReal64_WriteCallback,
                                       VI_NULL, 0));   
                                       
    checkErr( Ivi_AddAttributeViString (vi, CICS_ATTR_ID_QUERY_RESPONSE,
                                        "CICS_ATTR_ID_QUERY_RESPONSE", 
                                        id_buffer,
                                        IVI_VAL_NOT_USER_WRITABLE,
                                        ciCsAttrIdQueryResponse_ReadCallback,
                                        VI_NULL));
                                               
    /*- CICS_ATTR_INSTR_VOLTAGE_LIMIT_BEHAVIOR (hidden) -*/
    checkErr( Ivi_AddRepeatedAttributeViInt32 (vi,
                                               PHASE_REP_CAP_NAME,
                                               CICS_ATTR_INSTR_VOLTAGE_LIMIT_BEHAVIOR,
                                               "CICS_ATTR_INSTR_VOLTAGE_LIMIT_BEHAVIOR",
                                               CICS_VAL_VOLTAGE_REGULATE,
                                               IVI_VAL_HIDDEN,
                                               ciCsEnum_ReadCallback,
                                               ciCsEnum_WriteCallback,
                                               &attrVoltageLimitBehaviorRangeTable));
    
    /*- CICS_ATTR_OPC_TIMEOUT (hidden) -*/
    checkErr( Ivi_AddAttributeViInt32 (vi, CICS_ATTR_OPC_TIMEOUT,
                                       "CICS_ATTR_OPC_TIMEOUT",
                                       20000, IVI_VAL_HIDDEN | IVI_VAL_DONT_CHECK_STATUS,
                                       VI_NULL, VI_NULL, VI_NULL)); 
    
    /*- CICS_ATTR_INSTR_MODEL_TYPE (hidden) -*/
    checkErr( Ivi_AddAttributeViInt32 (vi, CICS_ATTR_INSTR_MODEL_TYPE,
                                       "CICS_ATTR_INSTR_MODEL_TYPE", 
                                       modelType,
                                       IVI_VAL_HIDDEN,
                                       VI_NULL,
                                       VI_NULL,
                                       VI_NULL));

    /*- CICS_ATTR_INSTR_ACTIVE_PHASE (Hidden) -*/
    checkErr( Ivi_AddAttributeViInt32 (vi, CICS_ATTR_INSTR_ACTIVE_PHASE,
                                       "CICS_ATTR_INSTR_ACTIVE_PHASE", 
                                       1, IVI_VAL_HIDDEN,
                                       ciCsViInt32_ReadCallback,
                                       ciCsViInt32_WriteCallback,
                                       0));
    /*- CICS_ATTR_INSTR_PHASE_COUPLING (hidden) -*/
    checkErr( Ivi_AddAttributeViInt32 (vi, CICS_ATTR_INSTR_PHASE_COUPLING,
                                       "CICS_ATTR_INSTR_PHASE_COUPLING", 
                                       CICS_VAL_COUPLING_ALL,
                                       IVI_VAL_HIDDEN,
                                       ciCsEnum_ReadCallback,
                                       ciCsEnum_WriteCallback,
                                       &attrInstrPhaseCouplingRangeTable));

    /*- CICS_ATTR_INSTR_CURRENT_RANGE (hidden) -*/
	checkErr( Ivi_AddRepeatedAttributeViReal64 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_INSTR_CURRENT_RANGE,
                                                "CICS_ATTR_INSTR_CURRENT_RANGE", 
                                                15.0,
                                                IVI_VAL_HIDDEN,
                                                ciCsViReal64_ReadCallback,
                                                ciCsViReal64_WriteCallback,
                                                VI_NULL, 0));                                                

    /*- CICS_ATTR_INSTR_VOLTAGE_RANGE (hidden) -*/
    checkErr( Ivi_AddRepeatedAttributeViReal64 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_INSTR_VOLTAGE_RANGE,
                                                "CICS_ATTR_INSTR_VOLTAGE_RANGE", 
                                                300.0,
                                                IVI_VAL_HIDDEN | IVI_VAL_USE_CALLBACKS_FOR_SIMULATION | IVI_VAL_NEVER_CACHE,
                                                ciCsViReal64_ReadCallback,
                                                ciCsViReal64_WriteCallback,
                                                VI_NULL, 0));      
	
	
	/*checkErr( Ivi_AddRepeatedAttributeViReal64 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_INSTR_VOLTAGE_RANGE,
                                                "CICS_ATTR_INSTR_VOLTAGE_RANGE", 
                                                300.0,
                                                IVI_VAL_HIDDEN | IVI_VAL_USE_CALLBACKS_FOR_SIMULATION | IVI_VAL_NEVER_CACHE,
                                                ciCsAttrInstrVoltageRange_ReadCallback,
                                                ciCsAttrInstrVoltageRange_WriteCallback,
                                                VI_NULL, 0));*/
	
	
	/*checkErr( Ivi_AddRepeatedAttributeViReal64 (vi,
                                                PHASE_REP_CAP_NAME,
                                                CICS_ATTR_INSTR_CURRENT_RANGE,
                                                "CICS_ATTR_INSTR_CURRENT_RANGE", 
                                                15.0,
                                                IVI_VAL_HIDDEN | IVI_VAL_USE_CALLBACKS_FOR_SIMULATION | IVI_VAL_NEVER_CACHE,
                                                ciCsAttrInstrCurrentRange_ReadCallback,
                                                ciCsAttrInstrCurrentRange_WriteCallback,
                                                VI_NULL, 0));*/
	

    /*- CICS_ATTR_INSTR_OUTPUT_PHASE_C (hidden) -*/
    checkErr( Ivi_AddAttributeViInt32 (vi, CICS_ATTR_INSTR_OUTPUT_PHASE_C,
                                       "CICS_ATTR_INSTR_OUTPUT_PHASE_C", 
                                       modelType,
                                       IVI_VAL_HIDDEN | IVI_VAL_USE_CALLBACKS_FOR_SIMULATION,
                                       ciCsViInt32_ReadCallback,
                                       VI_NULL,
                                       VI_NULL)); 
	
        /*- Setup attribute invalidations -----------------------------------*/            

    /*- CICS_ATTR_ACQUISITION_TIME_INTERVAL Invalidations -*/
    checkErr( Ivi_AddAttributeInvalidation (vi, CICS_ATTR_ACQUISITION_TIME_INTERVAL, 
                                            CICS_ATTR_ACQUISITION_START_TIME, 
                                            VI_FALSE));

    checkErr( ciCs_SetModelSpecificCallbacks(vi, modelType));

Error:
    return error;
}

/*****************************************************************************
 * Function: ciCs_SetModelSpecificCallbacks                                                  
 * Purpose:  This function sets specific callbacks for specified model
 *****************************************************************************/
 static ViStatus ciCs_SetModelSpecificCallbacks(
    ViSession   vi,
    ViInt32     modelType
 )
 {
    ViStatus    error = VI_SUCCESS;
    ViInt32     model_type = modelType;
    
    if (IS_4500Cs)
	{
		checkErr( Ivi_SetAttrWriteCallbackViString ( vi,
                                                    CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT,
                                                    VI_NULL));
        checkErr( Ivi_SetAttrWriteCallbackViString ( vi,
                                                   CICS_ATTR_INSTR_OUTPUT_PHASE_COUNT,
                                                   VI_NULL));
    }
    else
    {
        checkErr( Ivi_SetAttrCheckCallbackViBoolean(vi,
                                                    CICS_ATTR_OVP_ENABLED,
                                                    ciCsAttrOvpEnabled_CheckCallback));
        checkErr( Ivi_SetAttrCheckCallbackViReal64( vi,
                                                    CICS_ATTR_OVP_LIMIT,
                                                    ciCsPhaseNameViReal64_CheckCallback));
        checkErr( Ivi_SetAttrCheckCallbackViReal64( vi,
                                                    CICS_ATTR_TRIGGERED_CURRENT_LEVEL,
                                                    ciCsPhaseNameViReal64_CheckCallback));
    }
    
Error:
    return error;
 }

/*****************************************************************************
 *------------------- End Instrument Driver Source Code ---------------------*
 *****************************************************************************/



